/******************************************************************************
 * Top contributors (to current version):
 *   Gereon Kremer
 *
 * This file is part of the cvc5 project.
 *
 * Copyright (c) 2009-2025 by the authors listed in the file AUTHORS
 * in the top-level source directory and their institutional affiliations.
 * All rights reserved.  See the file COPYING in the top-level source
 * directory for licensing information.
 * ****************************************************************************
 *
 * Option template for option modules.
 *
 * For each <module>_options.toml configuration file, mkoptions.py
 * expands this template and generates a <module>_options.cpp file.
 */
#include "options/theory_options.h"

#include <iostream>

#include "base/check.h"
#include "options/option_exception.h"

namespace cvc5::internal::options {

// clang-format off
std::ostream& operator<<(std::ostream& os, ConflictProcessMode mode)
{
  switch(mode)
  {
    case ConflictProcessMode::NONE: return os << "none";
    case ConflictProcessMode::MINIMIZE: return os << "min";
    case ConflictProcessMode::MINIMIZE_EXT: return os << "min-ext";
    default: Unreachable();
  }
  return os;
}
ConflictProcessMode stringToConflictProcessMode(const std::string& optarg)
{
  if (optarg == "none") return ConflictProcessMode::NONE;
  else if (optarg == "min") return ConflictProcessMode::MINIMIZE;
  else if (optarg == "min-ext") return ConflictProcessMode::MINIMIZE_EXT;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Defines mode for processing theory conflicts.
Available modes for --conflict-process are:
+ none (default)
  Do not post-process conflicts from theory solvers.
+ min
  Do simple minimization for conflicts from theory solvers.
+ min-ext
  Do minimization for conflicts from theory solvers, relying on the extended
  rewriter.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --conflict-process: `") +
                        optarg + "'.  Try --conflict-process=help.");
}
std::ostream& operator<<(std::ostream& os, DefaultFunctionValueMode mode)
{
  switch(mode)
  {
    case DefaultFunctionValueMode::FIRST: return os << "first";
    case DefaultFunctionValueMode::FIRST_ENUM: return os << "first-enum";
    case DefaultFunctionValueMode::HOLE: return os << "hole";
    default: Unreachable();
  }
  return os;
}
DefaultFunctionValueMode stringToDefaultFunctionValueMode(const std::string& optarg)
{
  if (optarg == "first") return DefaultFunctionValueMode::FIRST;
  else if (optarg == "first-enum") return DefaultFunctionValueMode::FIRST_ENUM;
  else if (optarg == "hole") return DefaultFunctionValueMode::HOLE;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Defines mode for choosing default values for functions.
Available modes for --default-function-value-mode are:
+ first (default)
  The default value is equal to the value of the first application found of that
  function when applicable, or the first enumerated term otherwise.
+ first-enum
  The default value is equal to the first enumerated value of its range type.
+ hole
  The default value is equal to a distinguished variable.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --default-function-value-mode: `") +
                        optarg + "'.  Try --default-function-value-mode=help.");
}
std::ostream& operator<<(std::ostream& os, EqEngineMode mode)
{
  switch(mode)
  {
    case EqEngineMode::DISTRIBUTED: return os << "distributed";
    case EqEngineMode::CENTRAL: return os << "central";
    default: Unreachable();
  }
  return os;
}
EqEngineMode stringToEqEngineMode(const std::string& optarg)
{
  if (optarg == "distributed") return EqEngineMode::DISTRIBUTED;
  else if (optarg == "central") return EqEngineMode::CENTRAL;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Defines mode for managing equalities across theory solvers.
Available modes for --ee-mode are:
+ distributed (default)
  Each theory maintains its own equality engine.
+ central
  All applicable theories use the central equality engine.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --ee-mode: `") +
                        optarg + "'.  Try --ee-mode=help.");
}
std::ostream& operator<<(std::ostream& os, LemmaInprocessMode mode)
{
  switch(mode)
  {
    case LemmaInprocessMode::FULL: return os << "full";
    case LemmaInprocessMode::LIGHT: return os << "light";
    case LemmaInprocessMode::NONE: return os << "none";
    default: Unreachable();
  }
  return os;
}
LemmaInprocessMode stringToLemmaInprocessMode(const std::string& optarg)
{
  if (optarg == "full") return LemmaInprocessMode::FULL;
  else if (optarg == "light") return LemmaInprocessMode::LIGHT;
  else if (optarg == "none") return LemmaInprocessMode::NONE;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Modes for inprocessing lemmas.
Available modes for --lemma-inprocess are:
+ full
  Full inprocessing.
+ light
  Light inprocessing.
+ none (default)
  No inprocessing.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --lemma-inprocess: `") +
                        optarg + "'.  Try --lemma-inprocess=help.");
}
std::ostream& operator<<(std::ostream& os, LemmaInprocessSubsMode mode)
{
  switch(mode)
  {
    case LemmaInprocessSubsMode::ALL: return os << "all";
    case LemmaInprocessSubsMode::SIMPLE: return os << "simple";
    default: Unreachable();
  }
  return os;
}
LemmaInprocessSubsMode stringToLemmaInprocessSubsMode(const std::string& optarg)
{
  if (optarg == "all") return LemmaInprocessSubsMode::ALL;
  else if (optarg == "simple") return LemmaInprocessSubsMode::SIMPLE;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Modes for substitutions for inprocessing lemmas.
Available modes for --lemma-inprocess-subs are:
+ all
  All substitutions.
+ simple (default)
  Simple substitutions.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --lemma-inprocess-subs: `") +
                        optarg + "'.  Try --lemma-inprocess-subs=help.");
}
std::ostream& operator<<(std::ostream& os, TcMode mode)
{
  switch(mode)
  {
    case TcMode::CARE_GRAPH: return os << "care-graph";
    default: Unreachable();
  }
  return os;
}
TcMode stringToTcMode(const std::string& optarg)
{
  if (optarg == "care-graph") return TcMode::CARE_GRAPH;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Defines mode for theory combination.
Available modes for --tc-mode are:
+ care-graph (default)
  Use care graphs for theory combination.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --tc-mode: `") +
                        optarg + "'.  Try --tc-mode=help.");
}
std::ostream& operator<<(std::ostream& os, TheoryOfMode mode)
{
  switch(mode)
  {
    case TheoryOfMode::THEORY_OF_TYPE_BASED: return os << "type";
    case TheoryOfMode::THEORY_OF_TERM_BASED: return os << "term";
    default: Unreachable();
  }
  return os;
}
TheoryOfMode stringToTheoryOfMode(const std::string& optarg)
{
  if (optarg == "type") return TheoryOfMode::THEORY_OF_TYPE_BASED;
  else if (optarg == "term") return TheoryOfMode::THEORY_OF_TERM_BASED;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Defines how we associate theories with terms.
Available modes for --theoryof-mode are:
+ type (default)
  Type variables, constants and equalities by type.
+ term
  Type variables as uninterpreted, type constants by theory, equalities by the
  parametric theory.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --theoryof-mode: `") +
                        optarg + "'.  Try --theoryof-mode=help.");
}
// clang-format on

}  // namespace cvc5::internal::options
