/******************************************************************************
 * Top contributors (to current version):
 *   Gereon Kremer, Mathias Preiner, Alex Ozdemir
 *
 * This file is part of the cvc5 project.
 *
 * Copyright (c) 2009-2025 by the authors listed in the file AUTHORS
 * in the top-level source directory and their institutional affiliations.
 * All rights reserved.  See the file COPYING in the top-level source
 * directory for licensing information.
 * ****************************************************************************
 *
 * Global (command-line, set-option, ...) parameters for SMT.
 */

#include "base/check.h"
#include "base/output.h"
#include "options/io_utils.h"
#include "options/options.h"
#include "options/options_handler.h"
#include "options/options_listener.h"
#include "options/options_public.h"
#include "options/uf_options.h"

// clang-format off
#include "options/arith_options.h"
#include "options/arrays_options.h"
#include "options/bags_options.h"
#include "options/base_options.h"
#include "options/booleans_options.h"
#include "options/builtin_options.h"
#include "options/bv_options.h"
#include "options/datatypes_options.h"
#include "options/decision_options.h"
#include "options/expr_options.h"
#include "options/ff_options.h"
#include "options/fp_options.h"
#include "options/main_options.h"
#include "options/parallel_options.h"
#include "options/parser_options.h"
#include "options/printer_options.h"
#include "options/proof_options.h"
#include "options/prop_options.h"
#include "options/quantifiers_options.h"
#include "options/sep_options.h"
#include "options/sets_options.h"
#include "options/smt_options.h"
#include "options/strings_options.h"
#include "options/theory_options.h"
#include "options/uf_options.h"
#include "options/language.h"
#include <iostream>
#include "options/managed_streams.h"
#include <unordered_map>
#include <bitset>
// clang-format on

#include <cstring>
#include <iostream>
#include <limits>

namespace {
  // clang-format off
  enum class OptionEnum {
      ABSTRACT_VALUES,
      ACKERMANN,
      APPEND_LEARNED_LITERALS_TO_CUBES,
      APPROX_BRANCH_DEPTH,
      ARITH_BRAB,
      ARITH_EQ_SOLVER,
      ARITH_EXP,
      ARITH_PROP,
      ARITH_PROP_CLAUSES,
      ARITH_REWRITE_EQUALITIES,
      ARITH_STATIC_LEARNING,
      ARRAYS_EAGER_INDEX,
      ARRAYS_EAGER_LEMMAS,
      ARRAYS_EXP,
      ARRAYS_OPTIMIZE_LINEAR,
      ARRAYS_PROP,
      ARRAYS_REDUCE_SHARING,
      ARRAYS_WEAK_EQUIV,
      ASSIGN_FUNCTION_VALUES,
      BAGS,
      BITBLAST,
      BITWISE_EQ,
      BOOL_TO_BV,
      BV_ASSERT_INPUT,
      BV_EAGER_EVAL,
      BV_EQ_ENGINE,
      BV_GAUSS_ELIM,
      BV_INTRO_POW2,
      BV_PRINT_CONSTS_AS_INDEXED_SYMBOLS,
      BV_PROPAGATE,
      BV_RW_EXTEND_EQ,
      BV_SAT_SOLVER,
      BV_SOLVER,
      BV_TO_BOOL,
      BV_TO_INT_USE_POW2,
      BVAND_INTEGER_GRANULARITY,
      CBQI,
      CBQI_ALL_CONFLICT,
      CBQI_MODE,
      CBQI_SKIP_RD,
      CBQI_TCONSTRAINT,
      CBQI_VO_EXP,
      CDT_BISIMILAR,
      CEGIS_SAMPLE,
      CEGQI,
      CEGQI_ALL,
      CEGQI_BV,
      CEGQI_BV_CONCAT_INV,
      CEGQI_BV_INEQ,
      CEGQI_BV_INTERLEAVE_VALUE,
      CEGQI_BV_LINEAR,
      CEGQI_BV_RM_EXTRACT,
      CEGQI_BV_SOLVE_NL,
      CEGQI_FULL,
      CEGQI_INF_INT,
      CEGQI_INF_REAL,
      CEGQI_INNERMOST,
      CEGQI_MIDPOINT,
      CEGQI_MIN_BOUNDS,
      CEGQI_MULTI_INST,
      CEGQI_NESTED_QE,
      CEGQI_NOPT,
      CEGQI_ROUND_UP_LIA,
      CHECK_ABDUCTS,
      CHECK_INTERPOLANTS,
      CHECK_MODELS,
      CHECK_PROOF_STEPS,
      CHECK_PROOFS,
      CHECK_SYNTH_SOL,
      CHECK_UNSAT_CORES,
      CHECKS_BEFORE_PARTITION,
      CHECKS_BETWEEN_PARTITIONS,
      COLLECT_PIVOT_STATS,
      COMPUTE_PARTITIONS,
      COND_VAR_SPLIT_QUANT,
      CONDENSE_FUNCTION_VALUES,
      CONFLICT_PROCESS,
      CONJECTURE_GEN,
      CONJECTURE_GEN_GT_ENUM,
      CONJECTURE_GEN_MAX_DEPTH,
      CONJECTURE_GEN_PER_ROUND,
      CONS_EXP_TRIGGERS,
      COPYRIGHT,
      CUT_ALL_BOUNDED,
      DAG_THRESH,
      DATATYPES_EXP,
      DEBUG_CHECK_MODELS,
      DECISION,
      DEEP_RESTART,
      DEEP_RESTART_FACTOR,
      DEFAULT_FUNCTION_VALUE_MODE,
      DIFFICULTY_MODE,
      DIO_DECOMPS,
      DIO_SOLVER,
      DIO_TURNS,
      DRAT_BINARY_FORMAT,
      DT_BINARY_SPLIT,
      DT_BLAST_SPLITS,
      DT_CYCLIC,
      DT_INFER_AS_LEMMAS,
      DT_NESTED_REC,
      DT_POLITE_OPTIMIZE,
      DT_SHARE_SEL,
      DT_STC_IND,
      DT_VAR_EXP_QUANT,
      DUMP_DIFFICULTY,
      DUMP_INSTANTIATIONS,
      DUMP_INSTANTIATIONS_DEBUG,
      DUMP_MODELS,
      DUMP_PROOFS,
      DUMP_UNSAT_CORES,
      DUMP_UNSAT_CORES_LEMMAS,
      E_MATCHING,
      EAGER_ARITH_BV_CONV,
      EARLY_EXIT,
      EARLY_ITE_REMOVAL,
      EE_MODE,
      ELIM_TAUT_QUANT,
      ENUM_INST,
      ENUM_INST_INTERLEAVE,
      ENUM_INST_LIMIT,
      ENUM_INST_RD,
      ENUM_INST_STRATIFY,
      ENUM_INST_SUM,
      ERR,
      ERROR_SELECTION_RULE,
      EXPR_DEPTH,
      EXT_REW_PREP,
      EXT_REWRITE_QUANT,
      FC_PENALTIES,
      FF,
      FF_BITSUM,
      FF_ELIM_DISJUNCTIVE_BIT,
      FF_FIELD_POLYS,
      FF_SOLVER,
      FF_TRACE_GB,
      FILENAME,
      FILESYSTEM_ACCESS,
      FINITE_MODEL_FIND,
      FLATTEN_HO_CHAINS,
      FMF_BOUND,
      FMF_BOUND_BLAST,
      FMF_BOUND_LAZY,
      FMF_FUN,
      FMF_FUN_RLV,
      FMF_MBQI,
      FMF_TYPE_COMPLETION_THRESH,
      FORCE_LOGIC,
      FORCE_NO_LIMIT_CPU_WHILE_DUMP,
      FOREIGN_THEORY_REWRITE,
      FP,
      FP_EXP,
      FP_LAZY_WB,
      FRESH_BINDERS,
      FRESH_DECLARATIONS,
      FULL_SATURATE_QUANT,
      FULL_SYGUS_VERIFY,
      GLOBAL_DECLARATIONS,
      GLOBAL_NEGATE,
      HELP,
      HELP_REGULAR,
      HEURISTIC_PIVOTS,
      HO_ELIM,
      HO_ELIM_STORE_AX,
      HO_MATCHING,
      HO_MERGE_TERM_DB,
      IAND_MODE,
      IEVAL,
      IN,
      INCREMENT_TRIGGERS,
      INCREMENTAL,
      INST_MAX_LEVEL,
      INST_MAX_ROUNDS,
      INST_NO_ENTAIL,
      INST_WHEN,
      INST_WHEN_PHASE,
      INT_WF_IND,
      INTERACTIVE,
      INTERPOLANTS_MODE,
      ITE_DTT_SPLIT_QUANT,
      ITE_LIFT_QUANT,
      ITE_SIMP,
      JH_RLV_ORDER,
      JH_SKOLEM,
      JH_SKOLEM_RLV,
      LANG,
      LEARNED_REWRITE,
      LEMMA_INPROCESS,
      LEMMA_INPROCESS_INFER_EQ_LIT,
      LEMMA_INPROCESS_SUBS,
      LEMMAS_ON_REPLAY_FAILURE,
      LFSC_EXPAND_TRUST,
      LFSC_FLATTEN,
      LITERAL_MATCHING,
      MACROS_QUANT,
      MACROS_QUANT_MODE,
      MAXCUTSINCONTEXT,
      MBQI,
      MBQI_FAST_SYGUS,
      MBQI_INTERLEAVE,
      MBQI_ONE_INST_PER_ROUND,
      MINIMAL_UNSAT_CORES,
      MINISAT_DUMP_DIMACS,
      MINISAT_SIMPLIFICATION,
      MINISCOPE_QUANT,
      MINISCOPE_QUANT_USER,
      MIPLIB_TRICK,
      MIPLIB_TRICK_SUBS,
      MODEL_BASED_ARITH_BV_CONV,
      MODEL_CORES,
      MODEL_U_PRINT,
      MODEL_VAR_ELIM_UNEVAL,
      MULTI_TRIGGER_CACHE,
      MULTI_TRIGGER_LINEAR,
      MULTI_TRIGGER_PRIORITY,
      MULTI_TRIGGER_WHEN_SINGLE,
      NEW_PROP,
      NL_COV,
      NL_COV_FORCE,
      NL_COV_LIFT,
      NL_COV_LINEAR_MODEL,
      NL_COV_PROJ,
      NL_COV_PRUNE,
      NL_COV_VAR_ELIM,
      NL_EXT,
      NL_EXT_ENT_CONF,
      NL_EXT_FACTOR,
      NL_EXT_INC_PREC,
      NL_EXT_PURIFY,
      NL_EXT_RBOUND,
      NL_EXT_REWRITE,
      NL_EXT_SPLIT_ZERO,
      NL_EXT_TF_TAYLOR_DEG,
      NL_EXT_TF_TPLANES,
      NL_EXT_TPLANES,
      NL_EXT_TPLANES_INTERLEAVE,
      NL_ICP,
      NL_RLV,
      NL_RLV_ASSERT_BOUNDS,
      ON_REPEAT_ITE_SIMP,
      OPT_RES_RECONSTRUCTION_SIZE,
      ORACLES,
      OUT,
      OUTPUT,
      OUTPUT_LANG,
      PARSE_ONLY,
      PARSE_SKOLEM_DEFINITIONS,
      PARSING_MODE,
      PARTIAL_TRIGGERS,
      PARTITION_CHECK,
      PARTITION_CONFLICT_SIZE,
      PARTITION_START_TIME,
      PARTITION_STRATEGY,
      PARTITION_TIME_INTERVAL,
      PARTITION_TLIMIT,
      PARTITION_WHEN,
      PB_REWRITES,
      PIVOT_THRESHOLD,
      PLUGIN_NOTIFY_SAT_CLAUSE_IN_SOLVE,
      PLUGIN_SHARE_SKOLEMS,
      POOL_INST,
      PORTFOLIO_JOBS,
      PP_ASSERT_MAX_SUB_SIZE,
      PRE_SKOLEM_QUANT,
      PRE_SKOLEM_QUANT_NESTED,
      PRENEX_QUANT,
      PRENEX_QUANT_USER,
      PREPROCESS_ONLY,
      PREREGISTER_MODE,
      PRINT_ARITH_LIT_TOKEN,
      PRINT_CORES_FULL,
      PRINT_DOT_CLUSTERS,
      PRINT_INST,
      PRINT_INST_FULL,
      PRINT_SKOLEM_DEFINITIONS,
      PRINT_SUCCESS,
      PRODUCE_ABDUCTS,
      PRODUCE_ASSERTIONS,
      PRODUCE_ASSIGNMENTS,
      PRODUCE_DIFFICULTY,
      PRODUCE_INTERPOLANTS,
      PRODUCE_LEARNED_LITERALS,
      PRODUCE_MODELS,
      PRODUCE_PROOFS,
      PRODUCE_UNSAT_ASSUMPTIONS,
      PRODUCE_UNSAT_CORES,
      PROOF_ALETHE_DEFINE_SKOLEMS,
      PROOF_ALETHE_RES_PIVOTS,
      PROOF_ALLOW_TRUST,
      PROOF_CHECK,
      PROOF_DAG_GLOBAL,
      PROOF_DOT_DAG,
      PROOF_ELIM_SUBTYPES,
      PROOF_FORMAT_MODE,
      PROOF_GRANULARITY,
      PROOF_MODE,
      PROOF_PEDANTIC,
      PROOF_PP_MERGE,
      PROOF_PRINT_CONCLUSION,
      PROOF_PRINT_REFERENCE,
      PROOF_PRUNE_INPUT,
      PROOF_REWRITE_RCONS_REC_LIMIT,
      PROOF_REWRITE_RCONS_STEP_LIMIT,
      PROP_PROOF_MODE,
      PROP_ROW_LENGTH,
      PURIFY_TRIGGERS,
      QUANT_ALPHA_EQUIV,
      QUANT_DSPLIT,
      QUANT_FUN_WD,
      QUANT_IND,
      QUANT_REP_MODE,
      QUIET,
      RANDOM_FREQ,
      RANDOM_PARTITIONING,
      RE_ELIM,
      RE_INTER_MODE,
      REGISTER_QUANT_BODY_TERMS,
      RELATIONAL_TRIGGERS,
      RELEVANCE_FILTER,
      RELEVANT_TRIGGERS,
      RELS_EXP,
      REPEAT_SIMP,
      REPLAY_EARLY_CLOSE_DEPTH,
      REPLAY_LEMMA_REJECT_CUT,
      REPLAY_NUM_ERR_PENALTY,
      REPLAY_REJECT_CUT,
      RESTART_INT_BASE,
      RESTART_INT_INC,
      RESTRICT_PIVOTS,
      REVERT_ARITH_MODELS_ON_UNSAT,
      RLIMIT,
      RLIMIT_PER,
      RR_TURNS,
      RWEIGHT,
      SAFE_OPTIONS,
      SAT_PROOF_MIN_DIMACS,
      SAT_RANDOM_SEED,
      SAT_SOLVER,
      SE_SOLVE_INT,
      SEED,
      SEGV_SPIN,
      SEMANTIC_CHECKS,
      SEP,
      SEP_MIN_REFINE,
      SEP_PRE_SKOLEM_EMP,
      SEQ_ARRAY,
      SETS_CARD_EXP,
      SETS_EXP,
      SETS_PROXY_LEMMAS,
      SHOW_CONFIG,
      SHOW_TRACE_TAGS,
      SIMP_ITE_COMPRESS,
      SIMP_WITH_CARE,
      SIMPLEX_CHECK_PERIOD,
      SIMPLIFICATION,
      SIMPLIFICATION_BCP,
      SOI_QE,
      SOLVE_BV_AS_INT,
      SOLVE_INT_AS_BV,
      SOLVE_REAL_AS_INT,
      SORT_INFERENCE,
      STANDARD_EFFORT_VARIABLE_ORDER_PIVOTS,
      STATIC_LEARNING,
      STATS,
      STATS_ALL,
      STATS_EVERY_QUERY,
      STATS_INTERNAL,
      STRICT_PARSING,
      STRINGS_ALPHA_CARD,
      STRINGS_CHECK_ENTAIL_LEN,
      STRINGS_DEQ_EXT,
      STRINGS_EAGER_EVAL,
      STRINGS_EAGER_LEN_RE,
      STRINGS_EAGER_REG,
      STRINGS_EAGER_SOLVER,
      STRINGS_EXP,
      STRINGS_FF,
      STRINGS_FMF,
      STRINGS_INFER_AS_LEMMAS,
      STRINGS_INFER_SYM,
      STRINGS_LAZY_PP,
      STRINGS_LEN_NORM,
      STRINGS_MBR,
      STRINGS_MODEL_MAX_LEN,
      STRINGS_PROCESS_LOOP_MODE,
      STRINGS_RE_DERIVE_CONF,
      STRINGS_RE_POSC_EAGER,
      STRINGS_REC_ARITH_APPROX,
      STRINGS_REGEXP_INCLUSION,
      STRINGS_REXPLAIN_LEMMAS,
      SUB_CBQI,
      SUB_CBQI_TIMEOUT,
      SYGUS,
      SYGUS_ABORT_SIZE,
      SYGUS_ADD_CONST_GRAMMAR,
      SYGUS_ARG_RELEVANT,
      SYGUS_AUTO_UNFOLD,
      SYGUS_BOOL_ITE_RETURN_CONST,
      SYGUS_CORE_CONNECTIVE,
      SYGUS_CREPAIR_ABORT,
      SYGUS_ENUM,
      SYGUS_ENUM_FAST_NUM_CONSTS,
      SYGUS_ENUM_RANDOM_P,
      SYGUS_EVAL_UNFOLD,
      SYGUS_EXPR_MINER_CHECK_TIMEOUT,
      SYGUS_FAIR,
      SYGUS_FAIR_MAX,
      SYGUS_FILTER_SOL,
      SYGUS_FILTER_SOL_REV,
      SYGUS_GRAMMAR_CONS,
      SYGUS_GRAMMAR_NORM,
      SYGUS_GRAMMAR_USE_DISJ,
      SYGUS_INFERENCE,
      SYGUS_INST,
      SYGUS_INST_MODE,
      SYGUS_INST_SCOPE,
      SYGUS_INST_TERM_SEL,
      SYGUS_INV_TEMPL,
      SYGUS_INV_TEMPL_WHEN_SG,
      SYGUS_MIN_GRAMMAR,
      SYGUS_OUT,
      SYGUS_PBE,
      SYGUS_PBE_MULTI_FAIR,
      SYGUS_PBE_MULTI_FAIR_DIFF,
      SYGUS_QE_PREPROC,
      SYGUS_QUERY_GEN,
      SYGUS_QUERY_GEN_DUMP_FILES,
      SYGUS_QUERY_GEN_FILTER_SOLVED,
      SYGUS_QUERY_GEN_THRESH,
      SYGUS_REC_FUN,
      SYGUS_REC_FUN_EVAL_LIMIT,
      SYGUS_REPAIR_CONST,
      SYGUS_REPAIR_CONST_TIMEOUT,
      SYGUS_REWRITER,
      SYGUS_RR_SYNTH_ACCEL,
      SYGUS_RR_SYNTH_CHECK,
      SYGUS_RR_SYNTH_FILTER_CONG,
      SYGUS_RR_SYNTH_FILTER_MATCH,
      SYGUS_RR_SYNTH_FILTER_NL,
      SYGUS_RR_SYNTH_FILTER_ORDER,
      SYGUS_RR_SYNTH_INPUT_NVARS,
      SYGUS_RR_SYNTH_REC,
      SYGUS_SAMPLE_FP_UNIFORM,
      SYGUS_SAMPLE_GRAMMAR,
      SYGUS_SAMPLES,
      SYGUS_SI,
      SYGUS_SI_ABORT,
      SYGUS_SI_RCONS,
      SYGUS_SI_RCONS_LIMIT,
      SYGUS_SI_RCONS_P,
      SYGUS_SIMPLE_SYM_BREAK,
      SYGUS_STREAM,
      SYGUS_SYM_BREAK_LAZY,
      SYGUS_SYM_BREAK_PBE,
      SYGUS_SYM_BREAK_RLV,
      SYGUS_UNIF_COND_INDEPENDENT_NO_REPEAT_SOL,
      SYGUS_UNIF_PI,
      SYGUS_UNIF_SHUFFLE_COND,
      SYGUS_VERIFY_INST_MAX_ROUNDS,
      SYGUS_VERIFY_TIMEOUT,
      SYMMETRY_BREAKER,
      TC_MODE,
      TERM_DB_MODE,
      TERM_SORT_OVERLOAD,
      THEORYOF_MODE,
      TIMEOUT_CORE_TIMEOUT,
      TLIMIT,
      TLIMIT_PER,
      TRACE,
      TRIGGER_ACTIVE_SEL,
      TRIGGER_SEL,
      TYPE_CHECKING,
      UF_CARD_EXP,
      UF_HO_EXP,
      UF_HO_EXT,
      UF_LAMBDA_QE,
      UF_LAZY_LL,
      UF_SS,
      UF_SS_ABORT_CARD,
      UF_SS_FAIR,
      UF_SS_FAIR_MONOTONE,
      UNATE_LEMMAS,
      UNCONSTRAINED_SIMP,
      UNSAT_CORES_MODE,
      USE_APPROX,
      USE_FCSIMPLEX,
      USE_PORTFOLIO,
      USE_SOI,
      USER_PAT,
      USER_POOL,
      VAR_ELIM_QUANT,
      VAR_INEQ_ELIM_QUANT,
      VERBOSE,
      VERBOSITY,
      VERSION,
      WF_CHECKING,
      WRITE_PARTITIONS_TO,
    };
    const std::unordered_map<std::string, OptionEnum> NAME_TO_ENUM = {
      { "abstract-values", OptionEnum::ABSTRACT_VALUES },
      { "ackermann", OptionEnum::ACKERMANN },
      { "append-learned-literals-to-cubes", OptionEnum::APPEND_LEARNED_LITERALS_TO_CUBES },
      { "approx-branch-depth", OptionEnum::APPROX_BRANCH_DEPTH },
      { "arith-brab", OptionEnum::ARITH_BRAB },
      { "arith-eq-solver", OptionEnum::ARITH_EQ_SOLVER },
      { "arith-exp", OptionEnum::ARITH_EXP },
      { "arith-prop", OptionEnum::ARITH_PROP },
      { "arith-prop-clauses", OptionEnum::ARITH_PROP_CLAUSES },
      { "arith-rewrite-equalities", OptionEnum::ARITH_REWRITE_EQUALITIES },
      { "arith-static-learning", OptionEnum::ARITH_STATIC_LEARNING },
      { "arrays-eager-index", OptionEnum::ARRAYS_EAGER_INDEX },
      { "arrays-eager-lemmas", OptionEnum::ARRAYS_EAGER_LEMMAS },
      { "arrays-exp", OptionEnum::ARRAYS_EXP },
      { "arrays-optimize-linear", OptionEnum::ARRAYS_OPTIMIZE_LINEAR },
      { "arrays-prop", OptionEnum::ARRAYS_PROP },
      { "arrays-reduce-sharing", OptionEnum::ARRAYS_REDUCE_SHARING },
      { "arrays-weak-equiv", OptionEnum::ARRAYS_WEAK_EQUIV },
      { "assign-function-values", OptionEnum::ASSIGN_FUNCTION_VALUES },
      { "bags", OptionEnum::BAGS },
      { "bitblast", OptionEnum::BITBLAST },
      { "bitwise-eq", OptionEnum::BITWISE_EQ },
      { "bool-to-bv", OptionEnum::BOOL_TO_BV },
      { "bv-assert-input", OptionEnum::BV_ASSERT_INPUT },
      { "bv-eager-eval", OptionEnum::BV_EAGER_EVAL },
      { "bv-eq-engine", OptionEnum::BV_EQ_ENGINE },
      { "bv-gauss-elim", OptionEnum::BV_GAUSS_ELIM },
      { "bv-intro-pow2", OptionEnum::BV_INTRO_POW2 },
      { "bv-print-consts-as-indexed-symbols", OptionEnum::BV_PRINT_CONSTS_AS_INDEXED_SYMBOLS },
      { "bv-propagate", OptionEnum::BV_PROPAGATE },
      { "bv-rw-extend-eq", OptionEnum::BV_RW_EXTEND_EQ },
      { "bv-sat-solver", OptionEnum::BV_SAT_SOLVER },
      { "bv-solver", OptionEnum::BV_SOLVER },
      { "bv-to-bool", OptionEnum::BV_TO_BOOL },
      { "bv-to-int-use-pow2", OptionEnum::BV_TO_INT_USE_POW2 },
      { "bvand-integer-granularity", OptionEnum::BVAND_INTEGER_GRANULARITY },
      { "cbqi", OptionEnum::CBQI },
      { "cbqi-all-conflict", OptionEnum::CBQI_ALL_CONFLICT },
      { "cbqi-mode", OptionEnum::CBQI_MODE },
      { "cbqi-skip-rd", OptionEnum::CBQI_SKIP_RD },
      { "cbqi-tconstraint", OptionEnum::CBQI_TCONSTRAINT },
      { "cbqi-vo-exp", OptionEnum::CBQI_VO_EXP },
      { "cdt-bisimilar", OptionEnum::CDT_BISIMILAR },
      { "cegis-sample", OptionEnum::CEGIS_SAMPLE },
      { "cegqi", OptionEnum::CEGQI },
      { "cegqi-all", OptionEnum::CEGQI_ALL },
      { "cegqi-bv", OptionEnum::CEGQI_BV },
      { "cegqi-bv-concat-inv", OptionEnum::CEGQI_BV_CONCAT_INV },
      { "cegqi-bv-ineq", OptionEnum::CEGQI_BV_INEQ },
      { "cegqi-bv-interleave-value", OptionEnum::CEGQI_BV_INTERLEAVE_VALUE },
      { "cegqi-bv-linear", OptionEnum::CEGQI_BV_LINEAR },
      { "cegqi-bv-rm-extract", OptionEnum::CEGQI_BV_RM_EXTRACT },
      { "cegqi-bv-solve-nl", OptionEnum::CEGQI_BV_SOLVE_NL },
      { "cegqi-full", OptionEnum::CEGQI_FULL },
      { "cegqi-inf-int", OptionEnum::CEGQI_INF_INT },
      { "cegqi-inf-real", OptionEnum::CEGQI_INF_REAL },
      { "cegqi-innermost", OptionEnum::CEGQI_INNERMOST },
      { "cegqi-midpoint", OptionEnum::CEGQI_MIDPOINT },
      { "cegqi-min-bounds", OptionEnum::CEGQI_MIN_BOUNDS },
      { "cegqi-multi-inst", OptionEnum::CEGQI_MULTI_INST },
      { "cegqi-nested-qe", OptionEnum::CEGQI_NESTED_QE },
      { "cegqi-nopt", OptionEnum::CEGQI_NOPT },
      { "cegqi-round-up-lia", OptionEnum::CEGQI_ROUND_UP_LIA },
      { "check-abducts", OptionEnum::CHECK_ABDUCTS },
      { "check-interpolants", OptionEnum::CHECK_INTERPOLANTS },
      { "check-models", OptionEnum::CHECK_MODELS },
      { "check-proof-steps", OptionEnum::CHECK_PROOF_STEPS },
      { "check-proofs", OptionEnum::CHECK_PROOFS },
      { "check-synth-sol", OptionEnum::CHECK_SYNTH_SOL },
      { "check-unsat-cores", OptionEnum::CHECK_UNSAT_CORES },
      { "checks-before-partition", OptionEnum::CHECKS_BEFORE_PARTITION },
      { "checks-between-partitions", OptionEnum::CHECKS_BETWEEN_PARTITIONS },
      { "collect-pivot-stats", OptionEnum::COLLECT_PIVOT_STATS },
      { "compute-partitions", OptionEnum::COMPUTE_PARTITIONS },
      { "cond-var-split-quant", OptionEnum::COND_VAR_SPLIT_QUANT },
      { "condense-function-values", OptionEnum::CONDENSE_FUNCTION_VALUES },
      { "conflict-process", OptionEnum::CONFLICT_PROCESS },
      { "conjecture-gen", OptionEnum::CONJECTURE_GEN },
      { "conjecture-gen-gt-enum", OptionEnum::CONJECTURE_GEN_GT_ENUM },
      { "conjecture-gen-max-depth", OptionEnum::CONJECTURE_GEN_MAX_DEPTH },
      { "conjecture-gen-per-round", OptionEnum::CONJECTURE_GEN_PER_ROUND },
      { "cons-exp-triggers", OptionEnum::CONS_EXP_TRIGGERS },
      { "copyright", OptionEnum::COPYRIGHT },
      { "cut-all-bounded", OptionEnum::CUT_ALL_BOUNDED },
      { "dag-thresh", OptionEnum::DAG_THRESH },
      { "datatypes-exp", OptionEnum::DATATYPES_EXP },
      { "debug-check-models", OptionEnum::DEBUG_CHECK_MODELS },
      { "decision-mode", OptionEnum::DECISION },
      { "decision", OptionEnum::DECISION },
      { "deep-restart", OptionEnum::DEEP_RESTART },
      { "deep-restart-factor", OptionEnum::DEEP_RESTART_FACTOR },
      { "default-function-value-mode", OptionEnum::DEFAULT_FUNCTION_VALUE_MODE },
      { "difficulty-mode", OptionEnum::DIFFICULTY_MODE },
      { "dio-decomps", OptionEnum::DIO_DECOMPS },
      { "dio-solver", OptionEnum::DIO_SOLVER },
      { "dio-turns", OptionEnum::DIO_TURNS },
      { "drat-binary-format", OptionEnum::DRAT_BINARY_FORMAT },
      { "dt-binary-split", OptionEnum::DT_BINARY_SPLIT },
      { "dt-blast-splits", OptionEnum::DT_BLAST_SPLITS },
      { "dt-cyclic", OptionEnum::DT_CYCLIC },
      { "dt-infer-as-lemmas", OptionEnum::DT_INFER_AS_LEMMAS },
      { "dt-nested-rec", OptionEnum::DT_NESTED_REC },
      { "dt-polite-optimize", OptionEnum::DT_POLITE_OPTIMIZE },
      { "dt-share-sel", OptionEnum::DT_SHARE_SEL },
      { "dt-stc-ind", OptionEnum::DT_STC_IND },
      { "dt-var-exp-quant", OptionEnum::DT_VAR_EXP_QUANT },
      { "dump-difficulty", OptionEnum::DUMP_DIFFICULTY },
      { "dump-instantiations", OptionEnum::DUMP_INSTANTIATIONS },
      { "dump-instantiations-debug", OptionEnum::DUMP_INSTANTIATIONS_DEBUG },
      { "dump-models", OptionEnum::DUMP_MODELS },
      { "dump-proofs", OptionEnum::DUMP_PROOFS },
      { "dump-unsat-cores", OptionEnum::DUMP_UNSAT_CORES },
      { "dump-unsat-cores-lemmas", OptionEnum::DUMP_UNSAT_CORES_LEMMAS },
      { "e-matching", OptionEnum::E_MATCHING },
      { "eager-arith-bv-conv", OptionEnum::EAGER_ARITH_BV_CONV },
      { "early-exit", OptionEnum::EARLY_EXIT },
      { "early-ite-removal", OptionEnum::EARLY_ITE_REMOVAL },
      { "ee-mode", OptionEnum::EE_MODE },
      { "elim-taut-quant", OptionEnum::ELIM_TAUT_QUANT },
      { "enum-inst", OptionEnum::ENUM_INST },
      { "enum-inst-interleave", OptionEnum::ENUM_INST_INTERLEAVE },
      { "enum-inst-limit", OptionEnum::ENUM_INST_LIMIT },
      { "enum-inst-rd", OptionEnum::ENUM_INST_RD },
      { "enum-inst-stratify", OptionEnum::ENUM_INST_STRATIFY },
      { "enum-inst-sum", OptionEnum::ENUM_INST_SUM },
      { "diagnostic-output-channel", OptionEnum::ERR },
      { "err", OptionEnum::ERR },
      { "error-selection-rule", OptionEnum::ERROR_SELECTION_RULE },
      { "expr-depth", OptionEnum::EXPR_DEPTH },
      { "ext-rew-prep", OptionEnum::EXT_REW_PREP },
      { "ext-rewrite-quant", OptionEnum::EXT_REWRITE_QUANT },
      { "fc-penalties", OptionEnum::FC_PENALTIES },
      { "ff", OptionEnum::FF },
      { "ff-bitsum", OptionEnum::FF_BITSUM },
      { "ff-elim-disjunctive-bit", OptionEnum::FF_ELIM_DISJUNCTIVE_BIT },
      { "ff-field-polys", OptionEnum::FF_FIELD_POLYS },
      { "ff-solver", OptionEnum::FF_SOLVER },
      { "ff-trace-gb", OptionEnum::FF_TRACE_GB },
      { "filename", OptionEnum::FILENAME },
      { "filesystem-access", OptionEnum::FILESYSTEM_ACCESS },
      { "finite-model-find", OptionEnum::FINITE_MODEL_FIND },
      { "flatten-ho-chains", OptionEnum::FLATTEN_HO_CHAINS },
      { "fmf-bound", OptionEnum::FMF_BOUND },
      { "fmf-bound-blast", OptionEnum::FMF_BOUND_BLAST },
      { "fmf-bound-lazy", OptionEnum::FMF_BOUND_LAZY },
      { "fmf-fun", OptionEnum::FMF_FUN },
      { "fmf-fun-rlv", OptionEnum::FMF_FUN_RLV },
      { "fmf-mbqi", OptionEnum::FMF_MBQI },
      { "fmf-type-completion-thresh", OptionEnum::FMF_TYPE_COMPLETION_THRESH },
      { "force-logic", OptionEnum::FORCE_LOGIC },
      { "force-no-limit-cpu-while-dump", OptionEnum::FORCE_NO_LIMIT_CPU_WHILE_DUMP },
      { "foreign-theory-rewrite", OptionEnum::FOREIGN_THEORY_REWRITE },
      { "fp", OptionEnum::FP },
      { "fp-exp", OptionEnum::FP_EXP },
      { "fp-lazy-wb", OptionEnum::FP_LAZY_WB },
      { "fresh-binders", OptionEnum::FRESH_BINDERS },
      { "fresh-declarations", OptionEnum::FRESH_DECLARATIONS },
      { "full-saturate-quant", OptionEnum::FULL_SATURATE_QUANT },
      { "full-sygus-verify", OptionEnum::FULL_SYGUS_VERIFY },
      { "global-declarations", OptionEnum::GLOBAL_DECLARATIONS },
      { "global-negate", OptionEnum::GLOBAL_NEGATE },
      { "help", OptionEnum::HELP },
      { "help-regular", OptionEnum::HELP_REGULAR },
      { "heuristic-pivots", OptionEnum::HEURISTIC_PIVOTS },
      { "ho-elim", OptionEnum::HO_ELIM },
      { "ho-elim-store-ax", OptionEnum::HO_ELIM_STORE_AX },
      { "ho-matching", OptionEnum::HO_MATCHING },
      { "ho-merge-term-db", OptionEnum::HO_MERGE_TERM_DB },
      { "iand-mode", OptionEnum::IAND_MODE },
      { "ieval", OptionEnum::IEVAL },
      { "in", OptionEnum::IN },
      { "increment-triggers", OptionEnum::INCREMENT_TRIGGERS },
      { "incremental", OptionEnum::INCREMENTAL },
      { "inst-max-level", OptionEnum::INST_MAX_LEVEL },
      { "inst-max-rounds", OptionEnum::INST_MAX_ROUNDS },
      { "inst-no-entail", OptionEnum::INST_NO_ENTAIL },
      { "inst-when", OptionEnum::INST_WHEN },
      { "inst-when-phase", OptionEnum::INST_WHEN_PHASE },
      { "int-wf-ind", OptionEnum::INT_WF_IND },
      { "interactive", OptionEnum::INTERACTIVE },
      { "interpolants-mode", OptionEnum::INTERPOLANTS_MODE },
      { "ite-dtt-split-quant", OptionEnum::ITE_DTT_SPLIT_QUANT },
      { "ite-lift-quant", OptionEnum::ITE_LIFT_QUANT },
      { "ite-simp", OptionEnum::ITE_SIMP },
      { "jh-rlv-order", OptionEnum::JH_RLV_ORDER },
      { "jh-skolem", OptionEnum::JH_SKOLEM },
      { "jh-skolem-rlv", OptionEnum::JH_SKOLEM_RLV },
      { "lang", OptionEnum::LANG },
      { "input-language", OptionEnum::LANG },
      { "learned-rewrite", OptionEnum::LEARNED_REWRITE },
      { "lemma-inprocess", OptionEnum::LEMMA_INPROCESS },
      { "lemma-inprocess-infer-eq-lit", OptionEnum::LEMMA_INPROCESS_INFER_EQ_LIT },
      { "lemma-inprocess-subs", OptionEnum::LEMMA_INPROCESS_SUBS },
      { "lemmas-on-replay-failure", OptionEnum::LEMMAS_ON_REPLAY_FAILURE },
      { "lfsc-expand-trust", OptionEnum::LFSC_EXPAND_TRUST },
      { "lfsc-flatten", OptionEnum::LFSC_FLATTEN },
      { "literal-matching", OptionEnum::LITERAL_MATCHING },
      { "macros-quant", OptionEnum::MACROS_QUANT },
      { "macros-quant-mode", OptionEnum::MACROS_QUANT_MODE },
      { "maxCutsInContext", OptionEnum::MAXCUTSINCONTEXT },
      { "mbqi", OptionEnum::MBQI },
      { "mbqi-fast-sygus", OptionEnum::MBQI_FAST_SYGUS },
      { "mbqi-interleave", OptionEnum::MBQI_INTERLEAVE },
      { "mbqi-one-inst-per-round", OptionEnum::MBQI_ONE_INST_PER_ROUND },
      { "minimal-unsat-cores", OptionEnum::MINIMAL_UNSAT_CORES },
      { "minisat-dump-dimacs", OptionEnum::MINISAT_DUMP_DIMACS },
      { "minisat-simplification", OptionEnum::MINISAT_SIMPLIFICATION },
      { "miniscope-quant", OptionEnum::MINISCOPE_QUANT },
      { "miniscope-quant-user", OptionEnum::MINISCOPE_QUANT_USER },
      { "miplib-trick", OptionEnum::MIPLIB_TRICK },
      { "miplib-trick-subs", OptionEnum::MIPLIB_TRICK_SUBS },
      { "model-based-arith-bv-conv", OptionEnum::MODEL_BASED_ARITH_BV_CONV },
      { "model-cores", OptionEnum::MODEL_CORES },
      { "model-u-print", OptionEnum::MODEL_U_PRINT },
      { "model-var-elim-uneval", OptionEnum::MODEL_VAR_ELIM_UNEVAL },
      { "multi-trigger-cache", OptionEnum::MULTI_TRIGGER_CACHE },
      { "multi-trigger-linear", OptionEnum::MULTI_TRIGGER_LINEAR },
      { "multi-trigger-priority", OptionEnum::MULTI_TRIGGER_PRIORITY },
      { "multi-trigger-when-single", OptionEnum::MULTI_TRIGGER_WHEN_SINGLE },
      { "new-prop", OptionEnum::NEW_PROP },
      { "nl-cov", OptionEnum::NL_COV },
      { "nl-cov-force", OptionEnum::NL_COV_FORCE },
      { "nl-cov-lift", OptionEnum::NL_COV_LIFT },
      { "nl-cov-linear-model", OptionEnum::NL_COV_LINEAR_MODEL },
      { "nl-cov-proj", OptionEnum::NL_COV_PROJ },
      { "nl-cov-prune", OptionEnum::NL_COV_PRUNE },
      { "nl-cov-var-elim", OptionEnum::NL_COV_VAR_ELIM },
      { "nl-ext", OptionEnum::NL_EXT },
      { "nl-ext-ent-conf", OptionEnum::NL_EXT_ENT_CONF },
      { "nl-ext-factor", OptionEnum::NL_EXT_FACTOR },
      { "nl-ext-inc-prec", OptionEnum::NL_EXT_INC_PREC },
      { "nl-ext-purify", OptionEnum::NL_EXT_PURIFY },
      { "nl-ext-rbound", OptionEnum::NL_EXT_RBOUND },
      { "nl-ext-rewrite", OptionEnum::NL_EXT_REWRITE },
      { "nl-ext-split-zero", OptionEnum::NL_EXT_SPLIT_ZERO },
      { "nl-ext-tf-taylor-deg", OptionEnum::NL_EXT_TF_TAYLOR_DEG },
      { "nl-ext-tf-tplanes", OptionEnum::NL_EXT_TF_TPLANES },
      { "nl-ext-tplanes", OptionEnum::NL_EXT_TPLANES },
      { "nl-ext-tplanes-interleave", OptionEnum::NL_EXT_TPLANES_INTERLEAVE },
      { "nl-icp", OptionEnum::NL_ICP },
      { "nl-rlv", OptionEnum::NL_RLV },
      { "nl-rlv-assert-bounds", OptionEnum::NL_RLV_ASSERT_BOUNDS },
      { "on-repeat-ite-simp", OptionEnum::ON_REPEAT_ITE_SIMP },
      { "opt-res-reconstruction-size", OptionEnum::OPT_RES_RECONSTRUCTION_SIZE },
      { "oracles", OptionEnum::ORACLES },
      { "regular-output-channel", OptionEnum::OUT },
      { "out", OptionEnum::OUT },
      { "output", OptionEnum::OUTPUT },
      { "output-lang", OptionEnum::OUTPUT_LANG },
      { "output-language", OptionEnum::OUTPUT_LANG },
      { "parse-only", OptionEnum::PARSE_ONLY },
      { "parse-skolem-definitions", OptionEnum::PARSE_SKOLEM_DEFINITIONS },
      { "parsing-mode", OptionEnum::PARSING_MODE },
      { "partial-triggers", OptionEnum::PARTIAL_TRIGGERS },
      { "check", OptionEnum::PARTITION_CHECK },
      { "partition-check", OptionEnum::PARTITION_CHECK },
      { "partition-conflict-size", OptionEnum::PARTITION_CONFLICT_SIZE },
      { "partition-start-time", OptionEnum::PARTITION_START_TIME },
      { "partition-strategy", OptionEnum::PARTITION_STRATEGY },
      { "partition", OptionEnum::PARTITION_STRATEGY },
      { "partition-time-interval", OptionEnum::PARTITION_TIME_INTERVAL },
      { "partition-tlimit", OptionEnum::PARTITION_TLIMIT },
      { "partition-when", OptionEnum::PARTITION_WHEN },
      { "pb-rewrites", OptionEnum::PB_REWRITES },
      { "pivot-threshold", OptionEnum::PIVOT_THRESHOLD },
      { "plugin-notify-sat-clause-in-solve", OptionEnum::PLUGIN_NOTIFY_SAT_CLAUSE_IN_SOLVE },
      { "plugin-share-skolems", OptionEnum::PLUGIN_SHARE_SKOLEMS },
      { "pool-inst", OptionEnum::POOL_INST },
      { "portfolio-jobs", OptionEnum::PORTFOLIO_JOBS },
      { "pp-assert-max-sub-size", OptionEnum::PP_ASSERT_MAX_SUB_SIZE },
      { "pre-skolem-quant", OptionEnum::PRE_SKOLEM_QUANT },
      { "pre-skolem-quant-nested", OptionEnum::PRE_SKOLEM_QUANT_NESTED },
      { "prenex-quant", OptionEnum::PRENEX_QUANT },
      { "prenex-quant-user", OptionEnum::PRENEX_QUANT_USER },
      { "preprocess-only", OptionEnum::PREPROCESS_ONLY },
      { "preregister-mode", OptionEnum::PREREGISTER_MODE },
      { "print-arith-lit-token", OptionEnum::PRINT_ARITH_LIT_TOKEN },
      { "print-cores-full", OptionEnum::PRINT_CORES_FULL },
      { "print-dot-clusters", OptionEnum::PRINT_DOT_CLUSTERS },
      { "print-inst", OptionEnum::PRINT_INST },
      { "print-inst-full", OptionEnum::PRINT_INST_FULL },
      { "print-skolem-definitions", OptionEnum::PRINT_SKOLEM_DEFINITIONS },
      { "print-success", OptionEnum::PRINT_SUCCESS },
      { "produce-abducts", OptionEnum::PRODUCE_ABDUCTS },
      { "interactive-mode", OptionEnum::PRODUCE_ASSERTIONS },
      { "produce-assertions", OptionEnum::PRODUCE_ASSERTIONS },
      { "produce-assignments", OptionEnum::PRODUCE_ASSIGNMENTS },
      { "produce-difficulty", OptionEnum::PRODUCE_DIFFICULTY },
      { "produce-interpolants", OptionEnum::PRODUCE_INTERPOLANTS },
      { "produce-learned-literals", OptionEnum::PRODUCE_LEARNED_LITERALS },
      { "produce-models", OptionEnum::PRODUCE_MODELS },
      { "produce-proofs", OptionEnum::PRODUCE_PROOFS },
      { "produce-unsat-assumptions", OptionEnum::PRODUCE_UNSAT_ASSUMPTIONS },
      { "produce-unsat-cores", OptionEnum::PRODUCE_UNSAT_CORES },
      { "proof-alethe-define-skolems", OptionEnum::PROOF_ALETHE_DEFINE_SKOLEMS },
      { "proof-alethe-res-pivots", OptionEnum::PROOF_ALETHE_RES_PIVOTS },
      { "proof-allow-trust", OptionEnum::PROOF_ALLOW_TRUST },
      { "proof-check", OptionEnum::PROOF_CHECK },
      { "proof-dag-global", OptionEnum::PROOF_DAG_GLOBAL },
      { "proof-dot-dag", OptionEnum::PROOF_DOT_DAG },
      { "proof-elim-subtypes", OptionEnum::PROOF_ELIM_SUBTYPES },
      { "proof-format-mode", OptionEnum::PROOF_FORMAT_MODE },
      { "proof-granularity", OptionEnum::PROOF_GRANULARITY },
      { "proof-mode", OptionEnum::PROOF_MODE },
      { "proof-pedantic", OptionEnum::PROOF_PEDANTIC },
      { "proof-pp-merge", OptionEnum::PROOF_PP_MERGE },
      { "proof-print-conclusion", OptionEnum::PROOF_PRINT_CONCLUSION },
      { "proof-print-reference", OptionEnum::PROOF_PRINT_REFERENCE },
      { "proof-prune-input", OptionEnum::PROOF_PRUNE_INPUT },
      { "proof-rewrite-rcons-rec-limit", OptionEnum::PROOF_REWRITE_RCONS_REC_LIMIT },
      { "proof-rewrite-rcons-step-limit", OptionEnum::PROOF_REWRITE_RCONS_STEP_LIMIT },
      { "prop-proof-mode", OptionEnum::PROP_PROOF_MODE },
      { "prop-row-length", OptionEnum::PROP_ROW_LENGTH },
      { "purify-triggers", OptionEnum::PURIFY_TRIGGERS },
      { "quant-alpha-equiv", OptionEnum::QUANT_ALPHA_EQUIV },
      { "quant-dsplit", OptionEnum::QUANT_DSPLIT },
      { "quant-fun-wd", OptionEnum::QUANT_FUN_WD },
      { "quant-ind", OptionEnum::QUANT_IND },
      { "quant-rep-mode", OptionEnum::QUANT_REP_MODE },
      { "quiet", OptionEnum::QUIET },
      { "random-freq", OptionEnum::RANDOM_FREQ },
      { "random-frequency", OptionEnum::RANDOM_FREQ },
      { "random-partitioning", OptionEnum::RANDOM_PARTITIONING },
      { "re-elim", OptionEnum::RE_ELIM },
      { "re-inter-mode", OptionEnum::RE_INTER_MODE },
      { "register-quant-body-terms", OptionEnum::REGISTER_QUANT_BODY_TERMS },
      { "relational-triggers", OptionEnum::RELATIONAL_TRIGGERS },
      { "relevance-filter", OptionEnum::RELEVANCE_FILTER },
      { "relevant-triggers", OptionEnum::RELEVANT_TRIGGERS },
      { "rels-exp", OptionEnum::RELS_EXP },
      { "repeat-simp", OptionEnum::REPEAT_SIMP },
      { "replay-early-close-depth", OptionEnum::REPLAY_EARLY_CLOSE_DEPTH },
      { "replay-lemma-reject-cut", OptionEnum::REPLAY_LEMMA_REJECT_CUT },
      { "replay-num-err-penalty", OptionEnum::REPLAY_NUM_ERR_PENALTY },
      { "replay-reject-cut", OptionEnum::REPLAY_REJECT_CUT },
      { "restart-int-base", OptionEnum::RESTART_INT_BASE },
      { "restart-int-inc", OptionEnum::RESTART_INT_INC },
      { "restrict-pivots", OptionEnum::RESTRICT_PIVOTS },
      { "revert-arith-models-on-unsat", OptionEnum::REVERT_ARITH_MODELS_ON_UNSAT },
      { "rlimit", OptionEnum::RLIMIT },
      { "rlimit-per", OptionEnum::RLIMIT_PER },
      { "reproducible-resource-limit", OptionEnum::RLIMIT_PER },
      { "rr-turns", OptionEnum::RR_TURNS },
      { "rweight", OptionEnum::RWEIGHT },
      { "safe-options", OptionEnum::SAFE_OPTIONS },
      { "sat-proof-min-dimacs", OptionEnum::SAT_PROOF_MIN_DIMACS },
      { "sat-random-seed", OptionEnum::SAT_RANDOM_SEED },
      { "sat-solver", OptionEnum::SAT_SOLVER },
      { "se-solve-int", OptionEnum::SE_SOLVE_INT },
      { "seed", OptionEnum::SEED },
      { "random-seed", OptionEnum::SEED },
      { "segv-spin", OptionEnum::SEGV_SPIN },
      { "semantic-checks", OptionEnum::SEMANTIC_CHECKS },
      { "sep", OptionEnum::SEP },
      { "sep-min-refine", OptionEnum::SEP_MIN_REFINE },
      { "sep-pre-skolem-emp", OptionEnum::SEP_PRE_SKOLEM_EMP },
      { "seq-array", OptionEnum::SEQ_ARRAY },
      { "sets-card-exp", OptionEnum::SETS_CARD_EXP },
      { "sets-exp", OptionEnum::SETS_EXP },
      { "sets-proxy-lemmas", OptionEnum::SETS_PROXY_LEMMAS },
      { "show-config", OptionEnum::SHOW_CONFIG },
      { "show-trace-tags", OptionEnum::SHOW_TRACE_TAGS },
      { "simp-ite-compress", OptionEnum::SIMP_ITE_COMPRESS },
      { "simp-with-care", OptionEnum::SIMP_WITH_CARE },
      { "simplex-check-period", OptionEnum::SIMPLEX_CHECK_PERIOD },
      { "simplification", OptionEnum::SIMPLIFICATION },
      { "simplification-mode", OptionEnum::SIMPLIFICATION },
      { "simplification-bcp", OptionEnum::SIMPLIFICATION_BCP },
      { "soi-qe", OptionEnum::SOI_QE },
      { "solve-bv-as-int", OptionEnum::SOLVE_BV_AS_INT },
      { "solve-int-as-bv", OptionEnum::SOLVE_INT_AS_BV },
      { "solve-real-as-int", OptionEnum::SOLVE_REAL_AS_INT },
      { "sort-inference", OptionEnum::SORT_INFERENCE },
      { "standard-effort-variable-order-pivots", OptionEnum::STANDARD_EFFORT_VARIABLE_ORDER_PIVOTS },
      { "static-learning", OptionEnum::STATIC_LEARNING },
      { "stats", OptionEnum::STATS },
      { "stats-all", OptionEnum::STATS_ALL },
      { "stats-every-query", OptionEnum::STATS_EVERY_QUERY },
      { "stats-internal", OptionEnum::STATS_INTERNAL },
      { "strict-parsing", OptionEnum::STRICT_PARSING },
      { "strings-alpha-card", OptionEnum::STRINGS_ALPHA_CARD },
      { "strings-check-entail-len", OptionEnum::STRINGS_CHECK_ENTAIL_LEN },
      { "strings-deq-ext", OptionEnum::STRINGS_DEQ_EXT },
      { "strings-eager-eval", OptionEnum::STRINGS_EAGER_EVAL },
      { "strings-eager-len-re", OptionEnum::STRINGS_EAGER_LEN_RE },
      { "strings-eager-reg", OptionEnum::STRINGS_EAGER_REG },
      { "strings-eager-solver", OptionEnum::STRINGS_EAGER_SOLVER },
      { "strings-exp", OptionEnum::STRINGS_EXP },
      { "strings-ff", OptionEnum::STRINGS_FF },
      { "strings-fmf", OptionEnum::STRINGS_FMF },
      { "strings-infer-as-lemmas", OptionEnum::STRINGS_INFER_AS_LEMMAS },
      { "strings-infer-sym", OptionEnum::STRINGS_INFER_SYM },
      { "strings-lazy-pp", OptionEnum::STRINGS_LAZY_PP },
      { "strings-len-norm", OptionEnum::STRINGS_LEN_NORM },
      { "strings-mbr", OptionEnum::STRINGS_MBR },
      { "strings-model-max-len", OptionEnum::STRINGS_MODEL_MAX_LEN },
      { "strings-process-loop-mode", OptionEnum::STRINGS_PROCESS_LOOP_MODE },
      { "strings-re-derive-conf", OptionEnum::STRINGS_RE_DERIVE_CONF },
      { "strings-re-posc-eager", OptionEnum::STRINGS_RE_POSC_EAGER },
      { "strings-rec-arith-approx", OptionEnum::STRINGS_REC_ARITH_APPROX },
      { "strings-regexp-inclusion", OptionEnum::STRINGS_REGEXP_INCLUSION },
      { "strings-rexplain-lemmas", OptionEnum::STRINGS_REXPLAIN_LEMMAS },
      { "sub-cbqi", OptionEnum::SUB_CBQI },
      { "sub-cbqi-timeout", OptionEnum::SUB_CBQI_TIMEOUT },
      { "sygus", OptionEnum::SYGUS },
      { "sygus-abort-size", OptionEnum::SYGUS_ABORT_SIZE },
      { "sygus-add-const-grammar", OptionEnum::SYGUS_ADD_CONST_GRAMMAR },
      { "sygus-arg-relevant", OptionEnum::SYGUS_ARG_RELEVANT },
      { "sygus-auto-unfold", OptionEnum::SYGUS_AUTO_UNFOLD },
      { "sygus-bool-ite-return-const", OptionEnum::SYGUS_BOOL_ITE_RETURN_CONST },
      { "sygus-core-connective", OptionEnum::SYGUS_CORE_CONNECTIVE },
      { "sygus-crepair-abort", OptionEnum::SYGUS_CREPAIR_ABORT },
      { "sygus-enum", OptionEnum::SYGUS_ENUM },
      { "sygus-enum-fast-num-consts", OptionEnum::SYGUS_ENUM_FAST_NUM_CONSTS },
      { "sygus-enum-random-p", OptionEnum::SYGUS_ENUM_RANDOM_P },
      { "sygus-eval-unfold", OptionEnum::SYGUS_EVAL_UNFOLD },
      { "sygus-expr-miner-check-timeout", OptionEnum::SYGUS_EXPR_MINER_CHECK_TIMEOUT },
      { "sygus-fair", OptionEnum::SYGUS_FAIR },
      { "sygus-fair-max", OptionEnum::SYGUS_FAIR_MAX },
      { "sygus-filter-sol", OptionEnum::SYGUS_FILTER_SOL },
      { "sygus-filter-sol-rev", OptionEnum::SYGUS_FILTER_SOL_REV },
      { "sygus-grammar-cons", OptionEnum::SYGUS_GRAMMAR_CONS },
      { "sygus-grammar-norm", OptionEnum::SYGUS_GRAMMAR_NORM },
      { "sygus-grammar-use-disj", OptionEnum::SYGUS_GRAMMAR_USE_DISJ },
      { "sygus-inference", OptionEnum::SYGUS_INFERENCE },
      { "sygus-inst", OptionEnum::SYGUS_INST },
      { "sygus-inst-mode", OptionEnum::SYGUS_INST_MODE },
      { "sygus-inst-scope", OptionEnum::SYGUS_INST_SCOPE },
      { "sygus-inst-term-sel", OptionEnum::SYGUS_INST_TERM_SEL },
      { "sygus-inv-templ", OptionEnum::SYGUS_INV_TEMPL },
      { "sygus-inv-templ-when-sg", OptionEnum::SYGUS_INV_TEMPL_WHEN_SG },
      { "sygus-min-grammar", OptionEnum::SYGUS_MIN_GRAMMAR },
      { "sygus-out", OptionEnum::SYGUS_OUT },
      { "sygus-pbe", OptionEnum::SYGUS_PBE },
      { "sygus-pbe-multi-fair", OptionEnum::SYGUS_PBE_MULTI_FAIR },
      { "sygus-pbe-multi-fair-diff", OptionEnum::SYGUS_PBE_MULTI_FAIR_DIFF },
      { "sygus-qe-preproc", OptionEnum::SYGUS_QE_PREPROC },
      { "sygus-query-gen", OptionEnum::SYGUS_QUERY_GEN },
      { "sygus-query-gen-dump-files", OptionEnum::SYGUS_QUERY_GEN_DUMP_FILES },
      { "sygus-query-gen-filter-solved", OptionEnum::SYGUS_QUERY_GEN_FILTER_SOLVED },
      { "sygus-query-gen-thresh", OptionEnum::SYGUS_QUERY_GEN_THRESH },
      { "sygus-rec-fun", OptionEnum::SYGUS_REC_FUN },
      { "sygus-rec-fun-eval-limit", OptionEnum::SYGUS_REC_FUN_EVAL_LIMIT },
      { "sygus-repair-const", OptionEnum::SYGUS_REPAIR_CONST },
      { "sygus-repair-const-timeout", OptionEnum::SYGUS_REPAIR_CONST_TIMEOUT },
      { "sygus-rewriter", OptionEnum::SYGUS_REWRITER },
      { "sygus-rr-synth-accel", OptionEnum::SYGUS_RR_SYNTH_ACCEL },
      { "sygus-rr-synth-check", OptionEnum::SYGUS_RR_SYNTH_CHECK },
      { "sygus-rr-synth-filter-cong", OptionEnum::SYGUS_RR_SYNTH_FILTER_CONG },
      { "sygus-rr-synth-filter-match", OptionEnum::SYGUS_RR_SYNTH_FILTER_MATCH },
      { "sygus-rr-synth-filter-nl", OptionEnum::SYGUS_RR_SYNTH_FILTER_NL },
      { "sygus-rr-synth-filter-order", OptionEnum::SYGUS_RR_SYNTH_FILTER_ORDER },
      { "sygus-rr-synth-input-nvars", OptionEnum::SYGUS_RR_SYNTH_INPUT_NVARS },
      { "sygus-rr-synth-rec", OptionEnum::SYGUS_RR_SYNTH_REC },
      { "sygus-sample-fp-uniform", OptionEnum::SYGUS_SAMPLE_FP_UNIFORM },
      { "sygus-sample-grammar", OptionEnum::SYGUS_SAMPLE_GRAMMAR },
      { "sygus-samples", OptionEnum::SYGUS_SAMPLES },
      { "sygus-si", OptionEnum::SYGUS_SI },
      { "sygus-si-abort", OptionEnum::SYGUS_SI_ABORT },
      { "sygus-si-rcons", OptionEnum::SYGUS_SI_RCONS },
      { "sygus-si-rcons-limit", OptionEnum::SYGUS_SI_RCONS_LIMIT },
      { "sygus-si-rcons-p", OptionEnum::SYGUS_SI_RCONS_P },
      { "sygus-simple-sym-break", OptionEnum::SYGUS_SIMPLE_SYM_BREAK },
      { "sygus-stream", OptionEnum::SYGUS_STREAM },
      { "sygus-sym-break-lazy", OptionEnum::SYGUS_SYM_BREAK_LAZY },
      { "sygus-sym-break-pbe", OptionEnum::SYGUS_SYM_BREAK_PBE },
      { "sygus-sym-break-rlv", OptionEnum::SYGUS_SYM_BREAK_RLV },
      { "sygus-unif-cond-independent-no-repeat-sol", OptionEnum::SYGUS_UNIF_COND_INDEPENDENT_NO_REPEAT_SOL },
      { "sygus-unif-pi", OptionEnum::SYGUS_UNIF_PI },
      { "sygus-unif-shuffle-cond", OptionEnum::SYGUS_UNIF_SHUFFLE_COND },
      { "sygus-verify-inst-max-rounds", OptionEnum::SYGUS_VERIFY_INST_MAX_ROUNDS },
      { "sygus-verify-timeout", OptionEnum::SYGUS_VERIFY_TIMEOUT },
      { "symmetry-breaker", OptionEnum::SYMMETRY_BREAKER },
      { "tc-mode", OptionEnum::TC_MODE },
      { "term-db-mode", OptionEnum::TERM_DB_MODE },
      { "term-sort-overload", OptionEnum::TERM_SORT_OVERLOAD },
      { "theoryof-mode", OptionEnum::THEORYOF_MODE },
      { "timeout-core-timeout", OptionEnum::TIMEOUT_CORE_TIMEOUT },
      { "tlimit", OptionEnum::TLIMIT },
      { "tlimit-per", OptionEnum::TLIMIT_PER },
      { "trace", OptionEnum::TRACE },
      { "trigger-active-sel", OptionEnum::TRIGGER_ACTIVE_SEL },
      { "trigger-sel", OptionEnum::TRIGGER_SEL },
      { "type-checking", OptionEnum::TYPE_CHECKING },
      { "uf-card-exp", OptionEnum::UF_CARD_EXP },
      { "uf-ho-exp", OptionEnum::UF_HO_EXP },
      { "uf-ho-ext", OptionEnum::UF_HO_EXT },
      { "uf-lambda-qe", OptionEnum::UF_LAMBDA_QE },
      { "uf-lazy-ll", OptionEnum::UF_LAZY_LL },
      { "uf-ss", OptionEnum::UF_SS },
      { "uf-ss-abort-card", OptionEnum::UF_SS_ABORT_CARD },
      { "uf-ss-fair", OptionEnum::UF_SS_FAIR },
      { "uf-ss-fair-monotone", OptionEnum::UF_SS_FAIR_MONOTONE },
      { "unate-lemmas", OptionEnum::UNATE_LEMMAS },
      { "unconstrained-simp", OptionEnum::UNCONSTRAINED_SIMP },
      { "unsat-cores-mode", OptionEnum::UNSAT_CORES_MODE },
      { "use-approx", OptionEnum::USE_APPROX },
      { "use-fcsimplex", OptionEnum::USE_FCSIMPLEX },
      { "use-portfolio", OptionEnum::USE_PORTFOLIO },
      { "use-soi", OptionEnum::USE_SOI },
      { "user-pat", OptionEnum::USER_PAT },
      { "user-pool", OptionEnum::USER_POOL },
      { "var-elim-quant", OptionEnum::VAR_ELIM_QUANT },
      { "var-ineq-elim-quant", OptionEnum::VAR_INEQ_ELIM_QUANT },
      { "verbose", OptionEnum::VERBOSE },
      { "verbosity", OptionEnum::VERBOSITY },
      { "version", OptionEnum::VERSION },
      { "wf-checking", OptionEnum::WF_CHECKING },
      { "partitions-out", OptionEnum::WRITE_PARTITIONS_TO },
      { "write-partitions-to", OptionEnum::WRITE_PARTITIONS_TO },
    };
  // clang-format on
}

namespace cvc5::internal::options
{
  // Contains the default option handlers (i.e. parsers)
  namespace handlers {

  /**
   * Utility function for handling numeric options. Takes care of checking for
   * unsignedness, parsing and handling parsing exceptions. Expects `conv` to be
   * a conversion function like `std::stod`, accepting a `std::string` and a
   * `size_t*`. The argument `type` is only used to generate proper error
   * messages and should be the string representation of `T`. If `T` is
   * unsigned, checks that `optionarg` contains no minus. Then `conv` is called
   * and the error conditions are handled: `conv` may throw an exception or
   * `pos` may be smaller than the size of `optionarg`, indicating that not the
   * entirety of `optionarg` was parsed.
   */
  template <typename T, typename FF>
  T parseNumber(const std::string& flag,
                const std::string& optionarg,
                FF&& conv,
                const std::string& type)
  {
    if (!std::numeric_limits<T>::is_signed
        && (optionarg.find('-') != std::string::npos))
    {
      std::stringstream ss;
      ss << "Argument '" << optionarg << "' for " << type << " option " << flag
         << " is negative";
      throw OptionException(ss.str());
    }
    size_t pos = 0;
    T res;
    try
    {
      res = conv(optionarg, &pos);
    }
    catch (const std::exception& e)
    {
      std::stringstream ss;
      ss << "Argument '" << optionarg << "' for " << type << " option " << flag
         << " did not parse as " << type;
      throw OptionException(ss.str());
    }
    if (pos < optionarg.size())
    {
      std::stringstream ss;
      ss << "Argument '" << optionarg << "' for " << type << " option " << flag
         << " did parse only partially as " << type << ", leaving '"
         << optionarg.substr(pos) << "'";
      throw OptionException(ss.str());
    }
    return res;
  }

  /** Default handler that triggers a compiler error */
  template <typename T>
  T handleOption(const std::string& flag, const std::string& optionarg)
  {
    T::unsupported_handleOption_specialization;
    return *static_cast<T*>(nullptr);
  }

  /** Handle a string option by returning it as is. */
  template <>
  std::string handleOption<std::string>(const std::string& flag,
                                        const std::string& optionarg)
  {
    return optionarg;
  }
  /** Handle a bool option, recognizing "true" or "false". */
  template <>
  bool handleOption<bool>(const std::string& flag, const std::string& optionarg)
  {
    if (optionarg == "true")
    {
      return true;
    }
    if (optionarg == "false")
    {
      return false;
    }
    throw OptionException("Argument '" + optionarg + "' for bool option " + flag
                          + " is not a bool constant");
  }

  /** Handle a double option, using `parseNumber` with `std::stod`. */
  template <>
  double handleOption<double>(const std::string& flag,
                              const std::string& optionarg)
  {
    return parseNumber<double>(
        flag,
        optionarg,
        [](const auto& s, auto p) { return std::stod(s, p); },
        "double");
  }

  /** Handle a int64_t option, using `parseNumber` with `std::stoll`. */
  template <>
  int64_t handleOption<int64_t>(const std::string& flag,
                                const std::string& optionarg)
  {
    return parseNumber<int64_t>(
        flag,
        optionarg,
        [](const auto& s, auto p) { return std::stoll(s, p); },
        "int64_t");
  }

  /** Handle a uint64_t option, using `parseNumber` with `std::stoull`. */
  template <>
  uint64_t handleOption<uint64_t>(const std::string& flag,
                                  const std::string& optionarg)
  {
    return parseNumber<uint64_t>(
        flag,
        optionarg,
        [](const auto& s, auto p) { return std::stoull(s, p); },
        "uint64_t");
  }

  /** Handle a ManagedIn option. */
  template <>
  ManagedIn handleOption<ManagedIn>(const std::string& flag,
                                    const std::string& optionarg)
  {
    ManagedIn res;
    res.open(optionarg);
    return res;
  }

  /** Handle a ManagedErr option. */
  template <>
  ManagedErr handleOption<ManagedErr>(const std::string& flag,
                                      const std::string& optionarg)
  {
    ManagedErr res;
    res.open(optionarg);
    return res;
  }

  /** Handle a ManagedOut option. */
  template <>
  ManagedOut handleOption<ManagedOut>(const std::string& flag,
                                      const std::string& optionarg)
  {
    ManagedOut res;
    res.open(optionarg);
    return res;
  }
  }

  std::vector<std::string> getNames()
  {
    return {
        // clang-format off
    "abstract-values", "ackermann", "append-learned-literals-to-cubes",
    "approx-branch-depth", "arith-brab", "arith-eq-solver", "arith-exp",
    "arith-prop", "arith-prop-clauses", "arith-rewrite-equalities",
    "arith-static-learning", "arrays-eager-index", "arrays-eager-lemmas",
    "arrays-exp", "arrays-optimize-linear", "arrays-prop",
    "arrays-reduce-sharing", "arrays-weak-equiv", "assign-function-values",
    "bags", "bitblast", "bitwise-eq", "bool-to-bv", "bv-assert-input",
    "bv-eager-eval", "bv-eq-engine", "bv-gauss-elim", "bv-intro-pow2",
    "bv-print-consts-as-indexed-symbols", "bv-propagate", "bv-rw-extend-eq",
    "bv-sat-solver", "bv-solver", "bv-to-bool", "bv-to-int-use-pow2",
    "bvand-integer-granularity", "cbqi", "cbqi-all-conflict", "cbqi-mode",
    "cbqi-skip-rd", "cbqi-tconstraint", "cbqi-vo-exp", "cdt-bisimilar",
    "cegis-sample", "cegqi", "cegqi-all", "cegqi-bv", "cegqi-bv-concat-inv",
    "cegqi-bv-ineq", "cegqi-bv-interleave-value", "cegqi-bv-linear",
    "cegqi-bv-rm-extract", "cegqi-bv-solve-nl", "cegqi-full", "cegqi-inf-int",
    "cegqi-inf-real", "cegqi-innermost", "cegqi-midpoint", "cegqi-min-bounds",
    "cegqi-multi-inst", "cegqi-nested-qe", "cegqi-nopt", "cegqi-round-up-lia",
    "check", "check-abducts", "check-interpolants", "check-models",
    "check-proof-steps", "check-proofs", "check-synth-sol", "check-unsat-cores",
    "checks-before-partition", "checks-between-partitions",
    "collect-pivot-stats", "compute-partitions", "cond-var-split-quant",
    "condense-function-values", "conflict-process", "conjecture-gen",
    "conjecture-gen-gt-enum", "conjecture-gen-max-depth",
    "conjecture-gen-per-round", "cons-exp-triggers", "copyright",
    "cut-all-bounded", "dag-thresh", "datatypes-exp", "debug-check-models",
    "decision", "decision-mode", "deep-restart", "deep-restart-factor",
    "default-function-value-mode", "diagnostic-output-channel",
    "difficulty-mode", "dio-decomps", "dio-solver", "dio-turns",
    "drat-binary-format", "dt-binary-split", "dt-blast-splits", "dt-cyclic",
    "dt-infer-as-lemmas", "dt-nested-rec", "dt-polite-optimize", "dt-share-sel",
    "dt-stc-ind", "dt-var-exp-quant", "dump-difficulty", "dump-instantiations",
    "dump-instantiations-debug", "dump-models", "dump-proofs",
    "dump-unsat-cores", "dump-unsat-cores-lemmas", "e-matching",
    "eager-arith-bv-conv", "early-exit", "early-ite-removal", "ee-mode",
    "elim-taut-quant", "enum-inst", "enum-inst-interleave", "enum-inst-limit",
    "enum-inst-rd", "enum-inst-stratify", "enum-inst-sum", "err",
    "error-selection-rule", "expr-depth", "ext-rew-prep", "ext-rewrite-quant",
    "fc-penalties", "ff", "ff-bitsum", "ff-elim-disjunctive-bit",
    "ff-field-polys", "ff-solver", "ff-trace-gb", "filename",
    "filesystem-access", "finite-model-find", "flatten-ho-chains", "fmf-bound",
    "fmf-bound-blast", "fmf-bound-lazy", "fmf-fun", "fmf-fun-rlv", "fmf-mbqi",
    "fmf-type-completion-thresh", "force-logic",
    "force-no-limit-cpu-while-dump", "foreign-theory-rewrite", "fp", "fp-exp",
    "fp-lazy-wb", "fresh-binders", "fresh-declarations", "full-saturate-quant",
    "full-sygus-verify", "global-declarations", "global-negate", "help",
    "help-regular", "heuristic-pivots", "ho-elim", "ho-elim-store-ax",
    "ho-matching", "ho-merge-term-db", "iand-mode", "ieval", "in",
    "increment-triggers", "incremental", "input-language", "inst-max-level",
    "inst-max-rounds", "inst-no-entail", "inst-when", "inst-when-phase",
    "int-wf-ind", "interactive", "interactive-mode", "interpolants-mode",
    "ite-dtt-split-quant", "ite-lift-quant", "ite-simp", "jh-rlv-order",
    "jh-skolem", "jh-skolem-rlv", "lang", "learned-rewrite", "lemma-inprocess",
    "lemma-inprocess-infer-eq-lit", "lemma-inprocess-subs",
    "lemmas-on-replay-failure", "lfsc-expand-trust", "lfsc-flatten",
    "literal-matching", "macros-quant", "macros-quant-mode", "maxCutsInContext",
    "mbqi", "mbqi-fast-sygus", "mbqi-interleave", "mbqi-one-inst-per-round",
    "minimal-unsat-cores", "minisat-dump-dimacs", "minisat-simplification",
    "miniscope-quant", "miniscope-quant-user", "miplib-trick",
    "miplib-trick-subs", "model-based-arith-bv-conv", "model-cores",
    "model-u-print", "model-var-elim-uneval", "multi-trigger-cache",
    "multi-trigger-linear", "multi-trigger-priority",
    "multi-trigger-when-single", "new-prop", "nl-cov", "nl-cov-force",
    "nl-cov-lift", "nl-cov-linear-model", "nl-cov-proj", "nl-cov-prune",
    "nl-cov-var-elim", "nl-ext", "nl-ext-ent-conf", "nl-ext-factor",
    "nl-ext-inc-prec", "nl-ext-purify", "nl-ext-rbound", "nl-ext-rewrite",
    "nl-ext-split-zero", "nl-ext-tf-taylor-deg", "nl-ext-tf-tplanes",
    "nl-ext-tplanes", "nl-ext-tplanes-interleave", "nl-icp", "nl-rlv",
    "nl-rlv-assert-bounds", "on-repeat-ite-simp", "opt-res-reconstruction-size",
    "oracles", "out", "output", "output-lang", "output-language", "parse-only",
    "parse-skolem-definitions", "parsing-mode", "partial-triggers", "partition",
    "partition-check", "partition-conflict-size", "partition-start-time",
    "partition-strategy", "partition-time-interval", "partition-tlimit",
    "partition-when", "partitions-out", "pb-rewrites", "pivot-threshold",
    "plugin-notify-sat-clause-in-solve", "plugin-share-skolems", "pool-inst",
    "portfolio-jobs", "pp-assert-max-sub-size", "pre-skolem-quant",
    "pre-skolem-quant-nested", "prenex-quant", "prenex-quant-user",
    "preprocess-only", "preregister-mode", "print-arith-lit-token",
    "print-cores-full", "print-dot-clusters", "print-inst", "print-inst-full",
    "print-skolem-definitions", "print-success", "produce-abducts",
    "produce-assertions", "produce-assignments", "produce-difficulty",
    "produce-interpolants", "produce-learned-literals", "produce-models",
    "produce-proofs", "produce-unsat-assumptions", "produce-unsat-cores",
    "proof-alethe-define-skolems", "proof-alethe-res-pivots",
    "proof-allow-trust", "proof-check", "proof-dag-global", "proof-dot-dag",
    "proof-elim-subtypes", "proof-format-mode", "proof-granularity",
    "proof-mode", "proof-pedantic", "proof-pp-merge", "proof-print-conclusion",
    "proof-print-reference", "proof-prune-input",
    "proof-rewrite-rcons-rec-limit", "proof-rewrite-rcons-step-limit",
    "prop-proof-mode", "prop-row-length", "purify-triggers",
    "quant-alpha-equiv", "quant-dsplit", "quant-fun-wd", "quant-ind",
    "quant-rep-mode", "quiet", "random-freq", "random-frequency",
    "random-partitioning", "random-seed", "re-elim", "re-inter-mode",
    "register-quant-body-terms", "regular-output-channel",
    "relational-triggers", "relevance-filter", "relevant-triggers", "rels-exp",
    "repeat-simp", "replay-early-close-depth", "replay-lemma-reject-cut",
    "replay-num-err-penalty", "replay-reject-cut",
    "reproducible-resource-limit", "restart-int-base", "restart-int-inc",
    "restrict-pivots", "revert-arith-models-on-unsat", "rlimit", "rlimit-per",
    "rr-turns", "rweight", "safe-options", "sat-proof-min-dimacs",
    "sat-random-seed", "sat-solver", "se-solve-int", "seed", "segv-spin",
    "semantic-checks", "sep", "sep-min-refine", "sep-pre-skolem-emp",
    "seq-array", "sets-card-exp", "sets-exp", "sets-proxy-lemmas",
    "show-config", "show-trace-tags", "simp-ite-compress", "simp-with-care",
    "simplex-check-period", "simplification", "simplification-bcp",
    "simplification-mode", "soi-qe", "solve-bv-as-int", "solve-int-as-bv",
    "solve-real-as-int", "sort-inference",
    "standard-effort-variable-order-pivots", "static-learning", "stats",
    "stats-all", "stats-every-query", "stats-internal", "strict-parsing",
    "strings-alpha-card", "strings-check-entail-len", "strings-deq-ext",
    "strings-eager-eval", "strings-eager-len-re", "strings-eager-reg",
    "strings-eager-solver", "strings-exp", "strings-ff", "strings-fmf",
    "strings-infer-as-lemmas", "strings-infer-sym", "strings-lazy-pp",
    "strings-len-norm", "strings-mbr", "strings-model-max-len",
    "strings-process-loop-mode", "strings-re-derive-conf",
    "strings-re-posc-eager", "strings-rec-arith-approx",
    "strings-regexp-inclusion", "strings-rexplain-lemmas", "sub-cbqi",
    "sub-cbqi-timeout", "sygus", "sygus-abort-size", "sygus-add-const-grammar",
    "sygus-arg-relevant", "sygus-auto-unfold", "sygus-bool-ite-return-const",
    "sygus-core-connective", "sygus-crepair-abort", "sygus-enum",
    "sygus-enum-fast-num-consts", "sygus-enum-random-p", "sygus-eval-unfold",
    "sygus-expr-miner-check-timeout", "sygus-fair", "sygus-fair-max",
    "sygus-filter-sol", "sygus-filter-sol-rev", "sygus-grammar-cons",
    "sygus-grammar-norm", "sygus-grammar-use-disj", "sygus-inference",
    "sygus-inst", "sygus-inst-mode", "sygus-inst-scope", "sygus-inst-term-sel",
    "sygus-inv-templ", "sygus-inv-templ-when-sg", "sygus-min-grammar",
    "sygus-out", "sygus-pbe", "sygus-pbe-multi-fair",
    "sygus-pbe-multi-fair-diff", "sygus-qe-preproc", "sygus-query-gen",
    "sygus-query-gen-dump-files", "sygus-query-gen-filter-solved",
    "sygus-query-gen-thresh", "sygus-rec-fun", "sygus-rec-fun-eval-limit",
    "sygus-repair-const", "sygus-repair-const-timeout", "sygus-rewriter",
    "sygus-rr-synth-accel", "sygus-rr-synth-check",
    "sygus-rr-synth-filter-cong", "sygus-rr-synth-filter-match",
    "sygus-rr-synth-filter-nl", "sygus-rr-synth-filter-order",
    "sygus-rr-synth-input-nvars", "sygus-rr-synth-rec",
    "sygus-sample-fp-uniform", "sygus-sample-grammar", "sygus-samples",
    "sygus-si", "sygus-si-abort", "sygus-si-rcons", "sygus-si-rcons-limit",
    "sygus-si-rcons-p", "sygus-simple-sym-break", "sygus-stream",
    "sygus-sym-break-lazy", "sygus-sym-break-pbe", "sygus-sym-break-rlv",
    "sygus-unif-cond-independent-no-repeat-sol", "sygus-unif-pi",
    "sygus-unif-shuffle-cond", "sygus-verify-inst-max-rounds",
    "sygus-verify-timeout", "symmetry-breaker", "tc-mode", "term-db-mode",
    "term-sort-overload", "theoryof-mode", "timeout-core-timeout", "tlimit",
    "tlimit-per", "trace", "trigger-active-sel", "trigger-sel", "type-checking",
    "uf-card-exp", "uf-ho-exp", "uf-ho-ext", "uf-lambda-qe", "uf-lazy-ll",
    "uf-ss", "uf-ss-abort-card", "uf-ss-fair", "uf-ss-fair-monotone",
    "unate-lemmas", "unconstrained-simp", "unsat-cores-mode", "use-approx",
    "use-fcsimplex", "use-portfolio", "use-soi", "user-pat", "user-pool",
    "var-elim-quant", "var-ineq-elim-quant", "verbose", "verbosity", "version",
    "wf-checking", "write-partitions-to"
        // clang-format on
    };
  }

  std::string get(const Options& options, const std::string& name)
  {
    Trace("options") << "Options::getOption(" << name << ")" << std::endl;
    // clang-format off
    auto it = NAME_TO_ENUM.find(name);
    if (it == NAME_TO_ENUM.end()) {
      throw OptionException("Unrecognized option key or setting: " + name);
    }
    switch (it->second) {
      case OptionEnum::ABSTRACT_VALUES: return options.smt.abstractValues ? "true" : "false";
      case OptionEnum::ACKERMANN: return options.smt.ackermann ? "true" : "false";
      case OptionEnum::APPEND_LEARNED_LITERALS_TO_CUBES: return options.parallel.appendLearnedLiteralsToCubes ? "true" : "false";
      case OptionEnum::APPROX_BRANCH_DEPTH: return std::to_string(options.arith.maxApproxDepth);
      case OptionEnum::ARITH_BRAB: return options.arith.brabTest ? "true" : "false";
      case OptionEnum::ARITH_EQ_SOLVER: return options.arith.arithEqSolver ? "true" : "false";
      case OptionEnum::ARITH_EXP: return options.arith.arithExp ? "true" : "false";
      case OptionEnum::ARITH_PROP: { std::stringstream s; s << options.arith.arithPropagationMode; return s.str(); }
      case OptionEnum::ARITH_PROP_CLAUSES: return std::to_string(options.arith.arithPropAsLemmaLength);
      case OptionEnum::ARITH_REWRITE_EQUALITIES: return options.arith.arithRewriteEq ? "true" : "false";
      case OptionEnum::ARITH_STATIC_LEARNING: return options.arith.arithStaticLearning ? "true" : "false";
      case OptionEnum::ARRAYS_EAGER_INDEX: return options.arrays.arraysEagerIndexSplitting ? "true" : "false";
      case OptionEnum::ARRAYS_EAGER_LEMMAS: return options.arrays.arraysEagerLemmas ? "true" : "false";
      case OptionEnum::ARRAYS_EXP: return options.arrays.arraysExp ? "true" : "false";
      case OptionEnum::ARRAYS_OPTIMIZE_LINEAR: return options.arrays.arraysOptimizeLinear ? "true" : "false";
      case OptionEnum::ARRAYS_PROP: return std::to_string(options.arrays.arraysPropagate);
      case OptionEnum::ARRAYS_REDUCE_SHARING: return options.arrays.arraysReduceSharing ? "true" : "false";
      case OptionEnum::ARRAYS_WEAK_EQUIV: return options.arrays.arraysWeakEquivalence ? "true" : "false";
      case OptionEnum::ASSIGN_FUNCTION_VALUES: return options.theory.assignFunctionValues ? "true" : "false";
      case OptionEnum::BAGS: return options.bags.bags ? "true" : "false";
      case OptionEnum::BITBLAST: { std::stringstream s; s << options.bv.bitblastMode; return s.str(); }
      case OptionEnum::BITWISE_EQ: return options.bv.bitwiseEq ? "true" : "false";
      case OptionEnum::BOOL_TO_BV: { std::stringstream s; s << options.bv.boolToBitvector; return s.str(); }
      case OptionEnum::BV_ASSERT_INPUT: return options.bv.bvAssertInput ? "true" : "false";
      case OptionEnum::BV_EAGER_EVAL: return options.bv.bvEagerEval ? "true" : "false";
      case OptionEnum::BV_EQ_ENGINE: return options.bv.bvEqEngine ? "true" : "false";
      case OptionEnum::BV_GAUSS_ELIM: return options.bv.bvGaussElim ? "true" : "false";
      case OptionEnum::BV_INTRO_POW2: return options.bv.bvIntroducePow2 ? "true" : "false";
      case OptionEnum::BV_PRINT_CONSTS_AS_INDEXED_SYMBOLS: return options.printer.bvPrintConstsAsIndexedSymbols ? "true" : "false";
      case OptionEnum::BV_PROPAGATE: return options.bv.bitvectorPropagate ? "true" : "false";
      case OptionEnum::BV_RW_EXTEND_EQ: return options.bv.rwExtendEq ? "true" : "false";
      case OptionEnum::BV_SAT_SOLVER: { std::stringstream s; s << options.bv.bvSatSolver; return s.str(); }
      case OptionEnum::BV_SOLVER: { std::stringstream s; s << options.bv.bvSolver; return s.str(); }
      case OptionEnum::BV_TO_BOOL: return options.bv.bitvectorToBool ? "true" : "false";
      case OptionEnum::BV_TO_INT_USE_POW2: return options.smt.bvToIntUsePow2 ? "true" : "false";
      case OptionEnum::BVAND_INTEGER_GRANULARITY: return std::to_string(options.smt.BVAndIntegerGranularity);
      case OptionEnum::CBQI: return options.quantifiers.conflictBasedInst ? "true" : "false";
      case OptionEnum::CBQI_ALL_CONFLICT: return options.quantifiers.cbqiAllConflict ? "true" : "false";
      case OptionEnum::CBQI_MODE: { std::stringstream s; s << options.quantifiers.cbqiMode; return s.str(); }
      case OptionEnum::CBQI_SKIP_RD: return options.quantifiers.cbqiSkipRd ? "true" : "false";
      case OptionEnum::CBQI_TCONSTRAINT: return options.quantifiers.cbqiTConstraint ? "true" : "false";
      case OptionEnum::CBQI_VO_EXP: return options.quantifiers.cbqiVoExp ? "true" : "false";
      case OptionEnum::CDT_BISIMILAR: return options.datatypes.cdtBisimilar ? "true" : "false";
      case OptionEnum::CEGIS_SAMPLE: { std::stringstream s; s << options.quantifiers.cegisSample; return s.str(); }
      case OptionEnum::CEGQI: return options.quantifiers.cegqi ? "true" : "false";
      case OptionEnum::CEGQI_ALL: return options.quantifiers.cegqiAll ? "true" : "false";
      case OptionEnum::CEGQI_BV: return options.quantifiers.cegqiBv ? "true" : "false";
      case OptionEnum::CEGQI_BV_CONCAT_INV: return options.quantifiers.cegqiBvConcInv ? "true" : "false";
      case OptionEnum::CEGQI_BV_INEQ: { std::stringstream s; s << options.quantifiers.cegqiBvIneqMode; return s.str(); }
      case OptionEnum::CEGQI_BV_INTERLEAVE_VALUE: return options.quantifiers.cegqiBvInterleaveValue ? "true" : "false";
      case OptionEnum::CEGQI_BV_LINEAR: return options.quantifiers.cegqiBvLinearize ? "true" : "false";
      case OptionEnum::CEGQI_BV_RM_EXTRACT: return options.quantifiers.cegqiBvRmExtract ? "true" : "false";
      case OptionEnum::CEGQI_BV_SOLVE_NL: return options.quantifiers.cegqiBvSolveNl ? "true" : "false";
      case OptionEnum::CEGQI_FULL: return options.quantifiers.cegqiFullEffort ? "true" : "false";
      case OptionEnum::CEGQI_INF_INT: return options.quantifiers.cegqiUseInfInt ? "true" : "false";
      case OptionEnum::CEGQI_INF_REAL: return options.quantifiers.cegqiUseInfReal ? "true" : "false";
      case OptionEnum::CEGQI_INNERMOST: return options.quantifiers.cegqiInnermost ? "true" : "false";
      case OptionEnum::CEGQI_MIDPOINT: return options.quantifiers.cegqiMidpoint ? "true" : "false";
      case OptionEnum::CEGQI_MIN_BOUNDS: return options.quantifiers.cegqiMinBounds ? "true" : "false";
      case OptionEnum::CEGQI_MULTI_INST: return options.quantifiers.cegqiMultiInst ? "true" : "false";
      case OptionEnum::CEGQI_NESTED_QE: return options.quantifiers.cegqiNestedQE ? "true" : "false";
      case OptionEnum::CEGQI_NOPT: return options.quantifiers.cegqiNopt ? "true" : "false";
      case OptionEnum::CEGQI_ROUND_UP_LIA: return options.quantifiers.cegqiRoundUpLowerLia ? "true" : "false";
      case OptionEnum::CHECK_ABDUCTS: return options.smt.checkAbducts ? "true" : "false";
      case OptionEnum::CHECK_INTERPOLANTS: return options.smt.checkInterpolants ? "true" : "false";
      case OptionEnum::CHECK_MODELS: return options.smt.checkModels ? "true" : "false";
      case OptionEnum::CHECK_PROOF_STEPS: return options.proof.checkProofSteps ? "true" : "false";
      case OptionEnum::CHECK_PROOFS: return options.smt.checkProofs ? "true" : "false";
      case OptionEnum::CHECK_SYNTH_SOL: return options.smt.checkSynthSol ? "true" : "false";
      case OptionEnum::CHECK_UNSAT_CORES: return options.smt.checkUnsatCores ? "true" : "false";
      case OptionEnum::CHECKS_BEFORE_PARTITION: return std::to_string(options.parallel.checksBeforePartitioning);
      case OptionEnum::CHECKS_BETWEEN_PARTITIONS: return std::to_string(options.parallel.checksBetweenPartitions);
      case OptionEnum::COLLECT_PIVOT_STATS: return options.arith.collectPivots ? "true" : "false";
      case OptionEnum::COMPUTE_PARTITIONS: return std::to_string(options.parallel.computePartitions);
      case OptionEnum::COND_VAR_SPLIT_QUANT: { std::stringstream s; s << options.quantifiers.condVarSplitQuant; return s.str(); }
      case OptionEnum::CONDENSE_FUNCTION_VALUES: return options.theory.condenseFunctionValues ? "true" : "false";
      case OptionEnum::CONFLICT_PROCESS: { std::stringstream s; s << options.theory.conflictProcessMode; return s.str(); }
      case OptionEnum::CONJECTURE_GEN: return options.quantifiers.conjectureGen ? "true" : "false";
      case OptionEnum::CONJECTURE_GEN_GT_ENUM: return std::to_string(options.quantifiers.conjectureGenGtEnum);
      case OptionEnum::CONJECTURE_GEN_MAX_DEPTH: return std::to_string(options.quantifiers.conjectureGenMaxDepth);
      case OptionEnum::CONJECTURE_GEN_PER_ROUND: return std::to_string(options.quantifiers.conjectureGenPerRound);
      case OptionEnum::CONS_EXP_TRIGGERS: return options.quantifiers.consExpandTriggers ? "true" : "false";
      case OptionEnum::COPYRIGHT: return options.driver.showCopyright ? "true" : "false";
      case OptionEnum::CUT_ALL_BOUNDED: return options.arith.doCutAllBounded ? "true" : "false";
      case OptionEnum::DAG_THRESH: return std::to_string(options.printer.dagThresh);
      case OptionEnum::DATATYPES_EXP: return options.datatypes.datatypesExp ? "true" : "false";
      case OptionEnum::DEBUG_CHECK_MODELS: return options.smt.debugCheckModels ? "true" : "false";
      case OptionEnum::DECISION: { std::stringstream s; s << options.decision.decisionMode; return s.str(); }
      case OptionEnum::DEEP_RESTART: { std::stringstream s; s << options.smt.deepRestartMode; return s.str(); }
      case OptionEnum::DEEP_RESTART_FACTOR: return std::to_string(options.smt.deepRestartFactor);
      case OptionEnum::DEFAULT_FUNCTION_VALUE_MODE: { std::stringstream s; s << options.theory.defaultFunctionValueMode; return s.str(); }
      case OptionEnum::DIFFICULTY_MODE: { std::stringstream s; s << options.smt.difficultyMode; return s.str(); }
      case OptionEnum::DIO_DECOMPS: return options.arith.exportDioDecompositions ? "true" : "false";
      case OptionEnum::DIO_SOLVER: return options.arith.arithDioSolver ? "true" : "false";
      case OptionEnum::DIO_TURNS: return std::to_string(options.arith.dioSolverTurns);
      case OptionEnum::DRAT_BINARY_FORMAT: return options.proof.dratBinaryFormat ? "true" : "false";
      case OptionEnum::DT_BINARY_SPLIT: return options.datatypes.dtBinarySplit ? "true" : "false";
      case OptionEnum::DT_BLAST_SPLITS: return options.datatypes.dtBlastSplits ? "true" : "false";
      case OptionEnum::DT_CYCLIC: return options.datatypes.dtCyclic ? "true" : "false";
      case OptionEnum::DT_INFER_AS_LEMMAS: return options.datatypes.dtInferAsLemmas ? "true" : "false";
      case OptionEnum::DT_NESTED_REC: return options.datatypes.dtNestedRec ? "true" : "false";
      case OptionEnum::DT_POLITE_OPTIMIZE: return options.datatypes.dtPoliteOptimize ? "true" : "false";
      case OptionEnum::DT_SHARE_SEL: return options.datatypes.dtSharedSelectors ? "true" : "false";
      case OptionEnum::DT_STC_IND: return options.quantifiers.dtStcInduction ? "true" : "false";
      case OptionEnum::DT_VAR_EXP_QUANT: return options.quantifiers.dtVarExpandQuant ? "true" : "false";
      case OptionEnum::DUMP_DIFFICULTY: return options.driver.dumpDifficulty ? "true" : "false";
      case OptionEnum::DUMP_INSTANTIATIONS: return options.driver.dumpInstantiations ? "true" : "false";
      case OptionEnum::DUMP_INSTANTIATIONS_DEBUG: return options.driver.dumpInstantiationsDebug ? "true" : "false";
      case OptionEnum::DUMP_MODELS: return options.driver.dumpModels ? "true" : "false";
      case OptionEnum::DUMP_PROOFS: return options.driver.dumpProofs ? "true" : "false";
      case OptionEnum::DUMP_UNSAT_CORES: return options.driver.dumpUnsatCores ? "true" : "false";
      case OptionEnum::DUMP_UNSAT_CORES_LEMMAS: return options.driver.dumpUnsatCoresLemmas ? "true" : "false";
      case OptionEnum::E_MATCHING: return options.quantifiers.eMatching ? "true" : "false";
      case OptionEnum::EAGER_ARITH_BV_CONV: return options.uf.eagerArithBvConv ? "true" : "false";
      case OptionEnum::EARLY_EXIT: return options.driver.earlyExit ? "true" : "false";
      case OptionEnum::EARLY_ITE_REMOVAL: return options.smt.earlyIteRemoval ? "true" : "false";
      case OptionEnum::EE_MODE: { std::stringstream s; s << options.theory.eeMode; return s.str(); }
      case OptionEnum::ELIM_TAUT_QUANT: return options.quantifiers.elimTautQuant ? "true" : "false";
      case OptionEnum::ENUM_INST: return options.quantifiers.enumInst ? "true" : "false";
      case OptionEnum::ENUM_INST_INTERLEAVE: return options.quantifiers.enumInstInterleave ? "true" : "false";
      case OptionEnum::ENUM_INST_LIMIT: return std::to_string(options.quantifiers.enumInstLimit);
      case OptionEnum::ENUM_INST_RD: return options.quantifiers.enumInstRd ? "true" : "false";
      case OptionEnum::ENUM_INST_STRATIFY: return options.quantifiers.enumInstStratify ? "true" : "false";
      case OptionEnum::ENUM_INST_SUM: return options.quantifiers.enumInstSum ? "true" : "false";
      case OptionEnum::ERR: { std::stringstream s; s << options.base.err; return s.str(); }
      case OptionEnum::ERROR_SELECTION_RULE: { std::stringstream s; s << options.arith.arithErrorSelectionRule; return s.str(); }
      case OptionEnum::EXPR_DEPTH: return std::to_string(options.printer.nodeDepth);
      case OptionEnum::EXT_REW_PREP: { std::stringstream s; s << options.smt.extRewPrep; return s.str(); }
      case OptionEnum::EXT_REWRITE_QUANT: return options.quantifiers.extRewriteQuant ? "true" : "false";
      case OptionEnum::FC_PENALTIES: return options.arith.havePenalties ? "true" : "false";
      case OptionEnum::FF: return options.ff.ff ? "true" : "false";
      case OptionEnum::FF_BITSUM: return options.ff.ffBitsum ? "true" : "false";
      case OptionEnum::FF_ELIM_DISJUNCTIVE_BIT: return options.ff.ffElimDisjunctiveBit ? "true" : "false";
      case OptionEnum::FF_FIELD_POLYS: return options.ff.ffFieldPolys ? "true" : "false";
      case OptionEnum::FF_SOLVER: { std::stringstream s; s << options.ff.ffSolver; return s.str(); }
      case OptionEnum::FF_TRACE_GB: return options.ff.ffTraceGb ? "true" : "false";
      case OptionEnum::FILENAME: return options.driver.filename;
      case OptionEnum::FILESYSTEM_ACCESS: return options.parser.filesystemAccess ? "true" : "false";
      case OptionEnum::FINITE_MODEL_FIND: return options.quantifiers.finiteModelFind ? "true" : "false";
      case OptionEnum::FLATTEN_HO_CHAINS: return options.printer.flattenHOChains ? "true" : "false";
      case OptionEnum::FMF_BOUND: return options.quantifiers.fmfBound ? "true" : "false";
      case OptionEnum::FMF_BOUND_BLAST: return options.quantifiers.fmfBoundBlast ? "true" : "false";
      case OptionEnum::FMF_BOUND_LAZY: return options.quantifiers.fmfBoundLazy ? "true" : "false";
      case OptionEnum::FMF_FUN: return options.quantifiers.fmfFunWellDefined ? "true" : "false";
      case OptionEnum::FMF_FUN_RLV: return options.quantifiers.fmfFunWellDefinedRelevant ? "true" : "false";
      case OptionEnum::FMF_MBQI: { std::stringstream s; s << options.quantifiers.fmfMbqiMode; return s.str(); }
      case OptionEnum::FMF_TYPE_COMPLETION_THRESH: return std::to_string(options.quantifiers.fmfTypeCompletionThresh);
      case OptionEnum::FORCE_LOGIC: return options.parser.forceLogicString;
      case OptionEnum::FORCE_NO_LIMIT_CPU_WHILE_DUMP: return options.driver.forceNoLimitCpuWhileDump ? "true" : "false";
      case OptionEnum::FOREIGN_THEORY_REWRITE: return options.smt.foreignTheoryRewrite ? "true" : "false";
      case OptionEnum::FP: return options.fp.fp ? "true" : "false";
      case OptionEnum::FP_EXP: return options.fp.fpExp ? "true" : "false";
      case OptionEnum::FP_LAZY_WB: return options.fp.fpLazyWb ? "true" : "false";
      case OptionEnum::FRESH_BINDERS: return options.parser.freshBinders ? "true" : "false";
      case OptionEnum::FRESH_DECLARATIONS: return options.parser.freshDeclarations ? "true" : "false";
      case OptionEnum::FULL_SATURATE_QUANT: return options.quantifiers.fullSaturateQuant ? "true" : "false";
      case OptionEnum::FULL_SYGUS_VERIFY: return options.quantifiers.fullSygusVerify ? "true" : "false";
      case OptionEnum::GLOBAL_DECLARATIONS: return options.parser.globalDeclarations ? "true" : "false";
      case OptionEnum::GLOBAL_NEGATE: return options.quantifiers.globalNegate ? "true" : "false";
      case OptionEnum::HELP: return options.driver.help ? "true" : "false";
      case OptionEnum::HELP_REGULAR: return options.driver.helpRegular ? "true" : "false";
      case OptionEnum::HEURISTIC_PIVOTS: return std::to_string(options.arith.arithHeuristicPivots);
      case OptionEnum::HO_ELIM: return options.quantifiers.hoElim ? "true" : "false";
      case OptionEnum::HO_ELIM_STORE_AX: return options.quantifiers.hoElimStoreAx ? "true" : "false";
      case OptionEnum::HO_MATCHING: return options.quantifiers.hoMatching ? "true" : "false";
      case OptionEnum::HO_MERGE_TERM_DB: return options.quantifiers.hoMergeTermDb ? "true" : "false";
      case OptionEnum::IAND_MODE: { std::stringstream s; s << options.smt.iandMode; return s.str(); }
      case OptionEnum::IEVAL: { std::stringstream s; s << options.quantifiers.ievalMode; return s.str(); }
      case OptionEnum::IN: { std::stringstream s; s << options.base.in; return s.str(); }
      case OptionEnum::INCREMENT_TRIGGERS: return options.quantifiers.incrementTriggers ? "true" : "false";
      case OptionEnum::INCREMENTAL: return options.base.incrementalSolving ? "true" : "false";
      case OptionEnum::INST_MAX_LEVEL: return std::to_string(options.quantifiers.instMaxLevel);
      case OptionEnum::INST_MAX_ROUNDS: return std::to_string(options.quantifiers.instMaxRounds);
      case OptionEnum::INST_NO_ENTAIL: return options.quantifiers.instNoEntail ? "true" : "false";
      case OptionEnum::INST_WHEN: { std::stringstream s; s << options.quantifiers.instWhenMode; return s.str(); }
      case OptionEnum::INST_WHEN_PHASE: return std::to_string(options.quantifiers.instWhenPhase);
      case OptionEnum::INT_WF_IND: return options.quantifiers.intWfInduction ? "true" : "false";
      case OptionEnum::INTERACTIVE: return options.driver.interactive ? "true" : "false";
      case OptionEnum::INTERPOLANTS_MODE: { std::stringstream s; s << options.smt.interpolantsMode; return s.str(); }
      case OptionEnum::ITE_DTT_SPLIT_QUANT: return options.quantifiers.iteDtTesterSplitQuant ? "true" : "false";
      case OptionEnum::ITE_LIFT_QUANT: { std::stringstream s; s << options.quantifiers.iteLiftQuant; return s.str(); }
      case OptionEnum::ITE_SIMP: return options.smt.doITESimp ? "true" : "false";
      case OptionEnum::JH_RLV_ORDER: return options.decision.jhRlvOrder ? "true" : "false";
      case OptionEnum::JH_SKOLEM: { std::stringstream s; s << options.decision.jhSkolemMode; return s.str(); }
      case OptionEnum::JH_SKOLEM_RLV: { std::stringstream s; s << options.decision.jhSkolemRlvMode; return s.str(); }
      case OptionEnum::LANG: { std::stringstream s; s << options.base.inputLanguage; return s.str(); }
      case OptionEnum::LEARNED_REWRITE: return options.smt.learnedRewrite ? "true" : "false";
      case OptionEnum::LEMMA_INPROCESS: { std::stringstream s; s << options.theory.lemmaInprocess; return s.str(); }
      case OptionEnum::LEMMA_INPROCESS_INFER_EQ_LIT: return options.theory.lemmaInprocessInferEqLit ? "true" : "false";
      case OptionEnum::LEMMA_INPROCESS_SUBS: { std::stringstream s; s << options.theory.lemmaInprocessSubsMode; return s.str(); }
      case OptionEnum::LEMMAS_ON_REPLAY_FAILURE: return options.arith.replayFailureLemma ? "true" : "false";
      case OptionEnum::LFSC_EXPAND_TRUST: return options.proof.lfscExpandTrust ? "true" : "false";
      case OptionEnum::LFSC_FLATTEN: return options.proof.lfscFlatten ? "true" : "false";
      case OptionEnum::LITERAL_MATCHING: { std::stringstream s; s << options.quantifiers.literalMatchMode; return s.str(); }
      case OptionEnum::MACROS_QUANT: return options.quantifiers.macrosQuant ? "true" : "false";
      case OptionEnum::MACROS_QUANT_MODE: { std::stringstream s; s << options.quantifiers.macrosQuantMode; return s.str(); }
      case OptionEnum::MAXCUTSINCONTEXT: return std::to_string(options.arith.maxCutsInContext);
      case OptionEnum::MBQI: return options.quantifiers.mbqi ? "true" : "false";
      case OptionEnum::MBQI_FAST_SYGUS: return options.quantifiers.mbqiFastSygus ? "true" : "false";
      case OptionEnum::MBQI_INTERLEAVE: return options.quantifiers.mbqiInterleave ? "true" : "false";
      case OptionEnum::MBQI_ONE_INST_PER_ROUND: return options.quantifiers.fmfOneInstPerRound ? "true" : "false";
      case OptionEnum::MINIMAL_UNSAT_CORES: return options.smt.minimalUnsatCores ? "true" : "false";
      case OptionEnum::MINISAT_DUMP_DIMACS: return options.prop.minisatDumpDimacs ? "true" : "false";
      case OptionEnum::MINISAT_SIMPLIFICATION: { std::stringstream s; s << options.prop.minisatSimpMode; return s.str(); }
      case OptionEnum::MINISCOPE_QUANT: { std::stringstream s; s << options.quantifiers.miniscopeQuant; return s.str(); }
      case OptionEnum::MINISCOPE_QUANT_USER: return options.quantifiers.miniscopeQuantUser ? "true" : "false";
      case OptionEnum::MIPLIB_TRICK: return options.arith.arithMLTrick ? "true" : "false";
      case OptionEnum::MIPLIB_TRICK_SUBS: return std::to_string(options.arith.arithMLTrickSubstitutions);
      case OptionEnum::MODEL_BASED_ARITH_BV_CONV: return options.uf.modelBasedArithBvConv ? "true" : "false";
      case OptionEnum::MODEL_CORES: { std::stringstream s; s << options.smt.modelCoresMode; return s.str(); }
      case OptionEnum::MODEL_U_PRINT: { std::stringstream s; s << options.printer.modelUninterpPrint; return s.str(); }
      case OptionEnum::MODEL_VAR_ELIM_UNEVAL: return options.smt.modelVarElimUneval ? "true" : "false";
      case OptionEnum::MULTI_TRIGGER_CACHE: return options.quantifiers.multiTriggerCache ? "true" : "false";
      case OptionEnum::MULTI_TRIGGER_LINEAR: return options.quantifiers.multiTriggerLinear ? "true" : "false";
      case OptionEnum::MULTI_TRIGGER_PRIORITY: return options.quantifiers.multiTriggerPriority ? "true" : "false";
      case OptionEnum::MULTI_TRIGGER_WHEN_SINGLE: return options.quantifiers.multiTriggerWhenSingle ? "true" : "false";
      case OptionEnum::NEW_PROP: return options.arith.newProp ? "true" : "false";
      case OptionEnum::NL_COV: return options.arith.nlCov ? "true" : "false";
      case OptionEnum::NL_COV_FORCE: return options.arith.nlCovForce ? "true" : "false";
      case OptionEnum::NL_COV_LIFT: { std::stringstream s; s << options.arith.nlCovLifting; return s.str(); }
      case OptionEnum::NL_COV_LINEAR_MODEL: { std::stringstream s; s << options.arith.nlCovLinearModel; return s.str(); }
      case OptionEnum::NL_COV_PROJ: { std::stringstream s; s << options.arith.nlCovProjection; return s.str(); }
      case OptionEnum::NL_COV_PRUNE: return options.arith.nlCovPrune ? "true" : "false";
      case OptionEnum::NL_COV_VAR_ELIM: return options.arith.nlCovVarElim ? "true" : "false";
      case OptionEnum::NL_EXT: { std::stringstream s; s << options.arith.nlExt; return s.str(); }
      case OptionEnum::NL_EXT_ENT_CONF: return options.arith.nlExtEntailConflicts ? "true" : "false";
      case OptionEnum::NL_EXT_FACTOR: return options.arith.nlExtFactor ? "true" : "false";
      case OptionEnum::NL_EXT_INC_PREC: return options.arith.nlExtIncPrecision ? "true" : "false";
      case OptionEnum::NL_EXT_PURIFY: return options.arith.nlExtPurify ? "true" : "false";
      case OptionEnum::NL_EXT_RBOUND: return options.arith.nlExtResBound ? "true" : "false";
      case OptionEnum::NL_EXT_REWRITE: return options.arith.nlExtRewrites ? "true" : "false";
      case OptionEnum::NL_EXT_SPLIT_ZERO: return options.arith.nlExtSplitZero ? "true" : "false";
      case OptionEnum::NL_EXT_TF_TAYLOR_DEG: return std::to_string(options.arith.nlExtTfTaylorDegree);
      case OptionEnum::NL_EXT_TF_TPLANES: return options.arith.nlExtTfTangentPlanes ? "true" : "false";
      case OptionEnum::NL_EXT_TPLANES: return options.arith.nlExtTangentPlanes ? "true" : "false";
      case OptionEnum::NL_EXT_TPLANES_INTERLEAVE: return options.arith.nlExtTangentPlanesInterleave ? "true" : "false";
      case OptionEnum::NL_ICP: return options.arith.nlICP ? "true" : "false";
      case OptionEnum::NL_RLV: { std::stringstream s; s << options.arith.nlRlvMode; return s.str(); }
      case OptionEnum::NL_RLV_ASSERT_BOUNDS: return options.arith.nlRlvAssertBounds ? "true" : "false";
      case OptionEnum::ON_REPEAT_ITE_SIMP: return options.smt.doITESimpOnRepeat ? "true" : "false";
      case OptionEnum::OPT_RES_RECONSTRUCTION_SIZE: return options.proof.optResReconSize ? "true" : "false";
      case OptionEnum::ORACLES: return options.quantifiers.oracles ? "true" : "false";
      case OptionEnum::OUT: { std::stringstream s; s << options.base.out; return s.str(); }
      case OptionEnum::OUTPUT_LANG: { std::stringstream s; s << options.printer.outputLanguage; return s.str(); }
      case OptionEnum::PARSE_ONLY: return options.base.parseOnly ? "true" : "false";
      case OptionEnum::PARSE_SKOLEM_DEFINITIONS: return options.parser.parseSkolemDefinitions ? "true" : "false";
      case OptionEnum::PARSING_MODE: { std::stringstream s; s << options.parser.parsingMode; return s.str(); }
      case OptionEnum::PARTIAL_TRIGGERS: return options.quantifiers.partialTriggers ? "true" : "false";
      case OptionEnum::PARTITION_CHECK: { std::stringstream s; s << options.parallel.partitionCheck; return s.str(); }
      case OptionEnum::PARTITION_CONFLICT_SIZE: return std::to_string(options.parallel.partitionConflictSize);
      case OptionEnum::PARTITION_START_TIME: return std::to_string(options.parallel.partitionStartTime);
      case OptionEnum::PARTITION_STRATEGY: { std::stringstream s; s << options.parallel.partitionStrategy; return s.str(); }
      case OptionEnum::PARTITION_TIME_INTERVAL: return std::to_string(options.parallel.partitionTimeInterval);
      case OptionEnum::PARTITION_TLIMIT: return std::to_string(options.parallel.partitionTimeLimit);
      case OptionEnum::PARTITION_WHEN: { std::stringstream s; s << options.parallel.partitionWhen; return s.str(); }
      case OptionEnum::PB_REWRITES: return options.arith.pbRewrites ? "true" : "false";
      case OptionEnum::PIVOT_THRESHOLD: return std::to_string(options.arith.arithPivotThreshold);
      case OptionEnum::PLUGIN_NOTIFY_SAT_CLAUSE_IN_SOLVE: return options.base.pluginNotifySatClauseInSolve ? "true" : "false";
      case OptionEnum::PLUGIN_SHARE_SKOLEMS: return options.base.pluginShareSkolems ? "true" : "false";
      case OptionEnum::POOL_INST: return options.quantifiers.poolInst ? "true" : "false";
      case OptionEnum::PORTFOLIO_JOBS: return std::to_string(options.driver.portfolioJobs);
      case OptionEnum::PP_ASSERT_MAX_SUB_SIZE: return std::to_string(options.arith.ppAssertMaxSubSize);
      case OptionEnum::PRE_SKOLEM_QUANT: { std::stringstream s; s << options.quantifiers.preSkolemQuant; return s.str(); }
      case OptionEnum::PRE_SKOLEM_QUANT_NESTED: return options.quantifiers.preSkolemQuantNested ? "true" : "false";
      case OptionEnum::PRENEX_QUANT: { std::stringstream s; s << options.quantifiers.prenexQuant; return s.str(); }
      case OptionEnum::PRENEX_QUANT_USER: return options.quantifiers.prenexQuantUser ? "true" : "false";
      case OptionEnum::PREPROCESS_ONLY: return options.base.preprocessOnly ? "true" : "false";
      case OptionEnum::PREREGISTER_MODE: { std::stringstream s; s << options.prop.preRegisterMode; return s.str(); }
      case OptionEnum::PRINT_ARITH_LIT_TOKEN: return options.printer.printArithLitToken ? "true" : "false";
      case OptionEnum::PRINT_CORES_FULL: return options.smt.printCoresFull ? "true" : "false";
      case OptionEnum::PRINT_DOT_CLUSTERS: return options.proof.printDotClusters ? "true" : "false";
      case OptionEnum::PRINT_INST: { std::stringstream s; s << options.quantifiers.printInstMode; return s.str(); }
      case OptionEnum::PRINT_INST_FULL: return options.quantifiers.printInstFull ? "true" : "false";
      case OptionEnum::PRINT_SKOLEM_DEFINITIONS: return options.printer.printSkolemDefinitions ? "true" : "false";
      case OptionEnum::PRINT_SUCCESS: return options.driver.printSuccess ? "true" : "false";
      case OptionEnum::PRODUCE_ABDUCTS: return options.smt.produceAbducts ? "true" : "false";
      case OptionEnum::PRODUCE_ASSERTIONS: return options.smt.produceAssertions ? "true" : "false";
      case OptionEnum::PRODUCE_ASSIGNMENTS: return options.smt.produceAssignments ? "true" : "false";
      case OptionEnum::PRODUCE_DIFFICULTY: return options.smt.produceDifficulty ? "true" : "false";
      case OptionEnum::PRODUCE_INTERPOLANTS: return options.smt.produceInterpolants ? "true" : "false";
      case OptionEnum::PRODUCE_LEARNED_LITERALS: return options.smt.produceLearnedLiterals ? "true" : "false";
      case OptionEnum::PRODUCE_MODELS: return options.smt.produceModels ? "true" : "false";
      case OptionEnum::PRODUCE_PROOFS: return options.smt.produceProofs ? "true" : "false";
      case OptionEnum::PRODUCE_UNSAT_ASSUMPTIONS: return options.smt.unsatAssumptions ? "true" : "false";
      case OptionEnum::PRODUCE_UNSAT_CORES: return options.smt.produceUnsatCores ? "true" : "false";
      case OptionEnum::PROOF_ALETHE_DEFINE_SKOLEMS: return options.proof.proofAletheDefineSkolems ? "true" : "false";
      case OptionEnum::PROOF_ALETHE_RES_PIVOTS: return options.proof.proofAletheResPivots ? "true" : "false";
      case OptionEnum::PROOF_ALLOW_TRUST: return options.proof.proofAllowTrust ? "true" : "false";
      case OptionEnum::PROOF_CHECK: { std::stringstream s; s << options.proof.proofCheck; return s.str(); }
      case OptionEnum::PROOF_DAG_GLOBAL: return options.proof.proofDagGlobal ? "true" : "false";
      case OptionEnum::PROOF_DOT_DAG: return options.proof.printDotAsDAG ? "true" : "false";
      case OptionEnum::PROOF_ELIM_SUBTYPES: return options.proof.proofElimSubtypes ? "true" : "false";
      case OptionEnum::PROOF_FORMAT_MODE: { std::stringstream s; s << options.proof.proofFormatMode; return s.str(); }
      case OptionEnum::PROOF_GRANULARITY: { std::stringstream s; s << options.proof.proofGranularityMode; return s.str(); }
      case OptionEnum::PROOF_MODE: { std::stringstream s; s << options.smt.proofMode; return s.str(); }
      case OptionEnum::PROOF_PEDANTIC: return std::to_string(options.proof.proofPedantic);
      case OptionEnum::PROOF_PP_MERGE: return options.proof.proofPpMerge ? "true" : "false";
      case OptionEnum::PROOF_PRINT_CONCLUSION: return options.proof.proofPrintConclusion ? "true" : "false";
      case OptionEnum::PROOF_PRINT_REFERENCE: return options.proof.proofPrintReference ? "true" : "false";
      case OptionEnum::PROOF_PRUNE_INPUT: return options.proof.proofPruneInput ? "true" : "false";
      case OptionEnum::PROOF_REWRITE_RCONS_REC_LIMIT: return std::to_string(options.proof.proofRewriteRconsRecLimit);
      case OptionEnum::PROOF_REWRITE_RCONS_STEP_LIMIT: return std::to_string(options.proof.proofRewriteRconsStepLimit);
      case OptionEnum::PROP_PROOF_MODE: { std::stringstream s; s << options.proof.propProofMode; return s.str(); }
      case OptionEnum::PROP_ROW_LENGTH: return std::to_string(options.arith.arithPropagateMaxLength);
      case OptionEnum::PURIFY_TRIGGERS: return options.quantifiers.purifyTriggers ? "true" : "false";
      case OptionEnum::QUANT_ALPHA_EQUIV: return options.quantifiers.quantAlphaEquiv ? "true" : "false";
      case OptionEnum::QUANT_DSPLIT: { std::stringstream s; s << options.quantifiers.quantDynamicSplit; return s.str(); }
      case OptionEnum::QUANT_FUN_WD: return options.quantifiers.quantFunWellDefined ? "true" : "false";
      case OptionEnum::QUANT_IND: return options.quantifiers.quantInduction ? "true" : "false";
      case OptionEnum::QUANT_REP_MODE: { std::stringstream s; s << options.quantifiers.quantRepMode; return s.str(); }
      case OptionEnum::RANDOM_FREQ: return std::to_string(options.prop.satRandomFreq);
      case OptionEnum::RANDOM_PARTITIONING: return options.parallel.randomPartitioning ? "true" : "false";
      case OptionEnum::RE_ELIM: { std::stringstream s; s << options.strings.regExpElim; return s.str(); }
      case OptionEnum::RE_INTER_MODE: { std::stringstream s; s << options.strings.stringRegExpInterMode; return s.str(); }
      case OptionEnum::REGISTER_QUANT_BODY_TERMS: return options.quantifiers.registerQuantBodyTerms ? "true" : "false";
      case OptionEnum::RELATIONAL_TRIGGERS: return options.quantifiers.relationalTriggers ? "true" : "false";
      case OptionEnum::RELEVANCE_FILTER: return options.theory.relevanceFilter ? "true" : "false";
      case OptionEnum::RELEVANT_TRIGGERS: return options.quantifiers.relevantTriggers ? "true" : "false";
      case OptionEnum::RELS_EXP: return options.sets.relsExp ? "true" : "false";
      case OptionEnum::REPEAT_SIMP: return options.smt.repeatSimp ? "true" : "false";
      case OptionEnum::REPLAY_EARLY_CLOSE_DEPTH: return std::to_string(options.arith.replayEarlyCloseDepths);
      case OptionEnum::REPLAY_LEMMA_REJECT_CUT: return std::to_string(options.arith.lemmaRejectCutSize);
      case OptionEnum::REPLAY_NUM_ERR_PENALTY: return std::to_string(options.arith.replayNumericFailurePenalty);
      case OptionEnum::REPLAY_REJECT_CUT: return std::to_string(options.arith.replayRejectCutSize);
      case OptionEnum::RESTART_INT_BASE: return std::to_string(options.prop.satRestartFirst);
      case OptionEnum::RESTART_INT_INC: return std::to_string(options.prop.satRestartInc);
      case OptionEnum::RESTRICT_PIVOTS: return options.arith.restrictedPivots ? "true" : "false";
      case OptionEnum::REVERT_ARITH_MODELS_ON_UNSAT: return options.arith.revertArithModels ? "true" : "false";
      case OptionEnum::RLIMIT: return std::to_string(options.base.cumulativeResourceLimit);
      case OptionEnum::RLIMIT_PER: return std::to_string(options.base.perCallResourceLimit);
      case OptionEnum::RR_TURNS: return std::to_string(options.arith.rrTurns);
      case OptionEnum::SAFE_OPTIONS: return options.base.safeOptions ? "true" : "false";
      case OptionEnum::SAT_PROOF_MIN_DIMACS: return options.proof.satProofMinDimacs ? "true" : "false";
      case OptionEnum::SAT_RANDOM_SEED: return std::to_string(options.prop.satRandomSeed);
      case OptionEnum::SAT_SOLVER: { std::stringstream s; s << options.prop.satSolver; return s.str(); }
      case OptionEnum::SE_SOLVE_INT: return options.arith.trySolveIntStandardEffort ? "true" : "false";
      case OptionEnum::SEED: return std::to_string(options.driver.seed);
      case OptionEnum::SEGV_SPIN: return options.driver.segvSpin ? "true" : "false";
      case OptionEnum::SEMANTIC_CHECKS: return options.parser.semanticChecks ? "true" : "false";
      case OptionEnum::SEP: return options.sep.sep ? "true" : "false";
      case OptionEnum::SEP_MIN_REFINE: return options.sep.sepMinimalRefine ? "true" : "false";
      case OptionEnum::SEP_PRE_SKOLEM_EMP: return options.sep.sepPreSkolemEmp ? "true" : "false";
      case OptionEnum::SEQ_ARRAY: { std::stringstream s; s << options.strings.seqArray; return s.str(); }
      case OptionEnum::SETS_CARD_EXP: return options.sets.setsCardExp ? "true" : "false";
      case OptionEnum::SETS_EXP: return options.sets.setsExp ? "true" : "false";
      case OptionEnum::SETS_PROXY_LEMMAS: return options.sets.setsProxyLemmas ? "true" : "false";
      case OptionEnum::SHOW_CONFIG: return options.driver.showConfiguration ? "true" : "false";
      case OptionEnum::SHOW_TRACE_TAGS: return options.driver.showTraceTags ? "true" : "false";
      case OptionEnum::SIMP_ITE_COMPRESS: return options.smt.compressItes ? "true" : "false";
      case OptionEnum::SIMP_WITH_CARE: return options.smt.simplifyWithCareEnabled ? "true" : "false";
      case OptionEnum::SIMPLEX_CHECK_PERIOD: return std::to_string(options.arith.arithSimplexCheckPeriod);
      case OptionEnum::SIMPLIFICATION: { std::stringstream s; s << options.smt.simplificationMode; return s.str(); }
      case OptionEnum::SIMPLIFICATION_BCP: return options.smt.simplificationBoolConstProp ? "true" : "false";
      case OptionEnum::SOI_QE: return options.arith.soiQuickExplain ? "true" : "false";
      case OptionEnum::SOLVE_BV_AS_INT: { std::stringstream s; s << options.smt.solveBVAsInt; return s.str(); }
      case OptionEnum::SOLVE_INT_AS_BV: return std::to_string(options.smt.solveIntAsBV);
      case OptionEnum::SOLVE_REAL_AS_INT: return options.smt.solveRealAsInt ? "true" : "false";
      case OptionEnum::SORT_INFERENCE: return options.smt.sortInference ? "true" : "false";
      case OptionEnum::STANDARD_EFFORT_VARIABLE_ORDER_PIVOTS: return std::to_string(options.arith.arithStandardCheckVarOrderPivots);
      case OptionEnum::STATIC_LEARNING: return options.smt.staticLearning ? "true" : "false";
      case OptionEnum::STATS: return options.base.statistics ? "true" : "false";
      case OptionEnum::STATS_ALL: return options.base.statisticsAll ? "true" : "false";
      case OptionEnum::STATS_EVERY_QUERY: return options.base.statisticsEveryQuery ? "true" : "false";
      case OptionEnum::STATS_INTERNAL: return options.base.statisticsInternal ? "true" : "false";
      case OptionEnum::STRICT_PARSING: return options.parser.strictParsing ? "true" : "false";
      case OptionEnum::STRINGS_ALPHA_CARD: return std::to_string(options.strings.stringsAlphaCard);
      case OptionEnum::STRINGS_CHECK_ENTAIL_LEN: return options.strings.stringCheckEntailLen ? "true" : "false";
      case OptionEnum::STRINGS_DEQ_EXT: return options.strings.stringsDeqExt ? "true" : "false";
      case OptionEnum::STRINGS_EAGER_EVAL: return options.strings.stringEagerEval ? "true" : "false";
      case OptionEnum::STRINGS_EAGER_LEN_RE: return options.strings.stringsEagerLenEntRegexp ? "true" : "false";
      case OptionEnum::STRINGS_EAGER_REG: return options.strings.stringEagerReg ? "true" : "false";
      case OptionEnum::STRINGS_EAGER_SOLVER: return options.strings.stringEagerSolver ? "true" : "false";
      case OptionEnum::STRINGS_EXP: return options.strings.stringExp ? "true" : "false";
      case OptionEnum::STRINGS_FF: return options.strings.stringFlatForms ? "true" : "false";
      case OptionEnum::STRINGS_FMF: return options.strings.stringFMF ? "true" : "false";
      case OptionEnum::STRINGS_INFER_AS_LEMMAS: return options.strings.stringInferAsLemmas ? "true" : "false";
      case OptionEnum::STRINGS_INFER_SYM: return options.strings.stringInferSym ? "true" : "false";
      case OptionEnum::STRINGS_LAZY_PP: return options.strings.stringLazyPreproc ? "true" : "false";
      case OptionEnum::STRINGS_LEN_NORM: return options.strings.stringLenNorm ? "true" : "false";
      case OptionEnum::STRINGS_MBR: return options.strings.stringModelBasedReduction ? "true" : "false";
      case OptionEnum::STRINGS_MODEL_MAX_LEN: return std::to_string(options.strings.stringsModelMaxLength);
      case OptionEnum::STRINGS_PROCESS_LOOP_MODE: { std::stringstream s; s << options.strings.stringProcessLoopMode; return s.str(); }
      case OptionEnum::STRINGS_RE_DERIVE_CONF: return options.strings.stringRegexpDeriveConflicts ? "true" : "false";
      case OptionEnum::STRINGS_RE_POSC_EAGER: return options.strings.stringRegexpPosConcatEager ? "true" : "false";
      case OptionEnum::STRINGS_REC_ARITH_APPROX: return options.strings.stringRecArithApprox ? "true" : "false";
      case OptionEnum::STRINGS_REGEXP_INCLUSION: return options.strings.stringRegexpInclusion ? "true" : "false";
      case OptionEnum::STRINGS_REXPLAIN_LEMMAS: return options.strings.stringRExplainLemmas ? "true" : "false";
      case OptionEnum::SUB_CBQI: return options.quantifiers.quantSubCbqi ? "true" : "false";
      case OptionEnum::SUB_CBQI_TIMEOUT: return std::to_string(options.quantifiers.quantSubCbqiTimeout);
      case OptionEnum::SYGUS: return options.quantifiers.sygus ? "true" : "false";
      case OptionEnum::SYGUS_ABORT_SIZE: return std::to_string(options.datatypes.sygusAbortSize);
      case OptionEnum::SYGUS_ADD_CONST_GRAMMAR: return options.quantifiers.sygusAddConstGrammar ? "true" : "false";
      case OptionEnum::SYGUS_ARG_RELEVANT: return options.quantifiers.sygusArgRelevant ? "true" : "false";
      case OptionEnum::SYGUS_AUTO_UNFOLD: return options.quantifiers.sygusInvAutoUnfold ? "true" : "false";
      case OptionEnum::SYGUS_BOOL_ITE_RETURN_CONST: return options.quantifiers.sygusBoolIteReturnConst ? "true" : "false";
      case OptionEnum::SYGUS_CORE_CONNECTIVE: return options.quantifiers.sygusCoreConnective ? "true" : "false";
      case OptionEnum::SYGUS_CREPAIR_ABORT: return options.quantifiers.sygusConstRepairAbort ? "true" : "false";
      case OptionEnum::SYGUS_ENUM: { std::stringstream s; s << options.quantifiers.sygusEnumMode; return s.str(); }
      case OptionEnum::SYGUS_ENUM_FAST_NUM_CONSTS: return std::to_string(options.quantifiers.sygusEnumFastNumConsts);
      case OptionEnum::SYGUS_ENUM_RANDOM_P: return std::to_string(options.quantifiers.sygusEnumRandomP);
      case OptionEnum::SYGUS_EVAL_UNFOLD: { std::stringstream s; s << options.quantifiers.sygusEvalUnfoldMode; return s.str(); }
      case OptionEnum::SYGUS_EXPR_MINER_CHECK_TIMEOUT: return std::to_string(options.quantifiers.sygusExprMinerCheckTimeout);
      case OptionEnum::SYGUS_FAIR: { std::stringstream s; s << options.datatypes.sygusFair; return s.str(); }
      case OptionEnum::SYGUS_FAIR_MAX: return options.datatypes.sygusFairMax ? "true" : "false";
      case OptionEnum::SYGUS_FILTER_SOL: { std::stringstream s; s << options.quantifiers.sygusFilterSolMode; return s.str(); }
      case OptionEnum::SYGUS_FILTER_SOL_REV: return options.quantifiers.sygusFilterSolRevSubsume ? "true" : "false";
      case OptionEnum::SYGUS_GRAMMAR_CONS: { std::stringstream s; s << options.quantifiers.sygusGrammarConsMode; return s.str(); }
      case OptionEnum::SYGUS_GRAMMAR_NORM: return options.quantifiers.sygusGrammarNorm ? "true" : "false";
      case OptionEnum::SYGUS_GRAMMAR_USE_DISJ: return options.quantifiers.sygusGrammarUseDisj ? "true" : "false";
      case OptionEnum::SYGUS_INFERENCE: { std::stringstream s; s << options.quantifiers.sygusInference; return s.str(); }
      case OptionEnum::SYGUS_INST: return options.quantifiers.sygusInst ? "true" : "false";
      case OptionEnum::SYGUS_INST_MODE: { std::stringstream s; s << options.quantifiers.sygusInstMode; return s.str(); }
      case OptionEnum::SYGUS_INST_SCOPE: { std::stringstream s; s << options.quantifiers.sygusInstScope; return s.str(); }
      case OptionEnum::SYGUS_INST_TERM_SEL: { std::stringstream s; s << options.quantifiers.sygusInstTermSel; return s.str(); }
      case OptionEnum::SYGUS_INV_TEMPL: { std::stringstream s; s << options.quantifiers.sygusInvTemplMode; return s.str(); }
      case OptionEnum::SYGUS_INV_TEMPL_WHEN_SG: return options.quantifiers.sygusInvTemplWhenSyntax ? "true" : "false";
      case OptionEnum::SYGUS_MIN_GRAMMAR: return options.quantifiers.sygusMinGrammar ? "true" : "false";
      case OptionEnum::SYGUS_OUT: { std::stringstream s; s << options.quantifiers.sygusOut; return s.str(); }
      case OptionEnum::SYGUS_PBE: return options.quantifiers.sygusUnifPbe ? "true" : "false";
      case OptionEnum::SYGUS_PBE_MULTI_FAIR: return options.quantifiers.sygusPbeMultiFair ? "true" : "false";
      case OptionEnum::SYGUS_PBE_MULTI_FAIR_DIFF: return std::to_string(options.quantifiers.sygusPbeMultiFairDiff);
      case OptionEnum::SYGUS_QE_PREPROC: return options.quantifiers.sygusQePreproc ? "true" : "false";
      case OptionEnum::SYGUS_QUERY_GEN: { std::stringstream s; s << options.quantifiers.sygusQueryGen; return s.str(); }
      case OptionEnum::SYGUS_QUERY_GEN_DUMP_FILES: { std::stringstream s; s << options.quantifiers.sygusQueryGenDumpFiles; return s.str(); }
      case OptionEnum::SYGUS_QUERY_GEN_FILTER_SOLVED: return options.quantifiers.sygusQueryFilterSolved ? "true" : "false";
      case OptionEnum::SYGUS_QUERY_GEN_THRESH: return std::to_string(options.quantifiers.sygusQueryGenThresh);
      case OptionEnum::SYGUS_REC_FUN: return options.quantifiers.sygusRecFun ? "true" : "false";
      case OptionEnum::SYGUS_REC_FUN_EVAL_LIMIT: return std::to_string(options.quantifiers.sygusRecFunEvalLimit);
      case OptionEnum::SYGUS_REPAIR_CONST: return options.quantifiers.sygusRepairConst ? "true" : "false";
      case OptionEnum::SYGUS_REPAIR_CONST_TIMEOUT: return std::to_string(options.quantifiers.sygusRepairConstTimeout);
      case OptionEnum::SYGUS_REWRITER: { std::stringstream s; s << options.datatypes.sygusRewriter; return s.str(); }
      case OptionEnum::SYGUS_RR_SYNTH_ACCEL: return options.quantifiers.sygusRewSynthAccel ? "true" : "false";
      case OptionEnum::SYGUS_RR_SYNTH_CHECK: return options.quantifiers.sygusRewSynthCheck ? "true" : "false";
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_CONG: return options.quantifiers.sygusRewSynthFilterCong ? "true" : "false";
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_MATCH: return options.quantifiers.sygusRewSynthFilterMatch ? "true" : "false";
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_NL: return options.quantifiers.sygusRewSynthFilterNonLinear ? "true" : "false";
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_ORDER: return options.quantifiers.sygusRewSynthFilterOrder ? "true" : "false";
      case OptionEnum::SYGUS_RR_SYNTH_INPUT_NVARS: return std::to_string(options.quantifiers.sygusRewSynthInputNVars);
      case OptionEnum::SYGUS_RR_SYNTH_REC: return options.quantifiers.sygusRewSynthRec ? "true" : "false";
      case OptionEnum::SYGUS_SAMPLE_FP_UNIFORM: return options.quantifiers.sygusSampleFpUniform ? "true" : "false";
      case OptionEnum::SYGUS_SAMPLE_GRAMMAR: return options.quantifiers.sygusSampleGrammar ? "true" : "false";
      case OptionEnum::SYGUS_SAMPLES: return std::to_string(options.quantifiers.sygusSamples);
      case OptionEnum::SYGUS_SI: { std::stringstream s; s << options.quantifiers.cegqiSingleInvMode; return s.str(); }
      case OptionEnum::SYGUS_SI_ABORT: return options.quantifiers.cegqiSingleInvAbort ? "true" : "false";
      case OptionEnum::SYGUS_SI_RCONS: { std::stringstream s; s << options.quantifiers.cegqiSingleInvReconstruct; return s.str(); }
      case OptionEnum::SYGUS_SI_RCONS_LIMIT: return std::to_string(options.quantifiers.cegqiSingleInvReconstructLimit);
      case OptionEnum::SYGUS_SI_RCONS_P: return std::to_string(options.quantifiers.cegqiSingleInvReconstructP);
      case OptionEnum::SYGUS_SIMPLE_SYM_BREAK: { std::stringstream s; s << options.datatypes.sygusSimpleSymBreak; return s.str(); }
      case OptionEnum::SYGUS_STREAM: return options.quantifiers.sygusStream ? "true" : "false";
      case OptionEnum::SYGUS_SYM_BREAK_LAZY: return options.datatypes.sygusSymBreakLazy ? "true" : "false";
      case OptionEnum::SYGUS_SYM_BREAK_PBE: return options.datatypes.sygusSymBreakPbe ? "true" : "false";
      case OptionEnum::SYGUS_SYM_BREAK_RLV: return options.datatypes.sygusSymBreakRlv ? "true" : "false";
      case OptionEnum::SYGUS_UNIF_COND_INDEPENDENT_NO_REPEAT_SOL: return options.quantifiers.sygusUnifCondIndNoRepeatSol ? "true" : "false";
      case OptionEnum::SYGUS_UNIF_PI: { std::stringstream s; s << options.quantifiers.sygusUnifPi; return s.str(); }
      case OptionEnum::SYGUS_UNIF_SHUFFLE_COND: return options.quantifiers.sygusUnifShuffleCond ? "true" : "false";
      case OptionEnum::SYGUS_VERIFY_INST_MAX_ROUNDS: return std::to_string(options.quantifiers.sygusVerifyInstMaxRounds);
      case OptionEnum::SYGUS_VERIFY_TIMEOUT: return std::to_string(options.quantifiers.sygusVerifyTimeout);
      case OptionEnum::SYMMETRY_BREAKER: return options.uf.ufSymmetryBreaker ? "true" : "false";
      case OptionEnum::TC_MODE: { std::stringstream s; s << options.theory.tcMode; return s.str(); }
      case OptionEnum::TERM_DB_MODE: { std::stringstream s; s << options.quantifiers.termDbMode; return s.str(); }
      case OptionEnum::TERM_SORT_OVERLOAD: return options.parser.termSortOverloading ? "true" : "false";
      case OptionEnum::THEORYOF_MODE: { std::stringstream s; s << options.theory.theoryOfMode; return s.str(); }
      case OptionEnum::TIMEOUT_CORE_TIMEOUT: return std::to_string(options.smt.toCoreTimeout);
      case OptionEnum::TLIMIT: return std::to_string(options.base.cumulativeMillisecondLimit);
      case OptionEnum::TLIMIT_PER: return std::to_string(options.base.perCallMillisecondLimit);
      case OptionEnum::TRIGGER_ACTIVE_SEL: { std::stringstream s; s << options.quantifiers.triggerActiveSelMode; return s.str(); }
      case OptionEnum::TRIGGER_SEL: { std::stringstream s; s << options.quantifiers.triggerSelMode; return s.str(); }
      case OptionEnum::TYPE_CHECKING: return options.expr.typeChecking ? "true" : "false";
      case OptionEnum::UF_CARD_EXP: return options.uf.ufCardExp ? "true" : "false";
      case OptionEnum::UF_HO_EXP: return options.uf.ufHoExp ? "true" : "false";
      case OptionEnum::UF_HO_EXT: return options.uf.ufHoExt ? "true" : "false";
      case OptionEnum::UF_LAMBDA_QE: return options.uf.ufHoLambdaQe ? "true" : "false";
      case OptionEnum::UF_LAZY_LL: return options.uf.ufHoLazyLambdaLift ? "true" : "false";
      case OptionEnum::UF_SS: { std::stringstream s; s << options.uf.ufssMode; return s.str(); }
      case OptionEnum::UF_SS_ABORT_CARD: return std::to_string(options.uf.ufssAbortCardinality);
      case OptionEnum::UF_SS_FAIR: return options.uf.ufssFairness ? "true" : "false";
      case OptionEnum::UF_SS_FAIR_MONOTONE: return options.uf.ufssFairnessMonotone ? "true" : "false";
      case OptionEnum::UNATE_LEMMAS: { std::stringstream s; s << options.arith.arithUnateLemmaMode; return s.str(); }
      case OptionEnum::UNCONSTRAINED_SIMP: return options.smt.unconstrainedSimp ? "true" : "false";
      case OptionEnum::UNSAT_CORES_MODE: { std::stringstream s; s << options.smt.unsatCoresMode; return s.str(); }
      case OptionEnum::USE_APPROX: return options.arith.useApprox ? "true" : "false";
      case OptionEnum::USE_FCSIMPLEX: return options.arith.useFC ? "true" : "false";
      case OptionEnum::USE_PORTFOLIO: return options.driver.usePortfolio ? "true" : "false";
      case OptionEnum::USE_SOI: return options.arith.useSOI ? "true" : "false";
      case OptionEnum::USER_PAT: { std::stringstream s; s << options.quantifiers.userPatternsQuant; return s.str(); }
      case OptionEnum::USER_POOL: { std::stringstream s; s << options.quantifiers.userPoolQuant; return s.str(); }
      case OptionEnum::VAR_ELIM_QUANT: return options.quantifiers.varElimQuant ? "true" : "false";
      case OptionEnum::VAR_INEQ_ELIM_QUANT: return options.quantifiers.varIneqElimQuant ? "true" : "false";
      case OptionEnum::VERBOSITY: return std::to_string(options.base.verbosity);
      case OptionEnum::VERSION: return options.driver.showVersion ? "true" : "false";
      case OptionEnum::WF_CHECKING: return options.expr.wellFormedChecking ? "true" : "false";
      case OptionEnum::WRITE_PARTITIONS_TO: { std::stringstream s; s << options.parallel.partitionsOut; return s.str(); }
      default:
      {
        throw OptionException("Ungettable option key or setting: " + name);
      }
    }
    // clang-format on
    throw OptionException("Unrecognized option key or setting: " + name);
  }

  void set(
      Options & opts, const std::string& name, const std::string& optionarg)
  {
    Trace("options") << "set option " << name << " = " << optionarg
                     << std::endl;
    // clang-format off
    auto it = NAME_TO_ENUM.find(name);
    if (it == NAME_TO_ENUM.end()) {
      throw OptionException("Unrecognized option key or setting: " + name);
    }
    switch (it->second) {
      case OptionEnum::ABSTRACT_VALUES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().abstractValues = value;
        opts.write_smt().abstractValuesWasSetByUser = true;
        break;
      }
      case OptionEnum::ACKERMANN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().ackermann = value;
        opts.write_smt().ackermannWasSetByUser = true;
        break;
      }
      case OptionEnum::APPEND_LEARNED_LITERALS_TO_CUBES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parallel().appendLearnedLiteralsToCubes = value;
        opts.write_parallel().appendLearnedLiteralsToCubesWasSetByUser = true;
        break;
      }
      case OptionEnum::APPROX_BRANCH_DEPTH:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().maxApproxDepth = value;
        opts.write_arith().maxApproxDepthWasSetByUser = true;
        break;
      }
      case OptionEnum::ARITH_BRAB:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().brabTest = value;
        opts.write_arith().brabTestWasSetByUser = true;
        break;
      }
      case OptionEnum::ARITH_EQ_SOLVER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().arithEqSolver = value;
        opts.write_arith().arithEqSolverWasSetByUser = true;
        break;
      }
      case OptionEnum::ARITH_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().arithExp = value;
        opts.write_arith().arithExpWasSetByUser = true;
        break;
      }
      case OptionEnum::ARITH_PROP:
      {
        auto value = stringToArithPropagationMode(optionarg);
        opts.write_arith().arithPropagationMode = value;
        opts.write_arith().arithPropagationModeWasSetByUser = true;
        break;
      }
      case OptionEnum::ARITH_PROP_CLAUSES:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().arithPropAsLemmaLength = value;
        opts.write_arith().arithPropAsLemmaLengthWasSetByUser = true;
        break;
      }
      case OptionEnum::ARITH_REWRITE_EQUALITIES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().arithRewriteEq = value;
        opts.write_arith().arithRewriteEqWasSetByUser = true;
        break;
      }
      case OptionEnum::ARITH_STATIC_LEARNING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().arithStaticLearning = value;
        opts.write_arith().arithStaticLearningWasSetByUser = true;
        break;
      }
      case OptionEnum::ARRAYS_EAGER_INDEX:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arrays().arraysEagerIndexSplitting = value;
        opts.write_arrays().arraysEagerIndexSplittingWasSetByUser = true;
        break;
      }
      case OptionEnum::ARRAYS_EAGER_LEMMAS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arrays().arraysEagerLemmas = value;
        opts.write_arrays().arraysEagerLemmasWasSetByUser = true;
        break;
      }
      case OptionEnum::ARRAYS_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arrays().arraysExp = value;
        opts.write_arrays().arraysExpWasSetByUser = true;
        break;
      }
      case OptionEnum::ARRAYS_OPTIMIZE_LINEAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arrays().arraysOptimizeLinear = value;
        opts.write_arrays().arraysOptimizeLinearWasSetByUser = true;
        break;
      }
      case OptionEnum::ARRAYS_PROP:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arrays().arraysPropagate = value;
        opts.write_arrays().arraysPropagateWasSetByUser = true;
        break;
      }
      case OptionEnum::ARRAYS_REDUCE_SHARING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arrays().arraysReduceSharing = value;
        opts.write_arrays().arraysReduceSharingWasSetByUser = true;
        break;
      }
      case OptionEnum::ARRAYS_WEAK_EQUIV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arrays().arraysWeakEquivalence = value;
        opts.write_arrays().arraysWeakEquivalenceWasSetByUser = true;
        break;
      }
      case OptionEnum::ASSIGN_FUNCTION_VALUES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_theory().assignFunctionValues = value;
        opts.write_theory().assignFunctionValuesWasSetByUser = true;
        break;
      }
      case OptionEnum::BAGS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bags().bags = value;
        opts.write_bags().bagsWasSetByUser = true;
        break;
      }
      case OptionEnum::BITBLAST:
      {
        auto value = stringToBitblastMode(optionarg);
        opts.write_bv().bitblastMode = value;
        opts.write_bv().bitblastModeWasSetByUser = true;
        break;
      }
      case OptionEnum::BITWISE_EQ:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bitwiseEq = value;
        opts.write_bv().bitwiseEqWasSetByUser = true;
        break;
      }
      case OptionEnum::BOOL_TO_BV:
      {
        auto value = stringToBoolToBVMode(optionarg);
        opts.write_bv().boolToBitvector = value;
        opts.write_bv().boolToBitvectorWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_ASSERT_INPUT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bvAssertInput = value;
        opts.write_bv().bvAssertInputWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_EAGER_EVAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bvEagerEval = value;
        opts.write_bv().bvEagerEvalWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_EQ_ENGINE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bvEqEngine = value;
        opts.write_bv().bvEqEngineWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_GAUSS_ELIM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bvGaussElim = value;
        opts.write_bv().bvGaussElimWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_INTRO_POW2:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bvIntroducePow2 = value;
        opts.write_bv().bvIntroducePow2WasSetByUser = true;
        break;
      }
      case OptionEnum::BV_PRINT_CONSTS_AS_INDEXED_SYMBOLS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        ioutils::setDefaultBvPrintConstsAsIndexedSymbols(value);
        opts.write_printer().bvPrintConstsAsIndexedSymbols = value;
        opts.write_printer().bvPrintConstsAsIndexedSymbolsWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_PROPAGATE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bitvectorPropagate = value;
        opts.write_bv().bitvectorPropagateWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_RW_EXTEND_EQ:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().rwExtendEq = value;
        opts.write_bv().rwExtendEqWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_SAT_SOLVER:
      {
        auto value = stringToBvSatSolverMode(optionarg);
        opts.handler().checkBvSatSolver(name, value);
        opts.write_bv().bvSatSolver = value;
        opts.write_bv().bvSatSolverWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_SOLVER:
      {
        auto value = stringToBVSolver(optionarg);
        opts.write_bv().bvSolver = value;
        opts.write_bv().bvSolverWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_TO_BOOL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_bv().bitvectorToBool = value;
        opts.write_bv().bitvectorToBoolWasSetByUser = true;
        break;
      }
      case OptionEnum::BV_TO_INT_USE_POW2:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().bvToIntUsePow2 = value;
        opts.write_smt().bvToIntUsePow2WasSetByUser = true;
        break;
      }
      case OptionEnum::BVAND_INTEGER_GRANULARITY:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<uint64_t>(1));
        opts.handler().checkMaximum(name, value, static_cast<uint64_t>(8));
        opts.write_smt().BVAndIntegerGranularity = value;
        opts.write_smt().BVAndIntegerGranularityWasSetByUser = true;
        break;
      }
      case OptionEnum::CBQI:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().conflictBasedInst = value;
        opts.write_quantifiers().conflictBasedInstWasSetByUser = true;
        break;
      }
      case OptionEnum::CBQI_ALL_CONFLICT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cbqiAllConflict = value;
        opts.write_quantifiers().cbqiAllConflictWasSetByUser = true;
        break;
      }
      case OptionEnum::CBQI_MODE:
      {
        auto value = stringToQcfMode(optionarg);
        opts.write_quantifiers().cbqiMode = value;
        opts.write_quantifiers().cbqiModeWasSetByUser = true;
        break;
      }
      case OptionEnum::CBQI_SKIP_RD:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cbqiSkipRd = value;
        opts.write_quantifiers().cbqiSkipRdWasSetByUser = true;
        break;
      }
      case OptionEnum::CBQI_TCONSTRAINT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cbqiTConstraint = value;
        opts.write_quantifiers().cbqiTConstraintWasSetByUser = true;
        break;
      }
      case OptionEnum::CBQI_VO_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cbqiVoExp = value;
        opts.write_quantifiers().cbqiVoExpWasSetByUser = true;
        break;
      }
      case OptionEnum::CDT_BISIMILAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().cdtBisimilar = value;
        opts.write_datatypes().cdtBisimilarWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGIS_SAMPLE:
      {
        auto value = stringToCegisSampleMode(optionarg);
        opts.write_quantifiers().cegisSample = value;
        opts.write_quantifiers().cegisSampleWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqi = value;
        opts.write_quantifiers().cegqiWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_ALL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiAll = value;
        opts.write_quantifiers().cegqiAllWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_BV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiBv = value;
        opts.write_quantifiers().cegqiBvWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_BV_CONCAT_INV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiBvConcInv = value;
        opts.write_quantifiers().cegqiBvConcInvWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_BV_INEQ:
      {
        auto value = stringToCegqiBvIneqMode(optionarg);
        opts.write_quantifiers().cegqiBvIneqMode = value;
        opts.write_quantifiers().cegqiBvIneqModeWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_BV_INTERLEAVE_VALUE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiBvInterleaveValue = value;
        opts.write_quantifiers().cegqiBvInterleaveValueWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_BV_LINEAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiBvLinearize = value;
        opts.write_quantifiers().cegqiBvLinearizeWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_BV_RM_EXTRACT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiBvRmExtract = value;
        opts.write_quantifiers().cegqiBvRmExtractWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_BV_SOLVE_NL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiBvSolveNl = value;
        opts.write_quantifiers().cegqiBvSolveNlWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_FULL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiFullEffort = value;
        opts.write_quantifiers().cegqiFullEffortWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_INF_INT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiUseInfInt = value;
        opts.write_quantifiers().cegqiUseInfIntWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_INF_REAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiUseInfReal = value;
        opts.write_quantifiers().cegqiUseInfRealWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_INNERMOST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiInnermost = value;
        opts.write_quantifiers().cegqiInnermostWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_MIDPOINT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiMidpoint = value;
        opts.write_quantifiers().cegqiMidpointWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_MIN_BOUNDS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiMinBounds = value;
        opts.write_quantifiers().cegqiMinBoundsWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_MULTI_INST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiMultiInst = value;
        opts.write_quantifiers().cegqiMultiInstWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_NESTED_QE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiNestedQE = value;
        opts.write_quantifiers().cegqiNestedQEWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_NOPT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiNopt = value;
        opts.write_quantifiers().cegqiNoptWasSetByUser = true;
        break;
      }
      case OptionEnum::CEGQI_ROUND_UP_LIA:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiRoundUpLowerLia = value;
        opts.write_quantifiers().cegqiRoundUpLowerLiaWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECK_ABDUCTS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().checkAbducts = value;
        opts.write_smt().checkAbductsWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECK_INTERPOLANTS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().checkInterpolants = value;
        opts.write_smt().checkInterpolantsWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECK_MODELS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().checkModels = value;
        opts.write_smt().checkModelsWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECK_PROOF_STEPS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().checkProofSteps = value;
        opts.write_proof().checkProofStepsWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECK_PROOFS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().checkProofs = value;
        opts.write_smt().checkProofsWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECK_SYNTH_SOL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().checkSynthSol = value;
        opts.write_smt().checkSynthSolWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECK_UNSAT_CORES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().checkUnsatCores = value;
        opts.write_smt().checkUnsatCoresWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECKS_BEFORE_PARTITION:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_parallel().checksBeforePartitioning = value;
        opts.write_parallel().checksBeforePartitioningWasSetByUser = true;
        break;
      }
      case OptionEnum::CHECKS_BETWEEN_PARTITIONS:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_parallel().checksBetweenPartitions = value;
        opts.write_parallel().checksBetweenPartitionsWasSetByUser = true;
        break;
      }
      case OptionEnum::COLLECT_PIVOT_STATS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().collectPivots = value;
        opts.write_arith().collectPivotsWasSetByUser = true;
        break;
      }
      case OptionEnum::COMPUTE_PARTITIONS:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_parallel().computePartitions = value;
        opts.write_parallel().computePartitionsWasSetByUser = true;
        break;
      }
      case OptionEnum::COND_VAR_SPLIT_QUANT:
      {
        auto value = stringToCondVarSplitQuantMode(optionarg);
        opts.write_quantifiers().condVarSplitQuant = value;
        opts.write_quantifiers().condVarSplitQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::CONDENSE_FUNCTION_VALUES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_theory().condenseFunctionValues = value;
        opts.write_theory().condenseFunctionValuesWasSetByUser = true;
        break;
      }
      case OptionEnum::CONFLICT_PROCESS:
      {
        auto value = stringToConflictProcessMode(optionarg);
        opts.write_theory().conflictProcessMode = value;
        opts.write_theory().conflictProcessModeWasSetByUser = true;
        break;
      }
      case OptionEnum::CONJECTURE_GEN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().conjectureGen = value;
        opts.write_quantifiers().conjectureGenWasSetByUser = true;
        break;
      }
      case OptionEnum::CONJECTURE_GEN_GT_ENUM:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().conjectureGenGtEnum = value;
        opts.write_quantifiers().conjectureGenGtEnumWasSetByUser = true;
        break;
      }
      case OptionEnum::CONJECTURE_GEN_MAX_DEPTH:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().conjectureGenMaxDepth = value;
        opts.write_quantifiers().conjectureGenMaxDepthWasSetByUser = true;
        break;
      }
      case OptionEnum::CONJECTURE_GEN_PER_ROUND:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().conjectureGenPerRound = value;
        opts.write_quantifiers().conjectureGenPerRoundWasSetByUser = true;
        break;
      }
      case OptionEnum::CONS_EXP_TRIGGERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().consExpandTriggers = value;
        opts.write_quantifiers().consExpandTriggersWasSetByUser = true;
        break;
      }
      case OptionEnum::COPYRIGHT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().showCopyright(name, value);
        opts.write_driver().showCopyright = value;
        opts.write_driver().showCopyrightWasSetByUser = true;
        break;
      }
      case OptionEnum::CUT_ALL_BOUNDED:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().doCutAllBounded = value;
        opts.write_arith().doCutAllBoundedWasSetByUser = true;
        break;
      }
      case OptionEnum::DAG_THRESH:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<int64_t>(0));
        ioutils::setDefaultDagThresh(value);
        opts.write_printer().dagThresh = value;
        opts.write_printer().dagThreshWasSetByUser = true;
        break;
      }
      case OptionEnum::DATATYPES_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().datatypesExp = value;
        opts.write_datatypes().datatypesExpWasSetByUser = true;
        break;
      }
      case OptionEnum::DEBUG_CHECK_MODELS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().debugCheckModels = value;
        opts.write_smt().debugCheckModelsWasSetByUser = true;
        break;
      }
      case OptionEnum::DECISION:
      {
        auto value = stringToDecisionMode(optionarg);
        opts.write_decision().decisionMode = value;
        opts.write_decision().decisionModeWasSetByUser = true;
        break;
      }
      case OptionEnum::DEEP_RESTART:
      {
        auto value = stringToDeepRestartMode(optionarg);
        opts.write_smt().deepRestartMode = value;
        opts.write_smt().deepRestartModeWasSetByUser = true;
        break;
      }
      case OptionEnum::DEEP_RESTART_FACTOR:
      {
        auto value = handlers::handleOption<double>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<double>(0.0));
        opts.handler().checkMaximum(name, value, static_cast<double>(1000.0));
        opts.write_smt().deepRestartFactor = value;
        opts.write_smt().deepRestartFactorWasSetByUser = true;
        break;
      }
      case OptionEnum::DEFAULT_FUNCTION_VALUE_MODE:
      {
        auto value = stringToDefaultFunctionValueMode(optionarg);
        opts.write_theory().defaultFunctionValueMode = value;
        opts.write_theory().defaultFunctionValueModeWasSetByUser = true;
        break;
      }
      case OptionEnum::DIFFICULTY_MODE:
      {
        auto value = stringToDifficultyMode(optionarg);
        opts.write_smt().difficultyMode = value;
        opts.write_smt().difficultyModeWasSetByUser = true;
        break;
      }
      case OptionEnum::DIO_DECOMPS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().exportDioDecompositions = value;
        opts.write_arith().exportDioDecompositionsWasSetByUser = true;
        break;
      }
      case OptionEnum::DIO_SOLVER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().arithDioSolver = value;
        opts.write_arith().arithDioSolverWasSetByUser = true;
        break;
      }
      case OptionEnum::DIO_TURNS:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().dioSolverTurns = value;
        opts.write_arith().dioSolverTurnsWasSetByUser = true;
        break;
      }
      case OptionEnum::DRAT_BINARY_FORMAT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().dratBinaryFormat = value;
        opts.write_proof().dratBinaryFormatWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_BINARY_SPLIT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().dtBinarySplit = value;
        opts.write_datatypes().dtBinarySplitWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_BLAST_SPLITS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().dtBlastSplits = value;
        opts.write_datatypes().dtBlastSplitsWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_CYCLIC:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().dtCyclic = value;
        opts.write_datatypes().dtCyclicWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_INFER_AS_LEMMAS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().dtInferAsLemmas = value;
        opts.write_datatypes().dtInferAsLemmasWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_NESTED_REC:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().dtNestedRec = value;
        opts.write_datatypes().dtNestedRecWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_POLITE_OPTIMIZE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().dtPoliteOptimize = value;
        opts.write_datatypes().dtPoliteOptimizeWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_SHARE_SEL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().dtSharedSelectors = value;
        opts.write_datatypes().dtSharedSelectorsWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_STC_IND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().dtStcInduction = value;
        opts.write_quantifiers().dtStcInductionWasSetByUser = true;
        break;
      }
      case OptionEnum::DT_VAR_EXP_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().dtVarExpandQuant = value;
        opts.write_quantifiers().dtVarExpandQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::DUMP_DIFFICULTY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().dumpDifficulty = value;
        opts.write_driver().dumpDifficultyWasSetByUser = true;
        break;
      }
      case OptionEnum::DUMP_INSTANTIATIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().dumpInstantiations = value;
        opts.write_driver().dumpInstantiationsWasSetByUser = true;
        break;
      }
      case OptionEnum::DUMP_INSTANTIATIONS_DEBUG:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().dumpInstantiationsDebug = value;
        opts.write_driver().dumpInstantiationsDebugWasSetByUser = true;
        break;
      }
      case OptionEnum::DUMP_MODELS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().dumpModels = value;
        opts.write_driver().dumpModelsWasSetByUser = true;
        break;
      }
      case OptionEnum::DUMP_PROOFS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().dumpProofs = value;
        opts.write_driver().dumpProofsWasSetByUser = true;
        break;
      }
      case OptionEnum::DUMP_UNSAT_CORES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().dumpUnsatCores = value;
        opts.write_driver().dumpUnsatCoresWasSetByUser = true;
        break;
      }
      case OptionEnum::DUMP_UNSAT_CORES_LEMMAS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().dumpUnsatCoresLemmas = value;
        opts.write_driver().dumpUnsatCoresLemmasWasSetByUser = true;
        break;
      }
      case OptionEnum::E_MATCHING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().eMatching = value;
        opts.write_quantifiers().eMatchingWasSetByUser = true;
        break;
      }
      case OptionEnum::EAGER_ARITH_BV_CONV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().eagerArithBvConv = value;
        opts.write_uf().eagerArithBvConvWasSetByUser = true;
        break;
      }
      case OptionEnum::EARLY_EXIT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().earlyExit = value;
        opts.write_driver().earlyExitWasSetByUser = true;
        break;
      }
      case OptionEnum::EARLY_ITE_REMOVAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().earlyIteRemoval = value;
        opts.write_smt().earlyIteRemovalWasSetByUser = true;
        break;
      }
      case OptionEnum::EE_MODE:
      {
        auto value = stringToEqEngineMode(optionarg);
        opts.write_theory().eeMode = value;
        opts.write_theory().eeModeWasSetByUser = true;
        break;
      }
      case OptionEnum::ELIM_TAUT_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().elimTautQuant = value;
        opts.write_quantifiers().elimTautQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::ENUM_INST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().enumInst = value;
        opts.write_quantifiers().enumInstWasSetByUser = true;
        break;
      }
      case OptionEnum::ENUM_INST_INTERLEAVE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().enumInstInterleave = value;
        opts.write_quantifiers().enumInstInterleaveWasSetByUser = true;
        break;
      }
      case OptionEnum::ENUM_INST_LIMIT:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().enumInstLimit = value;
        opts.write_quantifiers().enumInstLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::ENUM_INST_RD:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().enumInstRd = value;
        opts.write_quantifiers().enumInstRdWasSetByUser = true;
        break;
      }
      case OptionEnum::ENUM_INST_STRATIFY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().enumInstStratify = value;
        opts.write_quantifiers().enumInstStratifyWasSetByUser = true;
        break;
      }
      case OptionEnum::ENUM_INST_SUM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().enumInstSum = value;
        opts.write_quantifiers().enumInstSumWasSetByUser = true;
        break;
      }
      case OptionEnum::ERR:
      {
        auto value = handlers::handleOption<ManagedErr>(name, optionarg);
        opts.handler().setErrStream(name, value);
        opts.write_base().err = value;
        opts.write_base().errWasSetByUser = true;
        break;
      }
      case OptionEnum::ERROR_SELECTION_RULE:
      {
        auto value = stringToErrorSelectionRule(optionarg);
        opts.write_arith().arithErrorSelectionRule = value;
        opts.write_arith().arithErrorSelectionRuleWasSetByUser = true;
        break;
      }
      case OptionEnum::EXPR_DEPTH:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<int64_t>(-1));
        ioutils::setDefaultNodeDepth(value);
        opts.write_printer().nodeDepth = value;
        opts.write_printer().nodeDepthWasSetByUser = true;
        break;
      }
      case OptionEnum::EXT_REW_PREP:
      {
        auto value = stringToExtRewPrepMode(optionarg);
        opts.write_smt().extRewPrep = value;
        opts.write_smt().extRewPrepWasSetByUser = true;
        break;
      }
      case OptionEnum::EXT_REWRITE_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().extRewriteQuant = value;
        opts.write_quantifiers().extRewriteQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::FC_PENALTIES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().havePenalties = value;
        opts.write_arith().havePenaltiesWasSetByUser = true;
        break;
      }
      case OptionEnum::FF:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_ff().ff = value;
        opts.write_ff().ffWasSetByUser = true;
        break;
      }
      case OptionEnum::FF_BITSUM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_ff().ffBitsum = value;
        opts.write_ff().ffBitsumWasSetByUser = true;
        break;
      }
      case OptionEnum::FF_ELIM_DISJUNCTIVE_BIT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_ff().ffElimDisjunctiveBit = value;
        opts.write_ff().ffElimDisjunctiveBitWasSetByUser = true;
        break;
      }
      case OptionEnum::FF_FIELD_POLYS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_ff().ffFieldPolys = value;
        opts.write_ff().ffFieldPolysWasSetByUser = true;
        break;
      }
      case OptionEnum::FF_SOLVER:
      {
        auto value = stringToFfSolver(optionarg);
        opts.write_ff().ffSolver = value;
        opts.write_ff().ffSolverWasSetByUser = true;
        break;
      }
      case OptionEnum::FF_TRACE_GB:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_ff().ffTraceGb = value;
        opts.write_ff().ffTraceGbWasSetByUser = true;
        break;
      }
      case OptionEnum::FILENAME:
      {
        auto value = handlers::handleOption<std::string>(name, optionarg);
        opts.write_driver().filename = value;
        opts.write_driver().filenameWasSetByUser = true;
        break;
      }
      case OptionEnum::FILESYSTEM_ACCESS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parser().filesystemAccess = value;
        opts.write_parser().filesystemAccessWasSetByUser = true;
        break;
      }
      case OptionEnum::FINITE_MODEL_FIND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().finiteModelFind = value;
        opts.write_quantifiers().finiteModelFindWasSetByUser = true;
        break;
      }
      case OptionEnum::FLATTEN_HO_CHAINS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        ioutils::setDefaultFlattenHOChains(value);
        opts.write_printer().flattenHOChains = value;
        opts.write_printer().flattenHOChainsWasSetByUser = true;
        break;
      }
      case OptionEnum::FMF_BOUND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fmfBound = value;
        opts.write_quantifiers().fmfBoundWasSetByUser = true;
        break;
      }
      case OptionEnum::FMF_BOUND_BLAST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fmfBoundBlast = value;
        opts.write_quantifiers().fmfBoundBlastWasSetByUser = true;
        break;
      }
      case OptionEnum::FMF_BOUND_LAZY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fmfBoundLazy = value;
        opts.write_quantifiers().fmfBoundLazyWasSetByUser = true;
        break;
      }
      case OptionEnum::FMF_FUN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fmfFunWellDefined = value;
        opts.write_quantifiers().fmfFunWellDefinedWasSetByUser = true;
        break;
      }
      case OptionEnum::FMF_FUN_RLV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fmfFunWellDefinedRelevant = value;
        opts.write_quantifiers().fmfFunWellDefinedRelevantWasSetByUser = true;
        break;
      }
      case OptionEnum::FMF_MBQI:
      {
        auto value = stringToFmfMbqiMode(optionarg);
        opts.write_quantifiers().fmfMbqiMode = value;
        opts.write_quantifiers().fmfMbqiModeWasSetByUser = true;
        break;
      }
      case OptionEnum::FMF_TYPE_COMPLETION_THRESH:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().fmfTypeCompletionThresh = value;
        opts.write_quantifiers().fmfTypeCompletionThreshWasSetByUser = true;
        break;
      }
      case OptionEnum::FORCE_LOGIC:
      {
        auto value = handlers::handleOption<std::string>(name, optionarg);
        opts.write_parser().forceLogicString = value;
        opts.write_parser().forceLogicStringWasSetByUser = true;
        break;
      }
      case OptionEnum::FORCE_NO_LIMIT_CPU_WHILE_DUMP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().forceNoLimitCpuWhileDump = value;
        opts.write_driver().forceNoLimitCpuWhileDumpWasSetByUser = true;
        break;
      }
      case OptionEnum::FOREIGN_THEORY_REWRITE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().foreignTheoryRewrite = value;
        opts.write_smt().foreignTheoryRewriteWasSetByUser = true;
        break;
      }
      case OptionEnum::FP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_fp().fp = value;
        opts.write_fp().fpWasSetByUser = true;
        break;
      }
      case OptionEnum::FP_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_fp().fpExp = value;
        opts.write_fp().fpExpWasSetByUser = true;
        break;
      }
      case OptionEnum::FP_LAZY_WB:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_fp().fpLazyWb = value;
        opts.write_fp().fpLazyWbWasSetByUser = true;
        break;
      }
      case OptionEnum::FRESH_BINDERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parser().freshBinders = value;
        opts.write_parser().freshBindersWasSetByUser = true;
        break;
      }
      case OptionEnum::FRESH_DECLARATIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parser().freshDeclarations = value;
        opts.write_parser().freshDeclarationsWasSetByUser = true;
        break;
      }
      case OptionEnum::FULL_SATURATE_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fullSaturateQuant = value;
        opts.write_quantifiers().fullSaturateQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::FULL_SYGUS_VERIFY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fullSygusVerify = value;
        opts.write_quantifiers().fullSygusVerifyWasSetByUser = true;
        break;
      }
      case OptionEnum::GLOBAL_DECLARATIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parser().globalDeclarations = value;
        opts.write_parser().globalDeclarationsWasSetByUser = true;
        break;
      }
      case OptionEnum::GLOBAL_NEGATE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().globalNegate = value;
        opts.write_quantifiers().globalNegateWasSetByUser = true;
        break;
      }
      case OptionEnum::HELP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().help = value;
        opts.write_driver().helpWasSetByUser = true;
        break;
      }
      case OptionEnum::HELP_REGULAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().helpRegular = value;
        opts.write_driver().helpRegularWasSetByUser = true;
        break;
      }
      case OptionEnum::HEURISTIC_PIVOTS:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().arithHeuristicPivots = value;
        opts.write_arith().arithHeuristicPivotsWasSetByUser = true;
        break;
      }
      case OptionEnum::HO_ELIM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().hoElim = value;
        opts.write_quantifiers().hoElimWasSetByUser = true;
        break;
      }
      case OptionEnum::HO_ELIM_STORE_AX:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().hoElimStoreAx = value;
        opts.write_quantifiers().hoElimStoreAxWasSetByUser = true;
        break;
      }
      case OptionEnum::HO_MATCHING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().hoMatching = value;
        opts.write_quantifiers().hoMatchingWasSetByUser = true;
        break;
      }
      case OptionEnum::HO_MERGE_TERM_DB:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().hoMergeTermDb = value;
        opts.write_quantifiers().hoMergeTermDbWasSetByUser = true;
        break;
      }
      case OptionEnum::IAND_MODE:
      {
        auto value = stringToIandMode(optionarg);
        opts.write_smt().iandMode = value;
        opts.write_smt().iandModeWasSetByUser = true;
        break;
      }
      case OptionEnum::IEVAL:
      {
        auto value = stringToIevalMode(optionarg);
        opts.write_quantifiers().ievalMode = value;
        opts.write_quantifiers().ievalModeWasSetByUser = true;
        break;
      }
      case OptionEnum::IN:
      {
        auto value = handlers::handleOption<ManagedIn>(name, optionarg);
        opts.write_base().in = value;
        opts.write_base().inWasSetByUser = true;
        break;
      }
      case OptionEnum::INCREMENT_TRIGGERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().incrementTriggers = value;
        opts.write_quantifiers().incrementTriggersWasSetByUser = true;
        break;
      }
      case OptionEnum::INCREMENTAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_base().incrementalSolving = value;
        opts.write_base().incrementalSolvingWasSetByUser = true;
        break;
      }
      case OptionEnum::INST_MAX_LEVEL:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().instMaxLevel = value;
        opts.write_quantifiers().instMaxLevelWasSetByUser = true;
        break;
      }
      case OptionEnum::INST_MAX_ROUNDS:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().instMaxRounds = value;
        opts.write_quantifiers().instMaxRoundsWasSetByUser = true;
        break;
      }
      case OptionEnum::INST_NO_ENTAIL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().instNoEntail = value;
        opts.write_quantifiers().instNoEntailWasSetByUser = true;
        break;
      }
      case OptionEnum::INST_WHEN:
      {
        auto value = stringToInstWhenMode(optionarg);
        opts.write_quantifiers().instWhenMode = value;
        opts.write_quantifiers().instWhenModeWasSetByUser = true;
        break;
      }
      case OptionEnum::INST_WHEN_PHASE:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().instWhenPhase = value;
        opts.write_quantifiers().instWhenPhaseWasSetByUser = true;
        break;
      }
      case OptionEnum::INT_WF_IND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().intWfInduction = value;
        opts.write_quantifiers().intWfInductionWasSetByUser = true;
        break;
      }
      case OptionEnum::INTERACTIVE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().interactive = value;
        opts.write_driver().interactiveWasSetByUser = true;
        break;
      }
      case OptionEnum::INTERPOLANTS_MODE:
      {
        auto value = stringToInterpolantsMode(optionarg);
        opts.write_smt().interpolantsMode = value;
        opts.write_smt().interpolantsModeWasSetByUser = true;
        break;
      }
      case OptionEnum::ITE_DTT_SPLIT_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().iteDtTesterSplitQuant = value;
        opts.write_quantifiers().iteDtTesterSplitQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::ITE_LIFT_QUANT:
      {
        auto value = stringToIteLiftQuantMode(optionarg);
        opts.write_quantifiers().iteLiftQuant = value;
        opts.write_quantifiers().iteLiftQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::ITE_SIMP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().doITESimp = value;
        opts.write_smt().doITESimpWasSetByUser = true;
        break;
      }
      case OptionEnum::JH_RLV_ORDER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_decision().jhRlvOrder = value;
        opts.write_decision().jhRlvOrderWasSetByUser = true;
        break;
      }
      case OptionEnum::JH_SKOLEM:
      {
        auto value = stringToJutificationSkolemMode(optionarg);
        opts.write_decision().jhSkolemMode = value;
        opts.write_decision().jhSkolemModeWasSetByUser = true;
        break;
      }
      case OptionEnum::JH_SKOLEM_RLV:
      {
        auto value = stringToJutificationSkolemRlvMode(optionarg);
        opts.write_decision().jhSkolemRlvMode = value;
        opts.write_decision().jhSkolemRlvModeWasSetByUser = true;
        break;
      }
      case OptionEnum::LANG:
      {
        auto value = opts.handler().stringToLanguage(name, optionarg);
        opts.handler().setInputLanguage(name, value);
        opts.write_base().inputLanguage = value;
        opts.write_base().inputLanguageWasSetByUser = true;
        break;
      }
      case OptionEnum::LEARNED_REWRITE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().learnedRewrite = value;
        opts.write_smt().learnedRewriteWasSetByUser = true;
        break;
      }
      case OptionEnum::LEMMA_INPROCESS:
      {
        auto value = stringToLemmaInprocessMode(optionarg);
        opts.write_theory().lemmaInprocess = value;
        opts.write_theory().lemmaInprocessWasSetByUser = true;
        break;
      }
      case OptionEnum::LEMMA_INPROCESS_INFER_EQ_LIT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_theory().lemmaInprocessInferEqLit = value;
        opts.write_theory().lemmaInprocessInferEqLitWasSetByUser = true;
        break;
      }
      case OptionEnum::LEMMA_INPROCESS_SUBS:
      {
        auto value = stringToLemmaInprocessSubsMode(optionarg);
        opts.write_theory().lemmaInprocessSubsMode = value;
        opts.write_theory().lemmaInprocessSubsModeWasSetByUser = true;
        break;
      }
      case OptionEnum::LEMMAS_ON_REPLAY_FAILURE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().replayFailureLemma = value;
        opts.write_arith().replayFailureLemmaWasSetByUser = true;
        break;
      }
      case OptionEnum::LFSC_EXPAND_TRUST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().lfscExpandTrust = value;
        opts.write_proof().lfscExpandTrustWasSetByUser = true;
        break;
      }
      case OptionEnum::LFSC_FLATTEN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().lfscFlatten = value;
        opts.write_proof().lfscFlattenWasSetByUser = true;
        break;
      }
      case OptionEnum::LITERAL_MATCHING:
      {
        auto value = stringToLiteralMatchMode(optionarg);
        opts.write_quantifiers().literalMatchMode = value;
        opts.write_quantifiers().literalMatchModeWasSetByUser = true;
        break;
      }
      case OptionEnum::MACROS_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().macrosQuant = value;
        opts.write_quantifiers().macrosQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::MACROS_QUANT_MODE:
      {
        auto value = stringToMacrosQuantMode(optionarg);
        opts.write_quantifiers().macrosQuantMode = value;
        opts.write_quantifiers().macrosQuantModeWasSetByUser = true;
        break;
      }
      case OptionEnum::MAXCUTSINCONTEXT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().maxCutsInContext = value;
        opts.write_arith().maxCutsInContextWasSetByUser = true;
        break;
      }
      case OptionEnum::MBQI:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().mbqi = value;
        opts.write_quantifiers().mbqiWasSetByUser = true;
        break;
      }
      case OptionEnum::MBQI_FAST_SYGUS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().mbqiFastSygus = value;
        opts.write_quantifiers().mbqiFastSygusWasSetByUser = true;
        break;
      }
      case OptionEnum::MBQI_INTERLEAVE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().mbqiInterleave = value;
        opts.write_quantifiers().mbqiInterleaveWasSetByUser = true;
        break;
      }
      case OptionEnum::MBQI_ONE_INST_PER_ROUND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().fmfOneInstPerRound = value;
        opts.write_quantifiers().fmfOneInstPerRoundWasSetByUser = true;
        break;
      }
      case OptionEnum::MINIMAL_UNSAT_CORES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().minimalUnsatCores = value;
        opts.write_smt().minimalUnsatCoresWasSetByUser = true;
        break;
      }
      case OptionEnum::MINISAT_DUMP_DIMACS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_prop().minisatDumpDimacs = value;
        opts.write_prop().minisatDumpDimacsWasSetByUser = true;
        break;
      }
      case OptionEnum::MINISAT_SIMPLIFICATION:
      {
        auto value = stringToMinisatSimpMode(optionarg);
        opts.write_prop().minisatSimpMode = value;
        opts.write_prop().minisatSimpModeWasSetByUser = true;
        break;
      }
      case OptionEnum::MINISCOPE_QUANT:
      {
        auto value = stringToMiniscopeQuantMode(optionarg);
        opts.write_quantifiers().miniscopeQuant = value;
        opts.write_quantifiers().miniscopeQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::MINISCOPE_QUANT_USER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().miniscopeQuantUser = value;
        opts.write_quantifiers().miniscopeQuantUserWasSetByUser = true;
        break;
      }
      case OptionEnum::MIPLIB_TRICK:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().arithMLTrick = value;
        opts.write_arith().arithMLTrickWasSetByUser = true;
        break;
      }
      case OptionEnum::MIPLIB_TRICK_SUBS:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().arithMLTrickSubstitutions = value;
        opts.write_arith().arithMLTrickSubstitutionsWasSetByUser = true;
        break;
      }
      case OptionEnum::MODEL_BASED_ARITH_BV_CONV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().modelBasedArithBvConv = value;
        opts.write_uf().modelBasedArithBvConvWasSetByUser = true;
        break;
      }
      case OptionEnum::MODEL_CORES:
      {
        auto value = stringToModelCoresMode(optionarg);
        opts.write_smt().modelCoresMode = value;
        opts.write_smt().modelCoresModeWasSetByUser = true;
        break;
      }
      case OptionEnum::MODEL_U_PRINT:
      {
        auto value = stringToModelUninterpPrintMode(optionarg);
        ioutils::setDefaultModelUninterpPrint(value);
        opts.write_printer().modelUninterpPrint = value;
        opts.write_printer().modelUninterpPrintWasSetByUser = true;
        break;
      }
      case OptionEnum::MODEL_VAR_ELIM_UNEVAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().modelVarElimUneval = value;
        opts.write_smt().modelVarElimUnevalWasSetByUser = true;
        break;
      }
      case OptionEnum::MULTI_TRIGGER_CACHE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().multiTriggerCache = value;
        opts.write_quantifiers().multiTriggerCacheWasSetByUser = true;
        break;
      }
      case OptionEnum::MULTI_TRIGGER_LINEAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().multiTriggerLinear = value;
        opts.write_quantifiers().multiTriggerLinearWasSetByUser = true;
        break;
      }
      case OptionEnum::MULTI_TRIGGER_PRIORITY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().multiTriggerPriority = value;
        opts.write_quantifiers().multiTriggerPriorityWasSetByUser = true;
        break;
      }
      case OptionEnum::MULTI_TRIGGER_WHEN_SINGLE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().multiTriggerWhenSingle = value;
        opts.write_quantifiers().multiTriggerWhenSingleWasSetByUser = true;
        break;
      }
      case OptionEnum::NEW_PROP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().newProp = value;
        opts.write_arith().newPropWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_COV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlCov = value;
        opts.write_arith().nlCovWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_COV_FORCE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlCovForce = value;
        opts.write_arith().nlCovForceWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_COV_LIFT:
      {
        auto value = stringTonlCovLiftingMode(optionarg);
        opts.write_arith().nlCovLifting = value;
        opts.write_arith().nlCovLiftingWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_COV_LINEAR_MODEL:
      {
        auto value = stringTonlCovLinearModelMode(optionarg);
        opts.write_arith().nlCovLinearModel = value;
        opts.write_arith().nlCovLinearModelWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_COV_PROJ:
      {
        auto value = stringTonlCovProjectionMode(optionarg);
        opts.write_arith().nlCovProjection = value;
        opts.write_arith().nlCovProjectionWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_COV_PRUNE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlCovPrune = value;
        opts.write_arith().nlCovPruneWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_COV_VAR_ELIM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlCovVarElim = value;
        opts.write_arith().nlCovVarElimWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT:
      {
        auto value = stringToNlExtMode(optionarg);
        opts.write_arith().nlExt = value;
        opts.write_arith().nlExtWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_ENT_CONF:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtEntailConflicts = value;
        opts.write_arith().nlExtEntailConflictsWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_FACTOR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtFactor = value;
        opts.write_arith().nlExtFactorWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_INC_PREC:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtIncPrecision = value;
        opts.write_arith().nlExtIncPrecisionWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_PURIFY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtPurify = value;
        opts.write_arith().nlExtPurifyWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_RBOUND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtResBound = value;
        opts.write_arith().nlExtResBoundWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_REWRITE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtRewrites = value;
        opts.write_arith().nlExtRewritesWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_SPLIT_ZERO:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtSplitZero = value;
        opts.write_arith().nlExtSplitZeroWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_TF_TAYLOR_DEG:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().nlExtTfTaylorDegree = value;
        opts.write_arith().nlExtTfTaylorDegreeWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_TF_TPLANES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtTfTangentPlanes = value;
        opts.write_arith().nlExtTfTangentPlanesWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_TPLANES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtTangentPlanes = value;
        opts.write_arith().nlExtTangentPlanesWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_EXT_TPLANES_INTERLEAVE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlExtTangentPlanesInterleave = value;
        opts.write_arith().nlExtTangentPlanesInterleaveWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_ICP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlICP = value;
        opts.write_arith().nlICPWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_RLV:
      {
        auto value = stringToNlRlvMode(optionarg);
        opts.write_arith().nlRlvMode = value;
        opts.write_arith().nlRlvModeWasSetByUser = true;
        break;
      }
      case OptionEnum::NL_RLV_ASSERT_BOUNDS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().nlRlvAssertBounds = value;
        opts.write_arith().nlRlvAssertBoundsWasSetByUser = true;
        break;
      }
      case OptionEnum::ON_REPEAT_ITE_SIMP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().doITESimpOnRepeat = value;
        opts.write_smt().doITESimpOnRepeatWasSetByUser = true;
        break;
      }
      case OptionEnum::OPT_RES_RECONSTRUCTION_SIZE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().optResReconSize = value;
        opts.write_proof().optResReconSizeWasSetByUser = true;
        break;
      }
      case OptionEnum::ORACLES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().oracles = value;
        opts.write_quantifiers().oraclesWasSetByUser = true;
        break;
      }
      case OptionEnum::OUT:
      {
        auto value = handlers::handleOption<ManagedOut>(name, optionarg);
        opts.write_base().out = value;
        opts.write_base().outWasSetByUser = true;
        break;
      }
      case OptionEnum::OUTPUT:
      {
        auto value = stringToOutputTag(optionarg);
        opts.handler().enableOutputTag(name, value);
        break;
      }
      case OptionEnum::OUTPUT_LANG:
      {
        auto value = opts.handler().stringToLanguage(name, optionarg);
        ioutils::setDefaultOutputLanguage(value);
        opts.write_printer().outputLanguage = value;
        opts.write_printer().outputLanguageWasSetByUser = true;
        break;
      }
      case OptionEnum::PARSE_ONLY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_base().parseOnly = value;
        opts.write_base().parseOnlyWasSetByUser = true;
        break;
      }
      case OptionEnum::PARSE_SKOLEM_DEFINITIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parser().parseSkolemDefinitions = value;
        opts.write_parser().parseSkolemDefinitionsWasSetByUser = true;
        break;
      }
      case OptionEnum::PARSING_MODE:
      {
        auto value = stringToParsingMode(optionarg);
        opts.write_parser().parsingMode = value;
        opts.write_parser().parsingModeWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTIAL_TRIGGERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().partialTriggers = value;
        opts.write_quantifiers().partialTriggersWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTITION_CHECK:
      {
        auto value = stringToCheckMode(optionarg);
        opts.write_parallel().partitionCheck = value;
        opts.write_parallel().partitionCheckWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTITION_CONFLICT_SIZE:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_parallel().partitionConflictSize = value;
        opts.write_parallel().partitionConflictSizeWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTITION_START_TIME:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_parallel().partitionStartTime = value;
        opts.write_parallel().partitionStartTimeWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTITION_STRATEGY:
      {
        auto value = stringToPartitionMode(optionarg);
        opts.write_parallel().partitionStrategy = value;
        opts.write_parallel().partitionStrategyWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTITION_TIME_INTERVAL:
      {
        auto value = handlers::handleOption<double>(name, optionarg);
        opts.write_parallel().partitionTimeInterval = value;
        opts.write_parallel().partitionTimeIntervalWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTITION_TLIMIT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_parallel().partitionTimeLimit = value;
        opts.write_parallel().partitionTimeLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::PARTITION_WHEN:
      {
        auto value = stringToPartitionWhenMode(optionarg);
        opts.write_parallel().partitionWhen = value;
        opts.write_parallel().partitionWhenWasSetByUser = true;
        break;
      }
      case OptionEnum::PB_REWRITES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().pbRewrites = value;
        opts.write_arith().pbRewritesWasSetByUser = true;
        break;
      }
      case OptionEnum::PIVOT_THRESHOLD:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().arithPivotThreshold = value;
        opts.write_arith().arithPivotThresholdWasSetByUser = true;
        break;
      }
      case OptionEnum::PLUGIN_NOTIFY_SAT_CLAUSE_IN_SOLVE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_base().pluginNotifySatClauseInSolve = value;
        opts.write_base().pluginNotifySatClauseInSolveWasSetByUser = true;
        break;
      }
      case OptionEnum::PLUGIN_SHARE_SKOLEMS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_base().pluginShareSkolems = value;
        opts.write_base().pluginShareSkolemsWasSetByUser = true;
        break;
      }
      case OptionEnum::POOL_INST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().poolInst = value;
        opts.write_quantifiers().poolInstWasSetByUser = true;
        break;
      }
      case OptionEnum::PORTFOLIO_JOBS:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_driver().portfolioJobs = value;
        opts.write_driver().portfolioJobsWasSetByUser = true;
        break;
      }
      case OptionEnum::PP_ASSERT_MAX_SUB_SIZE:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().ppAssertMaxSubSize = value;
        opts.write_arith().ppAssertMaxSubSizeWasSetByUser = true;
        break;
      }
      case OptionEnum::PRE_SKOLEM_QUANT:
      {
        auto value = stringToPreSkolemQuantMode(optionarg);
        opts.write_quantifiers().preSkolemQuant = value;
        opts.write_quantifiers().preSkolemQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::PRE_SKOLEM_QUANT_NESTED:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().preSkolemQuantNested = value;
        opts.write_quantifiers().preSkolemQuantNestedWasSetByUser = true;
        break;
      }
      case OptionEnum::PRENEX_QUANT:
      {
        auto value = stringToPrenexQuantMode(optionarg);
        opts.write_quantifiers().prenexQuant = value;
        opts.write_quantifiers().prenexQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::PRENEX_QUANT_USER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().prenexQuantUser = value;
        opts.write_quantifiers().prenexQuantUserWasSetByUser = true;
        break;
      }
      case OptionEnum::PREPROCESS_ONLY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_base().preprocessOnly = value;
        opts.write_base().preprocessOnlyWasSetByUser = true;
        break;
      }
      case OptionEnum::PREREGISTER_MODE:
      {
        auto value = stringToPreRegisterMode(optionarg);
        opts.write_prop().preRegisterMode = value;
        opts.write_prop().preRegisterModeWasSetByUser = true;
        break;
      }
      case OptionEnum::PRINT_ARITH_LIT_TOKEN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        ioutils::setDefaultPrintArithLitToken(value);
        opts.write_printer().printArithLitToken = value;
        opts.write_printer().printArithLitTokenWasSetByUser = true;
        break;
      }
      case OptionEnum::PRINT_CORES_FULL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().printCoresFull = value;
        opts.write_smt().printCoresFullWasSetByUser = true;
        break;
      }
      case OptionEnum::PRINT_DOT_CLUSTERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().printDotClusters = value;
        opts.write_proof().printDotClustersWasSetByUser = true;
        break;
      }
      case OptionEnum::PRINT_INST:
      {
        auto value = stringToPrintInstMode(optionarg);
        opts.write_quantifiers().printInstMode = value;
        opts.write_quantifiers().printInstModeWasSetByUser = true;
        break;
      }
      case OptionEnum::PRINT_INST_FULL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().printInstFull = value;
        opts.write_quantifiers().printInstFullWasSetByUser = true;
        break;
      }
      case OptionEnum::PRINT_SKOLEM_DEFINITIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        ioutils::setDefaultPrintSkolemDefinitions(value);
        opts.write_printer().printSkolemDefinitions = value;
        opts.write_printer().printSkolemDefinitionsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRINT_SUCCESS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().printSuccess = value;
        opts.write_driver().printSuccessWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_ABDUCTS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceAbducts = value;
        opts.write_smt().produceAbductsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_ASSERTIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceAssertions = value;
        opts.write_smt().produceAssertionsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_ASSIGNMENTS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceAssignments = value;
        opts.write_smt().produceAssignmentsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_DIFFICULTY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceDifficulty = value;
        opts.write_smt().produceDifficultyWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_INTERPOLANTS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceInterpolants = value;
        opts.write_smt().produceInterpolantsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_LEARNED_LITERALS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceLearnedLiterals = value;
        opts.write_smt().produceLearnedLiteralsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_MODELS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceModels = value;
        opts.write_smt().produceModelsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_PROOFS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceProofs = value;
        opts.write_smt().produceProofsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_UNSAT_ASSUMPTIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().unsatAssumptions = value;
        opts.write_smt().unsatAssumptionsWasSetByUser = true;
        break;
      }
      case OptionEnum::PRODUCE_UNSAT_CORES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().produceUnsatCores = value;
        opts.write_smt().produceUnsatCoresWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_ALETHE_DEFINE_SKOLEMS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofAletheDefineSkolems = value;
        opts.write_proof().proofAletheDefineSkolemsWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_ALETHE_RES_PIVOTS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofAletheResPivots = value;
        opts.write_proof().proofAletheResPivotsWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_ALLOW_TRUST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofAllowTrust = value;
        opts.write_proof().proofAllowTrustWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_CHECK:
      {
        auto value = stringToProofCheckMode(optionarg);
        opts.write_proof().proofCheck = value;
        opts.write_proof().proofCheckWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_DAG_GLOBAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofDagGlobal = value;
        opts.write_proof().proofDagGlobalWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_DOT_DAG:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().printDotAsDAG = value;
        opts.write_proof().printDotAsDAGWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_ELIM_SUBTYPES:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofElimSubtypes = value;
        opts.write_proof().proofElimSubtypesWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_FORMAT_MODE:
      {
        auto value = stringToProofFormatMode(optionarg);
        opts.write_proof().proofFormatMode = value;
        opts.write_proof().proofFormatModeWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_GRANULARITY:
      {
        auto value = stringToProofGranularityMode(optionarg);
        opts.write_proof().proofGranularityMode = value;
        opts.write_proof().proofGranularityModeWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_MODE:
      {
        auto value = stringToProofMode(optionarg);
        opts.write_smt().proofMode = value;
        opts.write_smt().proofModeWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_PEDANTIC:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.handler().checkMaximum(name, value, static_cast<uint64_t>(100));
        opts.write_proof().proofPedantic = value;
        opts.write_proof().proofPedanticWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_PP_MERGE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofPpMerge = value;
        opts.write_proof().proofPpMergeWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_PRINT_CONCLUSION:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofPrintConclusion = value;
        opts.write_proof().proofPrintConclusionWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_PRINT_REFERENCE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofPrintReference = value;
        opts.write_proof().proofPrintReferenceWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_PRUNE_INPUT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().proofPruneInput = value;
        opts.write_proof().proofPruneInputWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_REWRITE_RCONS_REC_LIMIT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_proof().proofRewriteRconsRecLimit = value;
        opts.write_proof().proofRewriteRconsRecLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::PROOF_REWRITE_RCONS_STEP_LIMIT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_proof().proofRewriteRconsStepLimit = value;
        opts.write_proof().proofRewriteRconsStepLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::PROP_PROOF_MODE:
      {
        auto value = stringToPropProofMode(optionarg);
        opts.write_proof().propProofMode = value;
        opts.write_proof().propProofModeWasSetByUser = true;
        break;
      }
      case OptionEnum::PROP_ROW_LENGTH:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().arithPropagateMaxLength = value;
        opts.write_arith().arithPropagateMaxLengthWasSetByUser = true;
        break;
      }
      case OptionEnum::PURIFY_TRIGGERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().purifyTriggers = value;
        opts.write_quantifiers().purifyTriggersWasSetByUser = true;
        break;
      }
      case OptionEnum::QUANT_ALPHA_EQUIV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().quantAlphaEquiv = value;
        opts.write_quantifiers().quantAlphaEquivWasSetByUser = true;
        break;
      }
      case OptionEnum::QUANT_DSPLIT:
      {
        auto value = stringToQuantDSplitMode(optionarg);
        opts.write_quantifiers().quantDynamicSplit = value;
        opts.write_quantifiers().quantDynamicSplitWasSetByUser = true;
        break;
      }
      case OptionEnum::QUANT_FUN_WD:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().quantFunWellDefined = value;
        opts.write_quantifiers().quantFunWellDefinedWasSetByUser = true;
        break;
      }
      case OptionEnum::QUANT_IND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().quantInduction = value;
        opts.write_quantifiers().quantInductionWasSetByUser = true;
        break;
      }
      case OptionEnum::QUANT_REP_MODE:
      {
        auto value = stringToQuantRepMode(optionarg);
        opts.write_quantifiers().quantRepMode = value;
        opts.write_quantifiers().quantRepModeWasSetByUser = true;
        break;
      }
      case OptionEnum::QUIET:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().decreaseVerbosity(name, value);
        break;
      }
      case OptionEnum::RANDOM_FREQ:
      {
        auto value = handlers::handleOption<double>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<double>(0.0));
        opts.handler().checkMaximum(name, value, static_cast<double>(1.0));
        opts.write_prop().satRandomFreq = value;
        opts.write_prop().satRandomFreqWasSetByUser = true;
        break;
      }
      case OptionEnum::RANDOM_PARTITIONING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parallel().randomPartitioning = value;
        opts.write_parallel().randomPartitioningWasSetByUser = true;
        break;
      }
      case OptionEnum::RE_ELIM:
      {
        auto value = stringToRegExpElimMode(optionarg);
        opts.write_strings().regExpElim = value;
        opts.write_strings().regExpElimWasSetByUser = true;
        break;
      }
      case OptionEnum::RE_INTER_MODE:
      {
        auto value = stringToRegExpInterMode(optionarg);
        opts.write_strings().stringRegExpInterMode = value;
        opts.write_strings().stringRegExpInterModeWasSetByUser = true;
        break;
      }
      case OptionEnum::REGISTER_QUANT_BODY_TERMS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().registerQuantBodyTerms = value;
        opts.write_quantifiers().registerQuantBodyTermsWasSetByUser = true;
        break;
      }
      case OptionEnum::RELATIONAL_TRIGGERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().relationalTriggers = value;
        opts.write_quantifiers().relationalTriggersWasSetByUser = true;
        break;
      }
      case OptionEnum::RELEVANCE_FILTER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_theory().relevanceFilter = value;
        opts.write_theory().relevanceFilterWasSetByUser = true;
        break;
      }
      case OptionEnum::RELEVANT_TRIGGERS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().relevantTriggers = value;
        opts.write_quantifiers().relevantTriggersWasSetByUser = true;
        break;
      }
      case OptionEnum::RELS_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_sets().relsExp = value;
        opts.write_sets().relsExpWasSetByUser = true;
        break;
      }
      case OptionEnum::REPEAT_SIMP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().repeatSimp = value;
        opts.write_smt().repeatSimpWasSetByUser = true;
        break;
      }
      case OptionEnum::REPLAY_EARLY_CLOSE_DEPTH:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().replayEarlyCloseDepths = value;
        opts.write_arith().replayEarlyCloseDepthsWasSetByUser = true;
        break;
      }
      case OptionEnum::REPLAY_LEMMA_REJECT_CUT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().lemmaRejectCutSize = value;
        opts.write_arith().lemmaRejectCutSizeWasSetByUser = true;
        break;
      }
      case OptionEnum::REPLAY_NUM_ERR_PENALTY:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().replayNumericFailurePenalty = value;
        opts.write_arith().replayNumericFailurePenaltyWasSetByUser = true;
        break;
      }
      case OptionEnum::REPLAY_REJECT_CUT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().replayRejectCutSize = value;
        opts.write_arith().replayRejectCutSizeWasSetByUser = true;
        break;
      }
      case OptionEnum::RESTART_INT_BASE:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_prop().satRestartFirst = value;
        opts.write_prop().satRestartFirstWasSetByUser = true;
        break;
      }
      case OptionEnum::RESTART_INT_INC:
      {
        auto value = handlers::handleOption<double>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<double>(0.0));
        opts.write_prop().satRestartInc = value;
        opts.write_prop().satRestartIncWasSetByUser = true;
        break;
      }
      case OptionEnum::RESTRICT_PIVOTS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().restrictedPivots = value;
        opts.write_arith().restrictedPivotsWasSetByUser = true;
        break;
      }
      case OptionEnum::REVERT_ARITH_MODELS_ON_UNSAT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().revertArithModels = value;
        opts.write_arith().revertArithModelsWasSetByUser = true;
        break;
      }
      case OptionEnum::RLIMIT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_base().cumulativeResourceLimit = value;
        opts.write_base().cumulativeResourceLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::RLIMIT_PER:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_base().perCallResourceLimit = value;
        opts.write_base().perCallResourceLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::RR_TURNS:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().rrTurns = value;
        opts.write_arith().rrTurnsWasSetByUser = true;
        break;
      }
      case OptionEnum::RWEIGHT:
      {
        auto value = handlers::handleOption<std::string>(name, optionarg);
        opts.handler().setResourceWeight(name, value);
        break;
      }
      case OptionEnum::SAFE_OPTIONS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_base().safeOptions = value;
        opts.write_base().safeOptionsWasSetByUser = true;
        break;
      }
      case OptionEnum::SAT_PROOF_MIN_DIMACS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_proof().satProofMinDimacs = value;
        opts.write_proof().satProofMinDimacsWasSetByUser = true;
        break;
      }
      case OptionEnum::SAT_RANDOM_SEED:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_prop().satRandomSeed = value;
        opts.write_prop().satRandomSeedWasSetByUser = true;
        break;
      }
      case OptionEnum::SAT_SOLVER:
      {
        auto value = stringToSatSolverMode(optionarg);
        opts.write_prop().satSolver = value;
        opts.write_prop().satSolverWasSetByUser = true;
        break;
      }
      case OptionEnum::SE_SOLVE_INT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().trySolveIntStandardEffort = value;
        opts.write_arith().trySolveIntStandardEffortWasSetByUser = true;
        break;
      }
      case OptionEnum::SEED:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_driver().seed = value;
        opts.write_driver().seedWasSetByUser = true;
        break;
      }
      case OptionEnum::SEGV_SPIN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().segvSpin = value;
        opts.write_driver().segvSpinWasSetByUser = true;
        break;
      }
      case OptionEnum::SEMANTIC_CHECKS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parser().semanticChecks = value;
        opts.write_parser().semanticChecksWasSetByUser = true;
        break;
      }
      case OptionEnum::SEP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_sep().sep = value;
        opts.write_sep().sepWasSetByUser = true;
        break;
      }
      case OptionEnum::SEP_MIN_REFINE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_sep().sepMinimalRefine = value;
        opts.write_sep().sepMinimalRefineWasSetByUser = true;
        break;
      }
      case OptionEnum::SEP_PRE_SKOLEM_EMP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_sep().sepPreSkolemEmp = value;
        opts.write_sep().sepPreSkolemEmpWasSetByUser = true;
        break;
      }
      case OptionEnum::SEQ_ARRAY:
      {
        auto value = stringToSeqArrayMode(optionarg);
        opts.write_strings().seqArray = value;
        opts.write_strings().seqArrayWasSetByUser = true;
        break;
      }
      case OptionEnum::SETS_CARD_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_sets().setsCardExp = value;
        opts.write_sets().setsCardExpWasSetByUser = true;
        break;
      }
      case OptionEnum::SETS_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_sets().setsExp = value;
        opts.write_sets().setsExpWasSetByUser = true;
        break;
      }
      case OptionEnum::SETS_PROXY_LEMMAS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_sets().setsProxyLemmas = value;
        opts.write_sets().setsProxyLemmasWasSetByUser = true;
        break;
      }
      case OptionEnum::SHOW_CONFIG:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().showConfiguration(name, value);
        opts.write_driver().showConfiguration = value;
        opts.write_driver().showConfigurationWasSetByUser = true;
        break;
      }
      case OptionEnum::SHOW_TRACE_TAGS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().showTraceTags(name, value);
        opts.write_driver().showTraceTags = value;
        opts.write_driver().showTraceTagsWasSetByUser = true;
        break;
      }
      case OptionEnum::SIMP_ITE_COMPRESS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().compressItes = value;
        opts.write_smt().compressItesWasSetByUser = true;
        break;
      }
      case OptionEnum::SIMP_WITH_CARE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().simplifyWithCareEnabled = value;
        opts.write_smt().simplifyWithCareEnabledWasSetByUser = true;
        break;
      }
      case OptionEnum::SIMPLEX_CHECK_PERIOD:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_arith().arithSimplexCheckPeriod = value;
        opts.write_arith().arithSimplexCheckPeriodWasSetByUser = true;
        break;
      }
      case OptionEnum::SIMPLIFICATION:
      {
        auto value = stringToSimplificationMode(optionarg);
        opts.write_smt().simplificationMode = value;
        opts.write_smt().simplificationModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SIMPLIFICATION_BCP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().simplificationBoolConstProp = value;
        opts.write_smt().simplificationBoolConstPropWasSetByUser = true;
        break;
      }
      case OptionEnum::SOI_QE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().soiQuickExplain = value;
        opts.write_arith().soiQuickExplainWasSetByUser = true;
        break;
      }
      case OptionEnum::SOLVE_BV_AS_INT:
      {
        auto value = stringToSolveBVAsIntMode(optionarg);
        opts.write_smt().solveBVAsInt = value;
        opts.write_smt().solveBVAsIntWasSetByUser = true;
        break;
      }
      case OptionEnum::SOLVE_INT_AS_BV:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.handler().checkMaximum(name, value, static_cast<uint64_t>(4294967295));
        opts.write_smt().solveIntAsBV = value;
        opts.write_smt().solveIntAsBVWasSetByUser = true;
        break;
      }
      case OptionEnum::SOLVE_REAL_AS_INT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().solveRealAsInt = value;
        opts.write_smt().solveRealAsIntWasSetByUser = true;
        break;
      }
      case OptionEnum::SORT_INFERENCE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().sortInference = value;
        opts.write_smt().sortInferenceWasSetByUser = true;
        break;
      }
      case OptionEnum::STANDARD_EFFORT_VARIABLE_ORDER_PIVOTS:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_arith().arithStandardCheckVarOrderPivots = value;
        opts.write_arith().arithStandardCheckVarOrderPivotsWasSetByUser = true;
        break;
      }
      case OptionEnum::STATIC_LEARNING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().staticLearning = value;
        opts.write_smt().staticLearningWasSetByUser = true;
        break;
      }
      case OptionEnum::STATS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().setStats(name, value);
        opts.write_base().statistics = value;
        opts.write_base().statisticsWasSetByUser = true;
        break;
      }
      case OptionEnum::STATS_ALL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().setStatsDetail(name, value);
        opts.write_base().statisticsAll = value;
        opts.write_base().statisticsAllWasSetByUser = true;
        break;
      }
      case OptionEnum::STATS_EVERY_QUERY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().setStatsDetail(name, value);
        opts.write_base().statisticsEveryQuery = value;
        opts.write_base().statisticsEveryQueryWasSetByUser = true;
        break;
      }
      case OptionEnum::STATS_INTERNAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().setStatsDetail(name, value);
        opts.write_base().statisticsInternal = value;
        opts.write_base().statisticsInternalWasSetByUser = true;
        break;
      }
      case OptionEnum::STRICT_PARSING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().strictParsing(name, value);
        opts.write_parser().strictParsing = value;
        opts.write_parser().strictParsingWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_ALPHA_CARD:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.handler().checkMaximum(name, value, static_cast<uint64_t>(196608));
        opts.write_strings().stringsAlphaCard = value;
        opts.write_strings().stringsAlphaCardWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_CHECK_ENTAIL_LEN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringCheckEntailLen = value;
        opts.write_strings().stringCheckEntailLenWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_DEQ_EXT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringsDeqExt = value;
        opts.write_strings().stringsDeqExtWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_EAGER_EVAL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringEagerEval = value;
        opts.write_strings().stringEagerEvalWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_EAGER_LEN_RE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringsEagerLenEntRegexp = value;
        opts.write_strings().stringsEagerLenEntRegexpWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_EAGER_REG:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringEagerReg = value;
        opts.write_strings().stringEagerRegWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_EAGER_SOLVER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringEagerSolver = value;
        opts.write_strings().stringEagerSolverWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringExp = value;
        opts.write_strings().stringExpWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_FF:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringFlatForms = value;
        opts.write_strings().stringFlatFormsWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_FMF:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringFMF = value;
        opts.write_strings().stringFMFWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_INFER_AS_LEMMAS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringInferAsLemmas = value;
        opts.write_strings().stringInferAsLemmasWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_INFER_SYM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringInferSym = value;
        opts.write_strings().stringInferSymWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_LAZY_PP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringLazyPreproc = value;
        opts.write_strings().stringLazyPreprocWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_LEN_NORM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringLenNorm = value;
        opts.write_strings().stringLenNormWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_MBR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringModelBasedReduction = value;
        opts.write_strings().stringModelBasedReductionWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_MODEL_MAX_LEN:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.handler().checkMaximum(name, value, static_cast<uint64_t>(2147483647));
        opts.write_strings().stringsModelMaxLength = value;
        opts.write_strings().stringsModelMaxLengthWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_PROCESS_LOOP_MODE:
      {
        auto value = stringToProcessLoopMode(optionarg);
        opts.write_strings().stringProcessLoopMode = value;
        opts.write_strings().stringProcessLoopModeWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_RE_DERIVE_CONF:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringRegexpDeriveConflicts = value;
        opts.write_strings().stringRegexpDeriveConflictsWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_RE_POSC_EAGER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringRegexpPosConcatEager = value;
        opts.write_strings().stringRegexpPosConcatEagerWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_REC_ARITH_APPROX:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringRecArithApprox = value;
        opts.write_strings().stringRecArithApproxWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_REGEXP_INCLUSION:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringRegexpInclusion = value;
        opts.write_strings().stringRegexpInclusionWasSetByUser = true;
        break;
      }
      case OptionEnum::STRINGS_REXPLAIN_LEMMAS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_strings().stringRExplainLemmas = value;
        opts.write_strings().stringRExplainLemmasWasSetByUser = true;
        break;
      }
      case OptionEnum::SUB_CBQI:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().quantSubCbqi = value;
        opts.write_quantifiers().quantSubCbqiWasSetByUser = true;
        break;
      }
      case OptionEnum::SUB_CBQI_TIMEOUT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_quantifiers().quantSubCbqiTimeout = value;
        opts.write_quantifiers().quantSubCbqiTimeoutWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygus = value;
        opts.write_quantifiers().sygusWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_ABORT_SIZE:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_datatypes().sygusAbortSize = value;
        opts.write_datatypes().sygusAbortSizeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_ADD_CONST_GRAMMAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusAddConstGrammar = value;
        opts.write_quantifiers().sygusAddConstGrammarWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_ARG_RELEVANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusArgRelevant = value;
        opts.write_quantifiers().sygusArgRelevantWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_AUTO_UNFOLD:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusInvAutoUnfold = value;
        opts.write_quantifiers().sygusInvAutoUnfoldWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_BOOL_ITE_RETURN_CONST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusBoolIteReturnConst = value;
        opts.write_quantifiers().sygusBoolIteReturnConstWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_CORE_CONNECTIVE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusCoreConnective = value;
        opts.write_quantifiers().sygusCoreConnectiveWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_CREPAIR_ABORT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusConstRepairAbort = value;
        opts.write_quantifiers().sygusConstRepairAbortWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_ENUM:
      {
        auto value = stringToSygusEnumMode(optionarg);
        opts.write_quantifiers().sygusEnumMode = value;
        opts.write_quantifiers().sygusEnumModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_ENUM_FAST_NUM_CONSTS:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_quantifiers().sygusEnumFastNumConsts = value;
        opts.write_quantifiers().sygusEnumFastNumConstsWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_ENUM_RANDOM_P:
      {
        auto value = handlers::handleOption<double>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<double>(0.0));
        opts.handler().checkMaximum(name, value, static_cast<double>(1.0));
        opts.write_quantifiers().sygusEnumRandomP = value;
        opts.write_quantifiers().sygusEnumRandomPWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_EVAL_UNFOLD:
      {
        auto value = stringToSygusEvalUnfoldMode(optionarg);
        opts.write_quantifiers().sygusEvalUnfoldMode = value;
        opts.write_quantifiers().sygusEvalUnfoldModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_EXPR_MINER_CHECK_TIMEOUT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_quantifiers().sygusExprMinerCheckTimeout = value;
        opts.write_quantifiers().sygusExprMinerCheckTimeoutWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_FAIR:
      {
        auto value = stringToSygusFairMode(optionarg);
        opts.write_datatypes().sygusFair = value;
        opts.write_datatypes().sygusFairWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_FAIR_MAX:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().sygusFairMax = value;
        opts.write_datatypes().sygusFairMaxWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_FILTER_SOL:
      {
        auto value = stringToSygusFilterSolMode(optionarg);
        opts.write_quantifiers().sygusFilterSolMode = value;
        opts.write_quantifiers().sygusFilterSolModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_FILTER_SOL_REV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusFilterSolRevSubsume = value;
        opts.write_quantifiers().sygusFilterSolRevSubsumeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_GRAMMAR_CONS:
      {
        auto value = stringToSygusGrammarConsMode(optionarg);
        opts.write_quantifiers().sygusGrammarConsMode = value;
        opts.write_quantifiers().sygusGrammarConsModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_GRAMMAR_NORM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusGrammarNorm = value;
        opts.write_quantifiers().sygusGrammarNormWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_GRAMMAR_USE_DISJ:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusGrammarUseDisj = value;
        opts.write_quantifiers().sygusGrammarUseDisjWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_INFERENCE:
      {
        auto value = stringToSygusInferenceMode(optionarg);
        opts.write_quantifiers().sygusInference = value;
        opts.write_quantifiers().sygusInferenceWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_INST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusInst = value;
        opts.write_quantifiers().sygusInstWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_INST_MODE:
      {
        auto value = stringToSygusInstMode(optionarg);
        opts.write_quantifiers().sygusInstMode = value;
        opts.write_quantifiers().sygusInstModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_INST_SCOPE:
      {
        auto value = stringToSygusInstScope(optionarg);
        opts.write_quantifiers().sygusInstScope = value;
        opts.write_quantifiers().sygusInstScopeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_INST_TERM_SEL:
      {
        auto value = stringToSygusInstTermSelMode(optionarg);
        opts.write_quantifiers().sygusInstTermSel = value;
        opts.write_quantifiers().sygusInstTermSelWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_INV_TEMPL:
      {
        auto value = stringToSygusInvTemplMode(optionarg);
        opts.write_quantifiers().sygusInvTemplMode = value;
        opts.write_quantifiers().sygusInvTemplModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_INV_TEMPL_WHEN_SG:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusInvTemplWhenSyntax = value;
        opts.write_quantifiers().sygusInvTemplWhenSyntaxWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_MIN_GRAMMAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusMinGrammar = value;
        opts.write_quantifiers().sygusMinGrammarWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_OUT:
      {
        auto value = stringToSygusSolutionOutMode(optionarg);
        opts.write_quantifiers().sygusOut = value;
        opts.write_quantifiers().sygusOutWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_PBE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusUnifPbe = value;
        opts.write_quantifiers().sygusUnifPbeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_PBE_MULTI_FAIR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusPbeMultiFair = value;
        opts.write_quantifiers().sygusPbeMultiFairWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_PBE_MULTI_FAIR_DIFF:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().sygusPbeMultiFairDiff = value;
        opts.write_quantifiers().sygusPbeMultiFairDiffWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_QE_PREPROC:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusQePreproc = value;
        opts.write_quantifiers().sygusQePreprocWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_QUERY_GEN:
      {
        auto value = stringToSygusQueryGenMode(optionarg);
        opts.write_quantifiers().sygusQueryGen = value;
        opts.write_quantifiers().sygusQueryGenWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_QUERY_GEN_DUMP_FILES:
      {
        auto value = stringToSygusQueryDumpFilesMode(optionarg);
        opts.write_quantifiers().sygusQueryGenDumpFiles = value;
        opts.write_quantifiers().sygusQueryGenDumpFilesWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_QUERY_GEN_FILTER_SOLVED:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusQueryFilterSolved = value;
        opts.write_quantifiers().sygusQueryFilterSolvedWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_QUERY_GEN_THRESH:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_quantifiers().sygusQueryGenThresh = value;
        opts.write_quantifiers().sygusQueryGenThreshWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_REC_FUN:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRecFun = value;
        opts.write_quantifiers().sygusRecFunWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_REC_FUN_EVAL_LIMIT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_quantifiers().sygusRecFunEvalLimit = value;
        opts.write_quantifiers().sygusRecFunEvalLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_REPAIR_CONST:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRepairConst = value;
        opts.write_quantifiers().sygusRepairConstWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_REPAIR_CONST_TIMEOUT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_quantifiers().sygusRepairConstTimeout = value;
        opts.write_quantifiers().sygusRepairConstTimeoutWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_REWRITER:
      {
        auto value = stringToSygusRewriterMode(optionarg);
        opts.write_datatypes().sygusRewriter = value;
        opts.write_datatypes().sygusRewriterWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_ACCEL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthAccel = value;
        opts.write_quantifiers().sygusRewSynthAccelWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_CHECK:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthCheck = value;
        opts.write_quantifiers().sygusRewSynthCheckWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_CONG:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthFilterCong = value;
        opts.write_quantifiers().sygusRewSynthFilterCongWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_MATCH:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthFilterMatch = value;
        opts.write_quantifiers().sygusRewSynthFilterMatchWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_NL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthFilterNonLinear = value;
        opts.write_quantifiers().sygusRewSynthFilterNonLinearWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_FILTER_ORDER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthFilterOrder = value;
        opts.write_quantifiers().sygusRewSynthFilterOrderWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_INPUT_NVARS:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthInputNVars = value;
        opts.write_quantifiers().sygusRewSynthInputNVarsWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_RR_SYNTH_REC:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusRewSynthRec = value;
        opts.write_quantifiers().sygusRewSynthRecWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SAMPLE_FP_UNIFORM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusSampleFpUniform = value;
        opts.write_quantifiers().sygusSampleFpUniformWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SAMPLE_GRAMMAR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusSampleGrammar = value;
        opts.write_quantifiers().sygusSampleGrammarWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SAMPLES:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().sygusSamples = value;
        opts.write_quantifiers().sygusSamplesWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SI:
      {
        auto value = stringToCegqiSingleInvMode(optionarg);
        opts.write_quantifiers().cegqiSingleInvMode = value;
        opts.write_quantifiers().cegqiSingleInvModeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SI_ABORT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().cegqiSingleInvAbort = value;
        opts.write_quantifiers().cegqiSingleInvAbortWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SI_RCONS:
      {
        auto value = stringToCegqiSingleInvRconsMode(optionarg);
        opts.write_quantifiers().cegqiSingleInvReconstruct = value;
        opts.write_quantifiers().cegqiSingleInvReconstructWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SI_RCONS_LIMIT:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().cegqiSingleInvReconstructLimit = value;
        opts.write_quantifiers().cegqiSingleInvReconstructLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SI_RCONS_P:
      {
        auto value = handlers::handleOption<double>(name, optionarg);
        opts.handler().checkMinimum(name, value, static_cast<double>(0.0));
        opts.handler().checkMaximum(name, value, static_cast<double>(1.0));
        opts.write_quantifiers().cegqiSingleInvReconstructP = value;
        opts.write_quantifiers().cegqiSingleInvReconstructPWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SIMPLE_SYM_BREAK:
      {
        auto value = stringToSygusSimpleSymBreakMode(optionarg);
        opts.write_datatypes().sygusSimpleSymBreak = value;
        opts.write_datatypes().sygusSimpleSymBreakWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_STREAM:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusStream = value;
        opts.write_quantifiers().sygusStreamWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SYM_BREAK_LAZY:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().sygusSymBreakLazy = value;
        opts.write_datatypes().sygusSymBreakLazyWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SYM_BREAK_PBE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().sygusSymBreakPbe = value;
        opts.write_datatypes().sygusSymBreakPbeWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_SYM_BREAK_RLV:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_datatypes().sygusSymBreakRlv = value;
        opts.write_datatypes().sygusSymBreakRlvWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_UNIF_COND_INDEPENDENT_NO_REPEAT_SOL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusUnifCondIndNoRepeatSol = value;
        opts.write_quantifiers().sygusUnifCondIndNoRepeatSolWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_UNIF_PI:
      {
        auto value = stringToSygusUnifPiMode(optionarg);
        opts.write_quantifiers().sygusUnifPi = value;
        opts.write_quantifiers().sygusUnifPiWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_UNIF_SHUFFLE_COND:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().sygusUnifShuffleCond = value;
        opts.write_quantifiers().sygusUnifShuffleCondWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_VERIFY_INST_MAX_ROUNDS:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_quantifiers().sygusVerifyInstMaxRounds = value;
        opts.write_quantifiers().sygusVerifyInstMaxRoundsWasSetByUser = true;
        break;
      }
      case OptionEnum::SYGUS_VERIFY_TIMEOUT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_quantifiers().sygusVerifyTimeout = value;
        opts.write_quantifiers().sygusVerifyTimeoutWasSetByUser = true;
        break;
      }
      case OptionEnum::SYMMETRY_BREAKER:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufSymmetryBreaker = value;
        opts.write_uf().ufSymmetryBreakerWasSetByUser = true;
        break;
      }
      case OptionEnum::TC_MODE:
      {
        auto value = stringToTcMode(optionarg);
        opts.write_theory().tcMode = value;
        opts.write_theory().tcModeWasSetByUser = true;
        break;
      }
      case OptionEnum::TERM_DB_MODE:
      {
        auto value = stringToTermDbMode(optionarg);
        opts.write_quantifiers().termDbMode = value;
        opts.write_quantifiers().termDbModeWasSetByUser = true;
        break;
      }
      case OptionEnum::TERM_SORT_OVERLOAD:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_parser().termSortOverloading = value;
        opts.write_parser().termSortOverloadingWasSetByUser = true;
        break;
      }
      case OptionEnum::THEORYOF_MODE:
      {
        auto value = stringToTheoryOfMode(optionarg);
        opts.write_theory().theoryOfMode = value;
        opts.write_theory().theoryOfModeWasSetByUser = true;
        break;
      }
      case OptionEnum::TIMEOUT_CORE_TIMEOUT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_smt().toCoreTimeout = value;
        opts.write_smt().toCoreTimeoutWasSetByUser = true;
        break;
      }
      case OptionEnum::TLIMIT:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_base().cumulativeMillisecondLimit = value;
        opts.write_base().cumulativeMillisecondLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::TLIMIT_PER:
      {
        auto value = handlers::handleOption<uint64_t>(name, optionarg);
        opts.write_base().perCallMillisecondLimit = value;
        opts.write_base().perCallMillisecondLimitWasSetByUser = true;
        break;
      }
      case OptionEnum::TRACE:
      {
        auto value = handlers::handleOption<std::string>(name, optionarg);
        opts.handler().enableTraceTag(name, value);
        break;
      }
      case OptionEnum::TRIGGER_ACTIVE_SEL:
      {
        auto value = stringToTriggerActiveSelMode(optionarg);
        opts.write_quantifiers().triggerActiveSelMode = value;
        opts.write_quantifiers().triggerActiveSelModeWasSetByUser = true;
        break;
      }
      case OptionEnum::TRIGGER_SEL:
      {
        auto value = stringToTriggerSelMode(optionarg);
        opts.write_quantifiers().triggerSelMode = value;
        opts.write_quantifiers().triggerSelModeWasSetByUser = true;
        break;
      }
      case OptionEnum::TYPE_CHECKING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_expr().typeChecking = value;
        opts.write_expr().typeCheckingWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_CARD_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufCardExp = value;
        opts.write_uf().ufCardExpWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_HO_EXP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufHoExp = value;
        opts.write_uf().ufHoExpWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_HO_EXT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufHoExt = value;
        opts.write_uf().ufHoExtWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_LAMBDA_QE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufHoLambdaQe = value;
        opts.write_uf().ufHoLambdaQeWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_LAZY_LL:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufHoLazyLambdaLift = value;
        opts.write_uf().ufHoLazyLambdaLiftWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_SS:
      {
        auto value = stringToUfssMode(optionarg);
        opts.write_uf().ufssMode = value;
        opts.write_uf().ufssModeWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_SS_ABORT_CARD:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.write_uf().ufssAbortCardinality = value;
        opts.write_uf().ufssAbortCardinalityWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_SS_FAIR:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufssFairness = value;
        opts.write_uf().ufssFairnessWasSetByUser = true;
        break;
      }
      case OptionEnum::UF_SS_FAIR_MONOTONE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_uf().ufssFairnessMonotone = value;
        opts.write_uf().ufssFairnessMonotoneWasSetByUser = true;
        break;
      }
      case OptionEnum::UNATE_LEMMAS:
      {
        auto value = stringToArithUnateLemmaMode(optionarg);
        opts.write_arith().arithUnateLemmaMode = value;
        opts.write_arith().arithUnateLemmaModeWasSetByUser = true;
        break;
      }
      case OptionEnum::UNCONSTRAINED_SIMP:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_smt().unconstrainedSimp = value;
        opts.write_smt().unconstrainedSimpWasSetByUser = true;
        break;
      }
      case OptionEnum::UNSAT_CORES_MODE:
      {
        auto value = stringToUnsatCoresMode(optionarg);
        opts.write_smt().unsatCoresMode = value;
        opts.write_smt().unsatCoresModeWasSetByUser = true;
        break;
      }
      case OptionEnum::USE_APPROX:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().useApprox = value;
        opts.write_arith().useApproxWasSetByUser = true;
        break;
      }
      case OptionEnum::USE_FCSIMPLEX:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().useFC = value;
        opts.write_arith().useFCWasSetByUser = true;
        break;
      }
      case OptionEnum::USE_PORTFOLIO:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_driver().usePortfolio = value;
        opts.write_driver().usePortfolioWasSetByUser = true;
        break;
      }
      case OptionEnum::USE_SOI:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_arith().useSOI = value;
        opts.write_arith().useSOIWasSetByUser = true;
        break;
      }
      case OptionEnum::USER_PAT:
      {
        auto value = stringToUserPatMode(optionarg);
        opts.write_quantifiers().userPatternsQuant = value;
        opts.write_quantifiers().userPatternsQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::USER_POOL:
      {
        auto value = stringToUserPoolMode(optionarg);
        opts.write_quantifiers().userPoolQuant = value;
        opts.write_quantifiers().userPoolQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::VAR_ELIM_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().varElimQuant = value;
        opts.write_quantifiers().varElimQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::VAR_INEQ_ELIM_QUANT:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_quantifiers().varIneqElimQuant = value;
        opts.write_quantifiers().varIneqElimQuantWasSetByUser = true;
        break;
      }
      case OptionEnum::VERBOSE:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().increaseVerbosity(name, value);
        break;
      }
      case OptionEnum::VERBOSITY:
      {
        auto value = handlers::handleOption<int64_t>(name, optionarg);
        opts.handler().setVerbosity(name, value);
        opts.write_base().verbosity = value;
        opts.write_base().verbosityWasSetByUser = true;
        break;
      }
      case OptionEnum::VERSION:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.handler().showVersion(name, value);
        opts.write_driver().showVersion = value;
        opts.write_driver().showVersionWasSetByUser = true;
        break;
      }
      case OptionEnum::WF_CHECKING:
      {
        auto value = handlers::handleOption<bool>(name, optionarg);
        opts.write_expr().wellFormedChecking = value;
        opts.write_expr().wellFormedCheckingWasSetByUser = true;
        break;
      }
      case OptionEnum::WRITE_PARTITIONS_TO:
      {
        auto value = handlers::handleOption<ManagedOut>(name, optionarg);
        opts.write_parallel().partitionsOut = value;
        opts.write_parallel().partitionsOutWasSetByUser = true;
        break;
      }
    }
    // clang-format on
  }

#if defined(CVC5_MUZZLED) || defined(CVC5_COMPETITION_MODE)
#define DO_SEMANTIC_CHECKS_BY_DEFAULT false
#else /* CVC5_MUZZLED || CVC5_COMPETITION_MODE */
#define DO_SEMANTIC_CHECKS_BY_DEFAULT true
#endif /* CVC5_MUZZLED || CVC5_COMPETITION_MODE */

OptionInfo getInfo(const Options& opts, const std::string& name)
{
  // clang-format off
  auto it = NAME_TO_ENUM.find(name);
  if (it == NAME_TO_ENUM.end()) {
    throw OptionException("Unrecognized option key or setting: " + name);
  }
  switch (it->second) {
    case OptionEnum::ABSTRACT_VALUES:
      return OptionInfo{"abstract-values", {}, opts.smt.abstractValuesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.abstractValues}};
    case OptionEnum::ACKERMANN:
      return OptionInfo{"ackermann", {}, opts.smt.ackermannWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.ackermann}};
    case OptionEnum::APPEND_LEARNED_LITERALS_TO_CUBES:
      return OptionInfo{"append-learned-literals-to-cubes", {}, opts.parallel.appendLearnedLiteralsToCubesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.parallel.appendLearnedLiteralsToCubes}};
    case OptionEnum::APPROX_BRANCH_DEPTH:
      return OptionInfo{"approx-branch-depth", {}, opts.arith.maxApproxDepthWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{200, opts.arith.maxApproxDepth, {}, {}}};
    case OptionEnum::ARITH_BRAB:
      return OptionInfo{"arith-brab", {}, opts.arith.brabTestWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arith.brabTest}};
    case OptionEnum::ARITH_EQ_SOLVER:
      return OptionInfo{"arith-eq-solver", {}, opts.arith.arithEqSolverWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.arithEqSolver}};
    case OptionEnum::ARITH_EXP:
      return OptionInfo{"arith-exp", {}, opts.arith.arithExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.arith.arithExp}};
    case OptionEnum::ARITH_PROP:
      return OptionInfo{"arith-prop", {}, opts.arith.arithPropagationModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"both", opts.arith.arithPropagationMode, { "bi", "both", "none", "unate" }}};
    case OptionEnum::ARITH_PROP_CLAUSES:
      return OptionInfo{"arith-prop-clauses", {}, opts.arith.arithPropAsLemmaLengthWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{8, opts.arith.arithPropAsLemmaLength, {}, {}}};
    case OptionEnum::ARITH_REWRITE_EQUALITIES:
      return OptionInfo{"arith-rewrite-equalities", {}, opts.arith.arithRewriteEqWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.arith.arithRewriteEq}};
    case OptionEnum::ARITH_STATIC_LEARNING:
      return OptionInfo{"arith-static-learning", {}, opts.arith.arithStaticLearningWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arith.arithStaticLearning}};
    case OptionEnum::ARRAYS_EAGER_INDEX:
      return OptionInfo{"arrays-eager-index", {}, opts.arrays.arraysEagerIndexSplittingWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arrays.arraysEagerIndexSplitting}};
    case OptionEnum::ARRAYS_EAGER_LEMMAS:
      return OptionInfo{"arrays-eager-lemmas", {}, opts.arrays.arraysEagerLemmasWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arrays.arraysEagerLemmas}};
    case OptionEnum::ARRAYS_EXP:
      return OptionInfo{"arrays-exp", {}, opts.arrays.arraysExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arrays.arraysExp}};
    case OptionEnum::ARRAYS_OPTIMIZE_LINEAR:
      return OptionInfo{"arrays-optimize-linear", {}, opts.arrays.arraysOptimizeLinearWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.arrays.arraysOptimizeLinear}};
    case OptionEnum::ARRAYS_PROP:
      return OptionInfo{"arrays-prop", {}, opts.arrays.arraysPropagateWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{2, opts.arrays.arraysPropagate, {}, {}}};
    case OptionEnum::ARRAYS_REDUCE_SHARING:
      return OptionInfo{"arrays-reduce-sharing", {}, opts.arrays.arraysReduceSharingWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arrays.arraysReduceSharing}};
    case OptionEnum::ARRAYS_WEAK_EQUIV:
      return OptionInfo{"arrays-weak-equiv", {}, opts.arrays.arraysWeakEquivalenceWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arrays.arraysWeakEquivalence}};
    case OptionEnum::ASSIGN_FUNCTION_VALUES:
      return OptionInfo{"assign-function-values", {}, opts.theory.assignFunctionValuesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.theory.assignFunctionValues}};
    case OptionEnum::BAGS:
      return OptionInfo{"bags", {}, opts.bags.bagsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.bags.bags}};
    case OptionEnum::BITBLAST:
      return OptionInfo{"bitblast", {}, opts.bv.bitblastModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"lazy", opts.bv.bitblastMode, { "eager", "lazy" }}};
    case OptionEnum::BITWISE_EQ:
      return OptionInfo{"bitwise-eq", {}, opts.bv.bitwiseEqWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.bv.bitwiseEq}};
    case OptionEnum::BOOL_TO_BV:
      return OptionInfo{"bool-to-bv", {}, opts.bv.boolToBitvectorWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"off", opts.bv.boolToBitvector, { "all", "ite", "off" }}};
    case OptionEnum::BV_ASSERT_INPUT:
      return OptionInfo{"bv-assert-input", {}, opts.bv.bvAssertInputWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.bv.bvAssertInput}};
    case OptionEnum::BV_EAGER_EVAL:
      return OptionInfo{"bv-eager-eval", {}, opts.bv.bvEagerEvalWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.bv.bvEagerEval}};
    case OptionEnum::BV_EQ_ENGINE:
      return OptionInfo{"bv-eq-engine", {}, opts.bv.bvEqEngineWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.bv.bvEqEngine}};
    case OptionEnum::BV_GAUSS_ELIM:
      return OptionInfo{"bv-gauss-elim", {}, opts.bv.bvGaussElimWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.bv.bvGaussElim}};
    case OptionEnum::BV_INTRO_POW2:
      return OptionInfo{"bv-intro-pow2", {}, opts.bv.bvIntroducePow2WasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.bv.bvIntroducePow2}};
    case OptionEnum::BV_PRINT_CONSTS_AS_INDEXED_SYMBOLS:
      return OptionInfo{"bv-print-consts-as-indexed-symbols", {}, opts.printer.bvPrintConstsAsIndexedSymbolsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.printer.bvPrintConstsAsIndexedSymbols}};
    case OptionEnum::BV_PROPAGATE:
      return OptionInfo{"bv-propagate", {}, opts.bv.bitvectorPropagateWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.bv.bitvectorPropagate}};
    case OptionEnum::BV_RW_EXTEND_EQ:
      return OptionInfo{"bv-rw-extend-eq", {}, opts.bv.rwExtendEqWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.bv.rwExtendEq}};
    case OptionEnum::BV_SAT_SOLVER:
      return OptionInfo{"bv-sat-solver", {}, opts.bv.bvSatSolverWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"cadical", opts.bv.bvSatSolver, { "cadical", "cryptominisat", "kissat", "minisat" }}};
    case OptionEnum::BV_SOLVER:
      return OptionInfo{"bv-solver", {}, opts.bv.bvSolverWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"bitblast", opts.bv.bvSolver, { "bitblast", "bitblast-internal" }}};
    case OptionEnum::BV_TO_BOOL:
      return OptionInfo{"bv-to-bool", {}, opts.bv.bitvectorToBoolWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.bv.bitvectorToBool}};
    case OptionEnum::BV_TO_INT_USE_POW2:
      return OptionInfo{"bv-to-int-use-pow2", {}, opts.smt.bvToIntUsePow2WasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.bvToIntUsePow2}};
    case OptionEnum::BVAND_INTEGER_GRANULARITY:
      return OptionInfo{"bvand-integer-granularity", {}, opts.smt.BVAndIntegerGranularityWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{1, opts.smt.BVAndIntegerGranularity, 1, 8}};
    case OptionEnum::CBQI:
      return OptionInfo{"cbqi", {}, opts.quantifiers.conflictBasedInstWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.conflictBasedInst}};
    case OptionEnum::CBQI_ALL_CONFLICT:
      return OptionInfo{"cbqi-all-conflict", {}, opts.quantifiers.cbqiAllConflictWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cbqiAllConflict}};
    case OptionEnum::CBQI_MODE:
      return OptionInfo{"cbqi-mode", {}, opts.quantifiers.cbqiModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"prop-eq", opts.quantifiers.cbqiMode, { "conflict", "prop-eq" }}};
    case OptionEnum::CBQI_SKIP_RD:
      return OptionInfo{"cbqi-skip-rd", {}, opts.quantifiers.cbqiSkipRdWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cbqiSkipRd}};
    case OptionEnum::CBQI_TCONSTRAINT:
      return OptionInfo{"cbqi-tconstraint", {}, opts.quantifiers.cbqiTConstraintWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cbqiTConstraint}};
    case OptionEnum::CBQI_VO_EXP:
      return OptionInfo{"cbqi-vo-exp", {}, opts.quantifiers.cbqiVoExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cbqiVoExp}};
    case OptionEnum::CDT_BISIMILAR:
      return OptionInfo{"cdt-bisimilar", {}, opts.datatypes.cdtBisimilarWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.datatypes.cdtBisimilar}};
    case OptionEnum::CEGIS_SAMPLE:
      return OptionInfo{"cegis-sample", {}, opts.quantifiers.cegisSampleWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"none", opts.quantifiers.cegisSample, { "none", "trust", "use" }}};
    case OptionEnum::CEGQI:
      return OptionInfo{"cegqi", {}, opts.quantifiers.cegqiWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqi}};
    case OptionEnum::CEGQI_ALL:
      return OptionInfo{"cegqi-all", {}, opts.quantifiers.cegqiAllWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiAll}};
    case OptionEnum::CEGQI_BV:
      return OptionInfo{"cegqi-bv", {}, opts.quantifiers.cegqiBvWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.cegqiBv}};
    case OptionEnum::CEGQI_BV_CONCAT_INV:
      return OptionInfo{"cegqi-bv-concat-inv", {}, opts.quantifiers.cegqiBvConcInvWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.cegqiBvConcInv}};
    case OptionEnum::CEGQI_BV_INEQ:
      return OptionInfo{"cegqi-bv-ineq", {}, opts.quantifiers.cegqiBvIneqModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"eq-boundary", opts.quantifiers.cegqiBvIneqMode, { "eq-boundary", "eq-slack", "keep" }}};
    case OptionEnum::CEGQI_BV_INTERLEAVE_VALUE:
      return OptionInfo{"cegqi-bv-interleave-value", {}, opts.quantifiers.cegqiBvInterleaveValueWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiBvInterleaveValue}};
    case OptionEnum::CEGQI_BV_LINEAR:
      return OptionInfo{"cegqi-bv-linear", {}, opts.quantifiers.cegqiBvLinearizeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.cegqiBvLinearize}};
    case OptionEnum::CEGQI_BV_RM_EXTRACT:
      return OptionInfo{"cegqi-bv-rm-extract", {}, opts.quantifiers.cegqiBvRmExtractWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.cegqiBvRmExtract}};
    case OptionEnum::CEGQI_BV_SOLVE_NL:
      return OptionInfo{"cegqi-bv-solve-nl", {}, opts.quantifiers.cegqiBvSolveNlWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiBvSolveNl}};
    case OptionEnum::CEGQI_FULL:
      return OptionInfo{"cegqi-full", {}, opts.quantifiers.cegqiFullEffortWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiFullEffort}};
    case OptionEnum::CEGQI_INF_INT:
      return OptionInfo{"cegqi-inf-int", {}, opts.quantifiers.cegqiUseInfIntWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiUseInfInt}};
    case OptionEnum::CEGQI_INF_REAL:
      return OptionInfo{"cegqi-inf-real", {}, opts.quantifiers.cegqiUseInfRealWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiUseInfReal}};
    case OptionEnum::CEGQI_INNERMOST:
      return OptionInfo{"cegqi-innermost", {}, opts.quantifiers.cegqiInnermostWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.cegqiInnermost}};
    case OptionEnum::CEGQI_MIDPOINT:
      return OptionInfo{"cegqi-midpoint", {}, opts.quantifiers.cegqiMidpointWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiMidpoint}};
    case OptionEnum::CEGQI_MIN_BOUNDS:
      return OptionInfo{"cegqi-min-bounds", {}, opts.quantifiers.cegqiMinBoundsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiMinBounds}};
    case OptionEnum::CEGQI_MULTI_INST:
      return OptionInfo{"cegqi-multi-inst", {}, opts.quantifiers.cegqiMultiInstWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiMultiInst}};
    case OptionEnum::CEGQI_NESTED_QE:
      return OptionInfo{"cegqi-nested-qe", {}, opts.quantifiers.cegqiNestedQEWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiNestedQE}};
    case OptionEnum::CEGQI_NOPT:
      return OptionInfo{"cegqi-nopt", {}, opts.quantifiers.cegqiNoptWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.cegqiNopt}};
    case OptionEnum::CEGQI_ROUND_UP_LIA:
      return OptionInfo{"cegqi-round-up-lia", {}, opts.quantifiers.cegqiRoundUpLowerLiaWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiRoundUpLowerLia}};
    case OptionEnum::CHECK_ABDUCTS:
      return OptionInfo{"check-abducts", {}, opts.smt.checkAbductsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.checkAbducts}};
    case OptionEnum::CHECK_INTERPOLANTS:
      return OptionInfo{"check-interpolants", {}, opts.smt.checkInterpolantsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.checkInterpolants}};
    case OptionEnum::CHECK_MODELS:
      return OptionInfo{"check-models", {}, opts.smt.checkModelsWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.smt.checkModels}};
    case OptionEnum::CHECK_PROOF_STEPS:
      return OptionInfo{"check-proof-steps", {}, opts.proof.checkProofStepsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.checkProofSteps}};
    case OptionEnum::CHECK_PROOFS:
      return OptionInfo{"check-proofs", {}, opts.smt.checkProofsWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.smt.checkProofs}};
    case OptionEnum::CHECK_SYNTH_SOL:
      return OptionInfo{"check-synth-sol", {}, opts.smt.checkSynthSolWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.checkSynthSol}};
    case OptionEnum::CHECK_UNSAT_CORES:
      return OptionInfo{"check-unsat-cores", {}, opts.smt.checkUnsatCoresWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.checkUnsatCores}};
    case OptionEnum::CHECKS_BEFORE_PARTITION:
      return OptionInfo{"checks-before-partition", {}, opts.parallel.checksBeforePartitioningWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{1, opts.parallel.checksBeforePartitioning, {}, {}}};
    case OptionEnum::CHECKS_BETWEEN_PARTITIONS:
      return OptionInfo{"checks-between-partitions", {}, opts.parallel.checksBetweenPartitionsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{1, opts.parallel.checksBetweenPartitions, {}, {}}};
    case OptionEnum::COLLECT_PIVOT_STATS:
      return OptionInfo{"collect-pivot-stats", {}, opts.arith.collectPivotsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.collectPivots}};
    case OptionEnum::COMPUTE_PARTITIONS:
      return OptionInfo{"compute-partitions", {}, opts.parallel.computePartitionsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{0, opts.parallel.computePartitions, {}, {}}};
    case OptionEnum::COND_VAR_SPLIT_QUANT:
      return OptionInfo{"cond-var-split-quant", {}, opts.quantifiers.condVarSplitQuantWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"on", opts.quantifiers.condVarSplitQuant, { "agg", "off", "on" }}};
    case OptionEnum::CONDENSE_FUNCTION_VALUES:
      return OptionInfo{"condense-function-values", {}, opts.theory.condenseFunctionValuesWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{true, opts.theory.condenseFunctionValues}};
    case OptionEnum::CONFLICT_PROCESS:
      return OptionInfo{"conflict-process", {}, opts.theory.conflictProcessModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.theory.conflictProcessMode, { "min", "min-ext", "none" }}};
    case OptionEnum::CONJECTURE_GEN:
      return OptionInfo{"conjecture-gen", {}, opts.quantifiers.conjectureGenWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.conjectureGen}};
    case OptionEnum::CONJECTURE_GEN_GT_ENUM:
      return OptionInfo{"conjecture-gen-gt-enum", {}, opts.quantifiers.conjectureGenGtEnumWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{50, opts.quantifiers.conjectureGenGtEnum, {}, {}}};
    case OptionEnum::CONJECTURE_GEN_MAX_DEPTH:
      return OptionInfo{"conjecture-gen-max-depth", {}, opts.quantifiers.conjectureGenMaxDepthWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{3, opts.quantifiers.conjectureGenMaxDepth, {}, {}}};
    case OptionEnum::CONJECTURE_GEN_PER_ROUND:
      return OptionInfo{"conjecture-gen-per-round", {}, opts.quantifiers.conjectureGenPerRoundWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{1, opts.quantifiers.conjectureGenPerRound, {}, {}}};
    case OptionEnum::CONS_EXP_TRIGGERS:
      return OptionInfo{"cons-exp-triggers", {}, opts.quantifiers.consExpandTriggersWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.consExpandTriggers}};
    case OptionEnum::COPYRIGHT:
      return OptionInfo{"copyright", {}, opts.driver.showCopyrightWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.driver.showCopyright}};
    case OptionEnum::CUT_ALL_BOUNDED:
      return OptionInfo{"cut-all-bounded", {}, opts.arith.doCutAllBoundedWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.doCutAllBounded}};
    case OptionEnum::DAG_THRESH:
      return OptionInfo{"dag-thresh", {}, opts.printer.dagThreshWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<int64_t>{1, opts.printer.dagThresh, 0, {}}};
    case OptionEnum::DATATYPES_EXP:
      return OptionInfo{"datatypes-exp", {}, opts.datatypes.datatypesExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.datatypes.datatypesExp}};
    case OptionEnum::DEBUG_CHECK_MODELS:
      return OptionInfo{"debug-check-models", {}, opts.smt.debugCheckModelsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.debugCheckModels}};
    case OptionEnum::DECISION:
      return OptionInfo{"decision", {"decision-mode"}, opts.decision.decisionModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"internal", opts.decision.decisionMode, { "internal", "justification", "stoponly" }}};
    case OptionEnum::DEEP_RESTART:
      return OptionInfo{"deep-restart", {}, opts.smt.deepRestartModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.smt.deepRestartMode, { "all", "input", "input-and-prop", "input-and-solvable", "none" }}};
    case OptionEnum::DEEP_RESTART_FACTOR:
      return OptionInfo{"deep-restart-factor", {}, opts.smt.deepRestartFactorWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<double>{3.0, opts.smt.deepRestartFactor, 0.0, 1000.0}};
    case OptionEnum::DEFAULT_FUNCTION_VALUE_MODE:
      return OptionInfo{"default-function-value-mode", {}, opts.theory.defaultFunctionValueModeWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ModeInfo{"first", opts.theory.defaultFunctionValueMode, { "first", "first-enum", "hole" }}};
    case OptionEnum::DIFFICULTY_MODE:
      return OptionInfo{"difficulty-mode", {}, opts.smt.difficultyModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"lemma-literal-all", opts.smt.difficultyMode, { "lemma-literal", "lemma-literal-all", "model-check" }}};
    case OptionEnum::DIO_DECOMPS:
      return OptionInfo{"dio-decomps", {}, opts.arith.exportDioDecompositionsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.exportDioDecompositions}};
    case OptionEnum::DIO_SOLVER:
      return OptionInfo{"dio-solver", {}, opts.arith.arithDioSolverWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.arith.arithDioSolver}};
    case OptionEnum::DIO_TURNS:
      return OptionInfo{"dio-turns", {}, opts.arith.dioSolverTurnsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{10, opts.arith.dioSolverTurns, {}, {}}};
    case OptionEnum::DRAT_BINARY_FORMAT:
      return OptionInfo{"drat-binary-format", {}, opts.proof.dratBinaryFormatWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.dratBinaryFormat}};
    case OptionEnum::DT_BINARY_SPLIT:
      return OptionInfo{"dt-binary-split", {}, opts.datatypes.dtBinarySplitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.datatypes.dtBinarySplit}};
    case OptionEnum::DT_BLAST_SPLITS:
      return OptionInfo{"dt-blast-splits", {}, opts.datatypes.dtBlastSplitsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.datatypes.dtBlastSplits}};
    case OptionEnum::DT_CYCLIC:
      return OptionInfo{"dt-cyclic", {}, opts.datatypes.dtCyclicWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.datatypes.dtCyclic}};
    case OptionEnum::DT_INFER_AS_LEMMAS:
      return OptionInfo{"dt-infer-as-lemmas", {}, opts.datatypes.dtInferAsLemmasWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.datatypes.dtInferAsLemmas}};
    case OptionEnum::DT_NESTED_REC:
      return OptionInfo{"dt-nested-rec", {}, opts.datatypes.dtNestedRecWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.datatypes.dtNestedRec}};
    case OptionEnum::DT_POLITE_OPTIMIZE:
      return OptionInfo{"dt-polite-optimize", {}, opts.datatypes.dtPoliteOptimizeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.datatypes.dtPoliteOptimize}};
    case OptionEnum::DT_SHARE_SEL:
      return OptionInfo{"dt-share-sel", {}, opts.datatypes.dtSharedSelectorsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.datatypes.dtSharedSelectors}};
    case OptionEnum::DT_STC_IND:
      return OptionInfo{"dt-stc-ind", {}, opts.quantifiers.dtStcInductionWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.dtStcInduction}};
    case OptionEnum::DT_VAR_EXP_QUANT:
      return OptionInfo{"dt-var-exp-quant", {}, opts.quantifiers.dtVarExpandQuantWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.dtVarExpandQuant}};
    case OptionEnum::DUMP_DIFFICULTY:
      return OptionInfo{"dump-difficulty", {}, opts.driver.dumpDifficultyWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.driver.dumpDifficulty}};
    case OptionEnum::DUMP_INSTANTIATIONS:
      return OptionInfo{"dump-instantiations", {}, opts.driver.dumpInstantiationsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.driver.dumpInstantiations}};
    case OptionEnum::DUMP_INSTANTIATIONS_DEBUG:
      return OptionInfo{"dump-instantiations-debug", {}, opts.driver.dumpInstantiationsDebugWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.driver.dumpInstantiationsDebug}};
    case OptionEnum::DUMP_MODELS:
      return OptionInfo{"dump-models", {}, opts.driver.dumpModelsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.driver.dumpModels}};
    case OptionEnum::DUMP_PROOFS:
      return OptionInfo{"dump-proofs", {}, opts.driver.dumpProofsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.driver.dumpProofs}};
    case OptionEnum::DUMP_UNSAT_CORES:
      return OptionInfo{"dump-unsat-cores", {}, opts.driver.dumpUnsatCoresWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.driver.dumpUnsatCores}};
    case OptionEnum::DUMP_UNSAT_CORES_LEMMAS:
      return OptionInfo{"dump-unsat-cores-lemmas", {}, opts.driver.dumpUnsatCoresLemmasWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.driver.dumpUnsatCoresLemmas}};
    case OptionEnum::E_MATCHING:
      return OptionInfo{"e-matching", {}, opts.quantifiers.eMatchingWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.eMatching}};
    case OptionEnum::EAGER_ARITH_BV_CONV:
      return OptionInfo{"eager-arith-bv-conv", {}, opts.uf.eagerArithBvConvWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.uf.eagerArithBvConv}};
    case OptionEnum::EARLY_EXIT:
      return OptionInfo{"early-exit", {}, opts.driver.earlyExitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.driver.earlyExit}};
    case OptionEnum::EARLY_ITE_REMOVAL:
      return OptionInfo{"early-ite-removal", {}, opts.smt.earlyIteRemovalWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.earlyIteRemoval}};
    case OptionEnum::EE_MODE:
      return OptionInfo{"ee-mode", {}, opts.theory.eeModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"distributed", opts.theory.eeMode, { "central", "distributed" }}};
    case OptionEnum::ELIM_TAUT_QUANT:
      return OptionInfo{"elim-taut-quant", {}, opts.quantifiers.elimTautQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.elimTautQuant}};
    case OptionEnum::ENUM_INST:
      return OptionInfo{"enum-inst", {}, opts.quantifiers.enumInstWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.enumInst}};
    case OptionEnum::ENUM_INST_INTERLEAVE:
      return OptionInfo{"enum-inst-interleave", {}, opts.quantifiers.enumInstInterleaveWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.enumInstInterleave}};
    case OptionEnum::ENUM_INST_LIMIT:
      return OptionInfo{"enum-inst-limit", {}, opts.quantifiers.enumInstLimitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{-1, opts.quantifiers.enumInstLimit, {}, {}}};
    case OptionEnum::ENUM_INST_RD:
      return OptionInfo{"enum-inst-rd", {}, opts.quantifiers.enumInstRdWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.enumInstRd}};
    case OptionEnum::ENUM_INST_STRATIFY:
      return OptionInfo{"enum-inst-stratify", {}, opts.quantifiers.enumInstStratifyWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.enumInstStratify}};
    case OptionEnum::ENUM_INST_SUM:
      return OptionInfo{"enum-inst-sum", {}, opts.quantifiers.enumInstSumWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.enumInstSum}};
    case OptionEnum::ERR:
      return OptionInfo{"err", {"diagnostic-output-channel"}, opts.base.errWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::VoidInfo{}};
    case OptionEnum::ERROR_SELECTION_RULE:
      return OptionInfo{"error-selection-rule", {}, opts.arith.arithErrorSelectionRuleWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"min", opts.arith.arithErrorSelectionRule, { "max", "min", "sum", "varord" }}};
    case OptionEnum::EXPR_DEPTH:
      return OptionInfo{"expr-depth", {}, opts.printer.nodeDepthWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<int64_t>{-1, opts.printer.nodeDepth, -1, {}}};
    case OptionEnum::EXT_REW_PREP:
      return OptionInfo{"ext-rew-prep", {}, opts.smt.extRewPrepWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"off", opts.smt.extRewPrep, { "agg", "off", "use" }}};
    case OptionEnum::EXT_REWRITE_QUANT:
      return OptionInfo{"ext-rewrite-quant", {}, opts.quantifiers.extRewriteQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.extRewriteQuant}};
    case OptionEnum::FC_PENALTIES:
      return OptionInfo{"fc-penalties", {}, opts.arith.havePenaltiesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.havePenalties}};
    case OptionEnum::FF:
      return OptionInfo{"ff", {}, opts.ff.ffWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.ff.ff}};
    case OptionEnum::FF_BITSUM:
      return OptionInfo{"ff-bitsum", {}, opts.ff.ffBitsumWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.ff.ffBitsum}};
    case OptionEnum::FF_ELIM_DISJUNCTIVE_BIT:
      return OptionInfo{"ff-elim-disjunctive-bit", {}, opts.ff.ffElimDisjunctiveBitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.ff.ffElimDisjunctiveBit}};
    case OptionEnum::FF_FIELD_POLYS:
      return OptionInfo{"ff-field-polys", {}, opts.ff.ffFieldPolysWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.ff.ffFieldPolys}};
    case OptionEnum::FF_SOLVER:
      return OptionInfo{"ff-solver", {}, opts.ff.ffSolverWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"gb", opts.ff.ffSolver, { "gb", "split" }}};
    case OptionEnum::FF_TRACE_GB:
      return OptionInfo{"ff-trace-gb", {}, opts.ff.ffTraceGbWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.ff.ffTraceGb}};
    case OptionEnum::FILENAME:
      return OptionInfo{"filename", {}, opts.driver.filenameWasSetByUser, OptionInfo::Category::UNDOCUMENTED, OptionInfo::ValueInfo<std::string>{"", opts.driver.filename}};
    case OptionEnum::FILESYSTEM_ACCESS:
      return OptionInfo{"filesystem-access", {}, opts.parser.filesystemAccessWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.parser.filesystemAccess}};
    case OptionEnum::FINITE_MODEL_FIND:
      return OptionInfo{"finite-model-find", {}, opts.quantifiers.finiteModelFindWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.finiteModelFind}};
    case OptionEnum::FLATTEN_HO_CHAINS:
      return OptionInfo{"flatten-ho-chains", {}, opts.printer.flattenHOChainsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.printer.flattenHOChains}};
    case OptionEnum::FMF_BOUND:
      return OptionInfo{"fmf-bound", {}, opts.quantifiers.fmfBoundWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fmfBound}};
    case OptionEnum::FMF_BOUND_BLAST:
      return OptionInfo{"fmf-bound-blast", {}, opts.quantifiers.fmfBoundBlastWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fmfBoundBlast}};
    case OptionEnum::FMF_BOUND_LAZY:
      return OptionInfo{"fmf-bound-lazy", {}, opts.quantifiers.fmfBoundLazyWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fmfBoundLazy}};
    case OptionEnum::FMF_FUN:
      return OptionInfo{"fmf-fun", {}, opts.quantifiers.fmfFunWellDefinedWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fmfFunWellDefined}};
    case OptionEnum::FMF_FUN_RLV:
      return OptionInfo{"fmf-fun-rlv", {}, opts.quantifiers.fmfFunWellDefinedRelevantWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fmfFunWellDefinedRelevant}};
    case OptionEnum::FMF_MBQI:
      return OptionInfo{"fmf-mbqi", {}, opts.quantifiers.fmfMbqiModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"fmc", opts.quantifiers.fmfMbqiMode, { "fmc", "none", "trust" }}};
    case OptionEnum::FMF_TYPE_COMPLETION_THRESH:
      return OptionInfo{"fmf-type-completion-thresh", {}, opts.quantifiers.fmfTypeCompletionThreshWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<int64_t>{1000, opts.quantifiers.fmfTypeCompletionThresh, {}, {}}};
    case OptionEnum::FORCE_LOGIC:
      return OptionInfo{"force-logic", {}, opts.parser.forceLogicStringWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<std::string>{"", opts.parser.forceLogicString}};
    case OptionEnum::FORCE_NO_LIMIT_CPU_WHILE_DUMP:
      return OptionInfo{"force-no-limit-cpu-while-dump", {}, opts.driver.forceNoLimitCpuWhileDumpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.driver.forceNoLimitCpuWhileDump}};
    case OptionEnum::FOREIGN_THEORY_REWRITE:
      return OptionInfo{"foreign-theory-rewrite", {}, opts.smt.foreignTheoryRewriteWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.foreignTheoryRewrite}};
    case OptionEnum::FP:
      return OptionInfo{"fp", {}, opts.fp.fpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.fp.fp}};
    case OptionEnum::FP_EXP:
      return OptionInfo{"fp-exp", {}, opts.fp.fpExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.fp.fpExp}};
    case OptionEnum::FP_LAZY_WB:
      return OptionInfo{"fp-lazy-wb", {}, opts.fp.fpLazyWbWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.fp.fpLazyWb}};
    case OptionEnum::FRESH_BINDERS:
      return OptionInfo{"fresh-binders", {}, opts.parser.freshBindersWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.parser.freshBinders}};
    case OptionEnum::FRESH_DECLARATIONS:
      return OptionInfo{"fresh-declarations", {}, opts.parser.freshDeclarationsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.parser.freshDeclarations}};
    case OptionEnum::FULL_SATURATE_QUANT:
      return OptionInfo{"full-saturate-quant", {}, opts.quantifiers.fullSaturateQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fullSaturateQuant}};
    case OptionEnum::FULL_SYGUS_VERIFY:
      return OptionInfo{"full-sygus-verify", {}, opts.quantifiers.fullSygusVerifyWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fullSygusVerify}};
    case OptionEnum::GLOBAL_DECLARATIONS:
      return OptionInfo{"global-declarations", {}, opts.parser.globalDeclarationsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.parser.globalDeclarations}};
    case OptionEnum::GLOBAL_NEGATE:
      return OptionInfo{"global-negate", {}, opts.quantifiers.globalNegateWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.globalNegate}};
    case OptionEnum::HELP:
      return OptionInfo{"help", {}, opts.driver.helpWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.driver.help}};
    case OptionEnum::HELP_REGULAR:
      return OptionInfo{"help-regular", {}, opts.driver.helpRegularWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.driver.helpRegular}};
    case OptionEnum::HEURISTIC_PIVOTS:
      return OptionInfo{"heuristic-pivots", {}, opts.arith.arithHeuristicPivotsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{0, opts.arith.arithHeuristicPivots, {}, {}}};
    case OptionEnum::HO_ELIM:
      return OptionInfo{"ho-elim", {}, opts.quantifiers.hoElimWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.hoElim}};
    case OptionEnum::HO_ELIM_STORE_AX:
      return OptionInfo{"ho-elim-store-ax", {}, opts.quantifiers.hoElimStoreAxWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.hoElimStoreAx}};
    case OptionEnum::HO_MATCHING:
      return OptionInfo{"ho-matching", {}, opts.quantifiers.hoMatchingWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.hoMatching}};
    case OptionEnum::HO_MERGE_TERM_DB:
      return OptionInfo{"ho-merge-term-db", {}, opts.quantifiers.hoMergeTermDbWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.hoMergeTermDb}};
    case OptionEnum::IAND_MODE:
      return OptionInfo{"iand-mode", {}, opts.smt.iandModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"value", opts.smt.iandMode, { "bitwise", "sum", "value" }}};
    case OptionEnum::IEVAL:
      return OptionInfo{"ieval", {}, opts.quantifiers.ievalModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"use", opts.quantifiers.ievalMode, { "off", "use", "use-learn" }}};
    case OptionEnum::IN:
      return OptionInfo{"in", {}, opts.base.inWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::VoidInfo{}};
    case OptionEnum::INCREMENT_TRIGGERS:
      return OptionInfo{"increment-triggers", {}, opts.quantifiers.incrementTriggersWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.incrementTriggers}};
    case OptionEnum::INCREMENTAL:
      return OptionInfo{"incremental", {}, opts.base.incrementalSolvingWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{true, opts.base.incrementalSolving}};
    case OptionEnum::INST_MAX_LEVEL:
      return OptionInfo{"inst-max-level", {}, opts.quantifiers.instMaxLevelWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{-1, opts.quantifiers.instMaxLevel, {}, {}}};
    case OptionEnum::INST_MAX_ROUNDS:
      return OptionInfo{"inst-max-rounds", {}, opts.quantifiers.instMaxRoundsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<int64_t>{-1, opts.quantifiers.instMaxRounds, {}, {}}};
    case OptionEnum::INST_NO_ENTAIL:
      return OptionInfo{"inst-no-entail", {}, opts.quantifiers.instNoEntailWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.instNoEntail}};
    case OptionEnum::INST_WHEN:
      return OptionInfo{"inst-when", {}, opts.quantifiers.instWhenModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"full-last-call", opts.quantifiers.instWhenMode, { "full", "full-delay", "full-delay-last-call", "full-last-call", "last-call" }}};
    case OptionEnum::INST_WHEN_PHASE:
      return OptionInfo{"inst-when-phase", {}, opts.quantifiers.instWhenPhaseWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{2, opts.quantifiers.instWhenPhase, {}, {}}};
    case OptionEnum::INT_WF_IND:
      return OptionInfo{"int-wf-ind", {}, opts.quantifiers.intWfInductionWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.intWfInduction}};
    case OptionEnum::INTERACTIVE:
      return OptionInfo{"interactive", {}, opts.driver.interactiveWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.driver.interactive}};
    case OptionEnum::INTERPOLANTS_MODE:
      return OptionInfo{"interpolants-mode", {}, opts.smt.interpolantsModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"default", opts.smt.interpolantsMode, { "all", "assumptions", "conjecture", "default", "shared" }}};
    case OptionEnum::ITE_DTT_SPLIT_QUANT:
      return OptionInfo{"ite-dtt-split-quant", {}, opts.quantifiers.iteDtTesterSplitQuantWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.iteDtTesterSplitQuant}};
    case OptionEnum::ITE_LIFT_QUANT:
      return OptionInfo{"ite-lift-quant", {}, opts.quantifiers.iteLiftQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"simple", opts.quantifiers.iteLiftQuant, { "all", "none", "simple" }}};
    case OptionEnum::ITE_SIMP:
      return OptionInfo{"ite-simp", {}, opts.smt.doITESimpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.doITESimp}};
    case OptionEnum::JH_RLV_ORDER:
      return OptionInfo{"jh-rlv-order", {}, opts.decision.jhRlvOrderWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.decision.jhRlvOrder}};
    case OptionEnum::JH_SKOLEM:
      return OptionInfo{"jh-skolem", {}, opts.decision.jhSkolemModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"first", opts.decision.jhSkolemMode, { "first", "last" }}};
    case OptionEnum::JH_SKOLEM_RLV:
      return OptionInfo{"jh-skolem-rlv", {}, opts.decision.jhSkolemRlvModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"assert", opts.decision.jhSkolemRlvMode, { "always", "assert" }}};
    case OptionEnum::LANG:
      return OptionInfo{"lang", {"input-language"}, opts.base.inputLanguageWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::VoidInfo{}};
    case OptionEnum::LEARNED_REWRITE:
      return OptionInfo{"learned-rewrite", {}, opts.smt.learnedRewriteWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.learnedRewrite}};
    case OptionEnum::LEMMA_INPROCESS:
      return OptionInfo{"lemma-inprocess", {}, opts.theory.lemmaInprocessWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.theory.lemmaInprocess, { "full", "light", "none" }}};
    case OptionEnum::LEMMA_INPROCESS_INFER_EQ_LIT:
      return OptionInfo{"lemma-inprocess-infer-eq-lit", {}, opts.theory.lemmaInprocessInferEqLitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.theory.lemmaInprocessInferEqLit}};
    case OptionEnum::LEMMA_INPROCESS_SUBS:
      return OptionInfo{"lemma-inprocess-subs", {}, opts.theory.lemmaInprocessSubsModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"simple", opts.theory.lemmaInprocessSubsMode, { "all", "simple" }}};
    case OptionEnum::LEMMAS_ON_REPLAY_FAILURE:
      return OptionInfo{"lemmas-on-replay-failure", {}, opts.arith.replayFailureLemmaWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.replayFailureLemma}};
    case OptionEnum::LFSC_EXPAND_TRUST:
      return OptionInfo{"lfsc-expand-trust", {}, opts.proof.lfscExpandTrustWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.lfscExpandTrust}};
    case OptionEnum::LFSC_FLATTEN:
      return OptionInfo{"lfsc-flatten", {}, opts.proof.lfscFlattenWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.lfscFlatten}};
    case OptionEnum::LITERAL_MATCHING:
      return OptionInfo{"literal-matching", {}, opts.quantifiers.literalMatchModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"use", opts.quantifiers.literalMatchMode, { "agg", "agg-predicate", "none", "use" }}};
    case OptionEnum::MACROS_QUANT:
      return OptionInfo{"macros-quant", {}, opts.quantifiers.macrosQuantWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.macrosQuant}};
    case OptionEnum::MACROS_QUANT_MODE:
      return OptionInfo{"macros-quant-mode", {}, opts.quantifiers.macrosQuantModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"ground-uf", opts.quantifiers.macrosQuantMode, { "all", "ground", "ground-uf" }}};
    case OptionEnum::MAXCUTSINCONTEXT:
      return OptionInfo{"maxCutsInContext", {}, opts.arith.maxCutsInContextWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{65535, opts.arith.maxCutsInContext, {}, {}}};
    case OptionEnum::MBQI:
      return OptionInfo{"mbqi", {}, opts.quantifiers.mbqiWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.mbqi}};
    case OptionEnum::MBQI_FAST_SYGUS:
      return OptionInfo{"mbqi-fast-sygus", {}, opts.quantifiers.mbqiFastSygusWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.mbqiFastSygus}};
    case OptionEnum::MBQI_INTERLEAVE:
      return OptionInfo{"mbqi-interleave", {}, opts.quantifiers.mbqiInterleaveWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.mbqiInterleave}};
    case OptionEnum::MBQI_ONE_INST_PER_ROUND:
      return OptionInfo{"mbqi-one-inst-per-round", {}, opts.quantifiers.fmfOneInstPerRoundWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.fmfOneInstPerRound}};
    case OptionEnum::MINIMAL_UNSAT_CORES:
      return OptionInfo{"minimal-unsat-cores", {}, opts.smt.minimalUnsatCoresWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.minimalUnsatCores}};
    case OptionEnum::MINISAT_DUMP_DIMACS:
      return OptionInfo{"minisat-dump-dimacs", {}, opts.prop.minisatDumpDimacsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.prop.minisatDumpDimacs}};
    case OptionEnum::MINISAT_SIMPLIFICATION:
      return OptionInfo{"minisat-simplification", {}, opts.prop.minisatSimpModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"all", opts.prop.minisatSimpMode, { "all", "clause-elim", "none" }}};
    case OptionEnum::MINISCOPE_QUANT:
      return OptionInfo{"miniscope-quant", {}, opts.quantifiers.miniscopeQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"conj-and-fv", opts.quantifiers.miniscopeQuant, { "agg", "conj", "conj-and-fv", "fv", "off" }}};
    case OptionEnum::MINISCOPE_QUANT_USER:
      return OptionInfo{"miniscope-quant-user", {}, opts.quantifiers.miniscopeQuantUserWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.miniscopeQuantUser}};
    case OptionEnum::MIPLIB_TRICK:
      return OptionInfo{"miplib-trick", {}, opts.arith.arithMLTrickWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.arithMLTrick}};
    case OptionEnum::MIPLIB_TRICK_SUBS:
      return OptionInfo{"miplib-trick-subs", {}, opts.arith.arithMLTrickSubstitutionsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{1, opts.arith.arithMLTrickSubstitutions, {}, {}}};
    case OptionEnum::MODEL_BASED_ARITH_BV_CONV:
      return OptionInfo{"model-based-arith-bv-conv", {}, opts.uf.modelBasedArithBvConvWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.uf.modelBasedArithBvConv}};
    case OptionEnum::MODEL_CORES:
      return OptionInfo{"model-cores", {}, opts.smt.modelCoresModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"none", opts.smt.modelCoresMode, { "non-implied", "none", "simple" }}};
    case OptionEnum::MODEL_U_PRINT:
      return OptionInfo{"model-u-print", {}, opts.printer.modelUninterpPrintWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"none", opts.printer.modelUninterpPrint, { "decl-fun", "decl-sort-and-fun", "dt", "none" }}};
    case OptionEnum::MODEL_VAR_ELIM_UNEVAL:
      return OptionInfo{"model-var-elim-uneval", {}, opts.smt.modelVarElimUnevalWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.smt.modelVarElimUneval}};
    case OptionEnum::MULTI_TRIGGER_CACHE:
      return OptionInfo{"multi-trigger-cache", {}, opts.quantifiers.multiTriggerCacheWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.multiTriggerCache}};
    case OptionEnum::MULTI_TRIGGER_LINEAR:
      return OptionInfo{"multi-trigger-linear", {}, opts.quantifiers.multiTriggerLinearWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.multiTriggerLinear}};
    case OptionEnum::MULTI_TRIGGER_PRIORITY:
      return OptionInfo{"multi-trigger-priority", {}, opts.quantifiers.multiTriggerPriorityWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.multiTriggerPriority}};
    case OptionEnum::MULTI_TRIGGER_WHEN_SINGLE:
      return OptionInfo{"multi-trigger-when-single", {}, opts.quantifiers.multiTriggerWhenSingleWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.multiTriggerWhenSingle}};
    case OptionEnum::NEW_PROP:
      return OptionInfo{"new-prop", {}, opts.arith.newPropWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.arith.newProp}};
    case OptionEnum::NL_COV:
      return OptionInfo{"nl-cov", {}, opts.arith.nlCovWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arith.nlCov}};
    case OptionEnum::NL_COV_FORCE:
      return OptionInfo{"nl-cov-force", {}, opts.arith.nlCovForceWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlCovForce}};
    case OptionEnum::NL_COV_LIFT:
      return OptionInfo{"nl-cov-lift", {}, opts.arith.nlCovLiftingWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"regular", opts.arith.nlCovLifting, { "lazard", "regular" }}};
    case OptionEnum::NL_COV_LINEAR_MODEL:
      return OptionInfo{"nl-cov-linear-model", {}, opts.arith.nlCovLinearModelWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"none", opts.arith.nlCovLinearModel, { "initial", "none", "persistent" }}};
    case OptionEnum::NL_COV_PROJ:
      return OptionInfo{"nl-cov-proj", {}, opts.arith.nlCovProjectionWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"mccallum", opts.arith.nlCovProjection, { "lazard", "lazard-mod", "mccallum" }}};
    case OptionEnum::NL_COV_PRUNE:
      return OptionInfo{"nl-cov-prune", {}, opts.arith.nlCovPruneWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlCovPrune}};
    case OptionEnum::NL_COV_VAR_ELIM:
      return OptionInfo{"nl-cov-var-elim", {}, opts.arith.nlCovVarElimWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.arith.nlCovVarElim}};
    case OptionEnum::NL_EXT:
      return OptionInfo{"nl-ext", {}, opts.arith.nlExtWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"full", opts.arith.nlExt, { "full", "light", "none" }}};
    case OptionEnum::NL_EXT_ENT_CONF:
      return OptionInfo{"nl-ext-ent-conf", {}, opts.arith.nlExtEntailConflictsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlExtEntailConflicts}};
    case OptionEnum::NL_EXT_FACTOR:
      return OptionInfo{"nl-ext-factor", {}, opts.arith.nlExtFactorWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arith.nlExtFactor}};
    case OptionEnum::NL_EXT_INC_PREC:
      return OptionInfo{"nl-ext-inc-prec", {}, opts.arith.nlExtIncPrecisionWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.arith.nlExtIncPrecision}};
    case OptionEnum::NL_EXT_PURIFY:
      return OptionInfo{"nl-ext-purify", {}, opts.arith.nlExtPurifyWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlExtPurify}};
    case OptionEnum::NL_EXT_RBOUND:
      return OptionInfo{"nl-ext-rbound", {}, opts.arith.nlExtResBoundWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlExtResBound}};
    case OptionEnum::NL_EXT_REWRITE:
      return OptionInfo{"nl-ext-rewrite", {}, opts.arith.nlExtRewritesWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arith.nlExtRewrites}};
    case OptionEnum::NL_EXT_SPLIT_ZERO:
      return OptionInfo{"nl-ext-split-zero", {}, opts.arith.nlExtSplitZeroWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlExtSplitZero}};
    case OptionEnum::NL_EXT_TF_TAYLOR_DEG:
      return OptionInfo{"nl-ext-tf-taylor-deg", {}, opts.arith.nlExtTfTaylorDegreeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{4, opts.arith.nlExtTfTaylorDegree, {}, {}}};
    case OptionEnum::NL_EXT_TF_TPLANES:
      return OptionInfo{"nl-ext-tf-tplanes", {}, opts.arith.nlExtTfTangentPlanesWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arith.nlExtTfTangentPlanes}};
    case OptionEnum::NL_EXT_TPLANES:
      return OptionInfo{"nl-ext-tplanes", {}, opts.arith.nlExtTangentPlanesWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.arith.nlExtTangentPlanes}};
    case OptionEnum::NL_EXT_TPLANES_INTERLEAVE:
      return OptionInfo{"nl-ext-tplanes-interleave", {}, opts.arith.nlExtTangentPlanesInterleaveWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.arith.nlExtTangentPlanesInterleave}};
    case OptionEnum::NL_ICP:
      return OptionInfo{"nl-icp", {}, opts.arith.nlICPWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlICP}};
    case OptionEnum::NL_RLV:
      return OptionInfo{"nl-rlv", {}, opts.arith.nlRlvModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.arith.nlRlvMode, { "always", "interleave", "none" }}};
    case OptionEnum::NL_RLV_ASSERT_BOUNDS:
      return OptionInfo{"nl-rlv-assert-bounds", {}, opts.arith.nlRlvAssertBoundsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.nlRlvAssertBounds}};
    case OptionEnum::ON_REPEAT_ITE_SIMP:
      return OptionInfo{"on-repeat-ite-simp", {}, opts.smt.doITESimpOnRepeatWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.doITESimpOnRepeat}};
    case OptionEnum::OPT_RES_RECONSTRUCTION_SIZE:
      return OptionInfo{"opt-res-reconstruction-size", {}, opts.proof.optResReconSizeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.proof.optResReconSize}};
    case OptionEnum::ORACLES:
      return OptionInfo{"oracles", {}, opts.quantifiers.oraclesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.oracles}};
    case OptionEnum::OUT:
      return OptionInfo{"out", {"regular-output-channel"}, opts.base.outWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::VoidInfo{}};
    case OptionEnum::OUTPUT:
      return OptionInfo{"output", {}, false, OptionInfo::Category::COMMON, OptionInfo::VoidInfo{}};
    case OptionEnum::OUTPUT_LANG:
      return OptionInfo{"output-lang", {"output-language"}, opts.printer.outputLanguageWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::VoidInfo{}};
    case OptionEnum::PARSE_ONLY:
      return OptionInfo{"parse-only", {}, opts.base.parseOnlyWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.base.parseOnly}};
    case OptionEnum::PARSE_SKOLEM_DEFINITIONS:
      return OptionInfo{"parse-skolem-definitions", {}, opts.parser.parseSkolemDefinitionsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.parser.parseSkolemDefinitions}};
    case OptionEnum::PARSING_MODE:
      return OptionInfo{"parsing-mode", {}, opts.parser.parsingModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"default", opts.parser.parsingMode, { "default", "lenient", "strict" }}};
    case OptionEnum::PARTIAL_TRIGGERS:
      return OptionInfo{"partial-triggers", {}, opts.quantifiers.partialTriggersWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.partialTriggers}};
    case OptionEnum::PARTITION_CHECK:
      return OptionInfo{"partition-check", {"check"}, opts.parallel.partitionCheckWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"standard", opts.parallel.partitionCheck, { "full", "standard" }}};
    case OptionEnum::PARTITION_CONFLICT_SIZE:
      return OptionInfo{"partition-conflict-size", {}, opts.parallel.partitionConflictSizeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{0, opts.parallel.partitionConflictSize, {}, {}}};
    case OptionEnum::PARTITION_START_TIME:
      return OptionInfo{"partition-start-time", {}, opts.parallel.partitionStartTimeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{30, opts.parallel.partitionStartTime, {}, {}}};
    case OptionEnum::PARTITION_STRATEGY:
      return OptionInfo{"partition-strategy", {"partition"}, opts.parallel.partitionStrategyWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"decision-scatter", opts.parallel.partitionStrategy, { "decision-cube", "decision-scatter", "heap-cube", "heap-scatter", "lemma-cube", "lemma-scatter" }}};
    case OptionEnum::PARTITION_TIME_INTERVAL:
      return OptionInfo{"partition-time-interval", {}, opts.parallel.partitionTimeIntervalWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<double>{1.0, opts.parallel.partitionTimeInterval, {}, {}}};
    case OptionEnum::PARTITION_TLIMIT:
      return OptionInfo{"partition-tlimit", {}, opts.parallel.partitionTimeLimitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{60, opts.parallel.partitionTimeLimit, {}, {}}};
    case OptionEnum::PARTITION_WHEN:
      return OptionInfo{"partition-when", {}, opts.parallel.partitionWhenWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"tlimit", opts.parallel.partitionWhen, { "climit", "tlimit" }}};
    case OptionEnum::PB_REWRITES:
      return OptionInfo{"pb-rewrites", {}, opts.arith.pbRewritesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.pbRewrites}};
    case OptionEnum::PIVOT_THRESHOLD:
      return OptionInfo{"pivot-threshold", {}, opts.arith.arithPivotThresholdWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{2, opts.arith.arithPivotThreshold, {}, {}}};
    case OptionEnum::PLUGIN_NOTIFY_SAT_CLAUSE_IN_SOLVE:
      return OptionInfo{"plugin-notify-sat-clause-in-solve", {}, opts.base.pluginNotifySatClauseInSolveWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.base.pluginNotifySatClauseInSolve}};
    case OptionEnum::PLUGIN_SHARE_SKOLEMS:
      return OptionInfo{"plugin-share-skolems", {}, opts.base.pluginShareSkolemsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.base.pluginShareSkolems}};
    case OptionEnum::POOL_INST:
      return OptionInfo{"pool-inst", {}, opts.quantifiers.poolInstWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.poolInst}};
    case OptionEnum::PORTFOLIO_JOBS:
      return OptionInfo{"portfolio-jobs", {}, opts.driver.portfolioJobsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{1, opts.driver.portfolioJobs, {}, {}}};
    case OptionEnum::PP_ASSERT_MAX_SUB_SIZE:
      return OptionInfo{"pp-assert-max-sub-size", {}, opts.arith.ppAssertMaxSubSizeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{2, opts.arith.ppAssertMaxSubSize, {}, {}}};
    case OptionEnum::PRE_SKOLEM_QUANT:
      return OptionInfo{"pre-skolem-quant", {}, opts.quantifiers.preSkolemQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"off", opts.quantifiers.preSkolemQuant, { "agg", "off", "on" }}};
    case OptionEnum::PRE_SKOLEM_QUANT_NESTED:
      return OptionInfo{"pre-skolem-quant-nested", {}, opts.quantifiers.preSkolemQuantNestedWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.preSkolemQuantNested}};
    case OptionEnum::PRENEX_QUANT:
      return OptionInfo{"prenex-quant", {}, opts.quantifiers.prenexQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"simple", opts.quantifiers.prenexQuant, { "none", "norm", "simple" }}};
    case OptionEnum::PRENEX_QUANT_USER:
      return OptionInfo{"prenex-quant-user", {}, opts.quantifiers.prenexQuantUserWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.prenexQuantUser}};
    case OptionEnum::PREPROCESS_ONLY:
      return OptionInfo{"preprocess-only", {}, opts.base.preprocessOnlyWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.base.preprocessOnly}};
    case OptionEnum::PREREGISTER_MODE:
      return OptionInfo{"preregister-mode", {}, opts.prop.preRegisterModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"eager", opts.prop.preRegisterMode, { "eager", "lazy" }}};
    case OptionEnum::PRINT_ARITH_LIT_TOKEN:
      return OptionInfo{"print-arith-lit-token", {}, opts.printer.printArithLitTokenWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.printer.printArithLitToken}};
    case OptionEnum::PRINT_CORES_FULL:
      return OptionInfo{"print-cores-full", {}, opts.smt.printCoresFullWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.printCoresFull}};
    case OptionEnum::PRINT_DOT_CLUSTERS:
      return OptionInfo{"print-dot-clusters", {}, opts.proof.printDotClustersWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.proof.printDotClusters}};
    case OptionEnum::PRINT_INST:
      return OptionInfo{"print-inst", {}, opts.quantifiers.printInstModeWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ModeInfo{"list", opts.quantifiers.printInstMode, { "list", "num" }}};
    case OptionEnum::PRINT_INST_FULL:
      return OptionInfo{"print-inst-full", {}, opts.quantifiers.printInstFullWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.printInstFull}};
    case OptionEnum::PRINT_SKOLEM_DEFINITIONS:
      return OptionInfo{"print-skolem-definitions", {}, opts.printer.printSkolemDefinitionsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.printer.printSkolemDefinitions}};
    case OptionEnum::PRINT_SUCCESS:
      return OptionInfo{"print-success", {}, opts.driver.printSuccessWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.driver.printSuccess}};
    case OptionEnum::PRODUCE_ABDUCTS:
      return OptionInfo{"produce-abducts", {}, opts.smt.produceAbductsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.produceAbducts}};
    case OptionEnum::PRODUCE_ASSERTIONS:
      return OptionInfo{"produce-assertions", {"interactive-mode"}, opts.smt.produceAssertionsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.smt.produceAssertions}};
    case OptionEnum::PRODUCE_ASSIGNMENTS:
      return OptionInfo{"produce-assignments", {}, opts.smt.produceAssignmentsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.produceAssignments}};
    case OptionEnum::PRODUCE_DIFFICULTY:
      return OptionInfo{"produce-difficulty", {}, opts.smt.produceDifficultyWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.produceDifficulty}};
    case OptionEnum::PRODUCE_INTERPOLANTS:
      return OptionInfo{"produce-interpolants", {}, opts.smt.produceInterpolantsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.produceInterpolants}};
    case OptionEnum::PRODUCE_LEARNED_LITERALS:
      return OptionInfo{"produce-learned-literals", {}, opts.smt.produceLearnedLiteralsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.produceLearnedLiterals}};
    case OptionEnum::PRODUCE_MODELS:
      return OptionInfo{"produce-models", {}, opts.smt.produceModelsWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.smt.produceModels}};
    case OptionEnum::PRODUCE_PROOFS:
      return OptionInfo{"produce-proofs", {}, opts.smt.produceProofsWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.smt.produceProofs}};
    case OptionEnum::PRODUCE_UNSAT_ASSUMPTIONS:
      return OptionInfo{"produce-unsat-assumptions", {}, opts.smt.unsatAssumptionsWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.unsatAssumptions}};
    case OptionEnum::PRODUCE_UNSAT_CORES:
      return OptionInfo{"produce-unsat-cores", {}, opts.smt.produceUnsatCoresWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.produceUnsatCores}};
    case OptionEnum::PROOF_ALETHE_DEFINE_SKOLEMS:
      return OptionInfo{"proof-alethe-define-skolems", {}, opts.proof.proofAletheDefineSkolemsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.proofAletheDefineSkolems}};
    case OptionEnum::PROOF_ALETHE_RES_PIVOTS:
      return OptionInfo{"proof-alethe-res-pivots", {}, opts.proof.proofAletheResPivotsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.proofAletheResPivots}};
    case OptionEnum::PROOF_ALLOW_TRUST:
      return OptionInfo{"proof-allow-trust", {}, opts.proof.proofAllowTrustWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.proof.proofAllowTrust}};
    case OptionEnum::PROOF_CHECK:
      return OptionInfo{"proof-check", {}, opts.proof.proofCheckWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.proof.proofCheck, { "eager", "eager-simple", "lazy", "none" }}};
    case OptionEnum::PROOF_DAG_GLOBAL:
      return OptionInfo{"proof-dag-global", {}, opts.proof.proofDagGlobalWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.proof.proofDagGlobal}};
    case OptionEnum::PROOF_DOT_DAG:
      return OptionInfo{"proof-dot-dag", {}, opts.proof.printDotAsDAGWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.printDotAsDAG}};
    case OptionEnum::PROOF_ELIM_SUBTYPES:
      return OptionInfo{"proof-elim-subtypes", {}, opts.proof.proofElimSubtypesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.proofElimSubtypes}};
    case OptionEnum::PROOF_FORMAT_MODE:
      return OptionInfo{"proof-format-mode", {}, opts.proof.proofFormatModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"cpc", opts.proof.proofFormatMode, { "alethe", "cpc", "dot", "lfsc", "none" }}};
    case OptionEnum::PROOF_GRANULARITY:
      return OptionInfo{"proof-granularity", {}, opts.proof.proofGranularityModeWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ModeInfo{"macro", opts.proof.proofGranularityMode, { "dsl-rewrite", "dsl-rewrite-strict", "macro", "rewrite", "theory-rewrite" }}};
    case OptionEnum::PROOF_MODE:
      return OptionInfo{"proof-mode", {}, opts.smt.proofModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"off", opts.smt.proofMode, { "full-proof", "full-proof-strict", "off", "pp-only", "sat-proof" }}};
    case OptionEnum::PROOF_PEDANTIC:
      return OptionInfo{"proof-pedantic", {}, opts.proof.proofPedanticWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{0, opts.proof.proofPedantic, {}, 100}};
    case OptionEnum::PROOF_PP_MERGE:
      return OptionInfo{"proof-pp-merge", {}, opts.proof.proofPpMergeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.proof.proofPpMerge}};
    case OptionEnum::PROOF_PRINT_CONCLUSION:
      return OptionInfo{"proof-print-conclusion", {}, opts.proof.proofPrintConclusionWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.proofPrintConclusion}};
    case OptionEnum::PROOF_PRINT_REFERENCE:
      return OptionInfo{"proof-print-reference", {}, opts.proof.proofPrintReferenceWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.proofPrintReference}};
    case OptionEnum::PROOF_PRUNE_INPUT:
      return OptionInfo{"proof-prune-input", {}, opts.proof.proofPruneInputWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.proof.proofPruneInput}};
    case OptionEnum::PROOF_REWRITE_RCONS_REC_LIMIT:
      return OptionInfo{"proof-rewrite-rcons-rec-limit", {}, opts.proof.proofRewriteRconsRecLimitWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<uint64_t>{5, opts.proof.proofRewriteRconsRecLimit, {}, {}}};
    case OptionEnum::PROOF_REWRITE_RCONS_STEP_LIMIT:
      return OptionInfo{"proof-rewrite-rcons-step-limit", {}, opts.proof.proofRewriteRconsStepLimitWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<uint64_t>{1000, opts.proof.proofRewriteRconsStepLimit, {}, {}}};
    case OptionEnum::PROP_PROOF_MODE:
      return OptionInfo{"prop-proof-mode", {}, opts.proof.propProofModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"proof", opts.proof.propProofMode, { "proof", "sat-external-prove" }}};
    case OptionEnum::PROP_ROW_LENGTH:
      return OptionInfo{"prop-row-length", {}, opts.arith.arithPropagateMaxLengthWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{16, opts.arith.arithPropagateMaxLength, {}, {}}};
    case OptionEnum::PURIFY_TRIGGERS:
      return OptionInfo{"purify-triggers", {}, opts.quantifiers.purifyTriggersWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.purifyTriggers}};
    case OptionEnum::QUANT_ALPHA_EQUIV:
      return OptionInfo{"quant-alpha-equiv", {}, opts.quantifiers.quantAlphaEquivWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.quantAlphaEquiv}};
    case OptionEnum::QUANT_DSPLIT:
      return OptionInfo{"quant-dsplit", {}, opts.quantifiers.quantDynamicSplitWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"default", opts.quantifiers.quantDynamicSplit, { "agg", "default", "none" }}};
    case OptionEnum::QUANT_FUN_WD:
      return OptionInfo{"quant-fun-wd", {}, opts.quantifiers.quantFunWellDefinedWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.quantFunWellDefined}};
    case OptionEnum::QUANT_IND:
      return OptionInfo{"quant-ind", {}, opts.quantifiers.quantInductionWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.quantInduction}};
    case OptionEnum::QUANT_REP_MODE:
      return OptionInfo{"quant-rep-mode", {}, opts.quantifiers.quantRepModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"first", opts.quantifiers.quantRepMode, { "depth", "ee", "first" }}};
    case OptionEnum::QUIET:
      return OptionInfo{"quiet", {}, false, OptionInfo::Category::COMMON, OptionInfo::VoidInfo{}};
    case OptionEnum::RANDOM_FREQ:
      return OptionInfo{"random-freq", {"random-frequency"}, opts.prop.satRandomFreqWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<double>{0.0, opts.prop.satRandomFreq, 0.0, 1.0}};
    case OptionEnum::RANDOM_PARTITIONING:
      return OptionInfo{"random-partitioning", {}, opts.parallel.randomPartitioningWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.parallel.randomPartitioning}};
    case OptionEnum::RE_ELIM:
      return OptionInfo{"re-elim", {}, opts.strings.regExpElimWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"off", opts.strings.regExpElim, { "agg", "off", "on" }}};
    case OptionEnum::RE_INTER_MODE:
      return OptionInfo{"re-inter-mode", {}, opts.strings.stringRegExpInterModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.strings.stringRegExpInterMode, { "all", "constant", "none", "one-constant" }}};
    case OptionEnum::REGISTER_QUANT_BODY_TERMS:
      return OptionInfo{"register-quant-body-terms", {}, opts.quantifiers.registerQuantBodyTermsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.registerQuantBodyTerms}};
    case OptionEnum::RELATIONAL_TRIGGERS:
      return OptionInfo{"relational-triggers", {}, opts.quantifiers.relationalTriggersWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.relationalTriggers}};
    case OptionEnum::RELEVANCE_FILTER:
      return OptionInfo{"relevance-filter", {}, opts.theory.relevanceFilterWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.theory.relevanceFilter}};
    case OptionEnum::RELEVANT_TRIGGERS:
      return OptionInfo{"relevant-triggers", {}, opts.quantifiers.relevantTriggersWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.relevantTriggers}};
    case OptionEnum::RELS_EXP:
      return OptionInfo{"rels-exp", {}, opts.sets.relsExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.sets.relsExp}};
    case OptionEnum::REPEAT_SIMP:
      return OptionInfo{"repeat-simp", {}, opts.smt.repeatSimpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.repeatSimp}};
    case OptionEnum::REPLAY_EARLY_CLOSE_DEPTH:
      return OptionInfo{"replay-early-close-depth", {}, opts.arith.replayEarlyCloseDepthsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{1, opts.arith.replayEarlyCloseDepths, {}, {}}};
    case OptionEnum::REPLAY_LEMMA_REJECT_CUT:
      return OptionInfo{"replay-lemma-reject-cut", {}, opts.arith.lemmaRejectCutSizeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{25500, opts.arith.lemmaRejectCutSize, {}, {}}};
    case OptionEnum::REPLAY_NUM_ERR_PENALTY:
      return OptionInfo{"replay-num-err-penalty", {}, opts.arith.replayNumericFailurePenaltyWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{4194304, opts.arith.replayNumericFailurePenalty, {}, {}}};
    case OptionEnum::REPLAY_REJECT_CUT:
      return OptionInfo{"replay-reject-cut", {}, opts.arith.replayRejectCutSizeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{25500, opts.arith.replayRejectCutSize, {}, {}}};
    case OptionEnum::RESTART_INT_BASE:
      return OptionInfo{"restart-int-base", {}, opts.prop.satRestartFirstWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{25, opts.prop.satRestartFirst, {}, {}}};
    case OptionEnum::RESTART_INT_INC:
      return OptionInfo{"restart-int-inc", {}, opts.prop.satRestartIncWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<double>{3.0, opts.prop.satRestartInc, 0.0, {}}};
    case OptionEnum::RESTRICT_PIVOTS:
      return OptionInfo{"restrict-pivots", {}, opts.arith.restrictedPivotsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.arith.restrictedPivots}};
    case OptionEnum::REVERT_ARITH_MODELS_ON_UNSAT:
      return OptionInfo{"revert-arith-models-on-unsat", {}, opts.arith.revertArithModelsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.revertArithModels}};
    case OptionEnum::RLIMIT:
      return OptionInfo{"rlimit", {}, opts.base.cumulativeResourceLimitWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<uint64_t>{0, opts.base.cumulativeResourceLimit, {}, {}}};
    case OptionEnum::RLIMIT_PER:
      return OptionInfo{"rlimit-per", {"reproducible-resource-limit"}, opts.base.perCallResourceLimitWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<uint64_t>{0, opts.base.perCallResourceLimit, {}, {}}};
    case OptionEnum::RR_TURNS:
      return OptionInfo{"rr-turns", {}, opts.arith.rrTurnsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{3, opts.arith.rrTurns, {}, {}}};
    case OptionEnum::RWEIGHT:
      return OptionInfo{"rweight", {}, false, OptionInfo::Category::EXPERT, OptionInfo::VoidInfo{}};
    case OptionEnum::SAFE_OPTIONS:
      return OptionInfo{"safe-options", {}, opts.base.safeOptionsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.base.safeOptions}};
    case OptionEnum::SAT_PROOF_MIN_DIMACS:
      return OptionInfo{"sat-proof-min-dimacs", {}, opts.proof.satProofMinDimacsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.proof.satProofMinDimacs}};
    case OptionEnum::SAT_RANDOM_SEED:
      return OptionInfo{"sat-random-seed", {}, opts.prop.satRandomSeedWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<uint64_t>{0, opts.prop.satRandomSeed, {}, {}}};
    case OptionEnum::SAT_SOLVER:
      return OptionInfo{"sat-solver", {}, opts.prop.satSolverWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"minisat", opts.prop.satSolver, { "cadical", "minisat" }}};
    case OptionEnum::SE_SOLVE_INT:
      return OptionInfo{"se-solve-int", {}, opts.arith.trySolveIntStandardEffortWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.trySolveIntStandardEffort}};
    case OptionEnum::SEED:
      return OptionInfo{"seed", {"random-seed"}, opts.driver.seedWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<uint64_t>{0, opts.driver.seed, {}, {}}};
    case OptionEnum::SEGV_SPIN:
      return OptionInfo{"segv-spin", {}, opts.driver.segvSpinWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.driver.segvSpin}};
    case OptionEnum::SEMANTIC_CHECKS:
      return OptionInfo{"semantic-checks", {}, opts.parser.semanticChecksWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{DO_SEMANTIC_CHECKS_BY_DEFAULT, opts.parser.semanticChecks}};
    case OptionEnum::SEP:
      return OptionInfo{"sep", {}, opts.sep.sepWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.sep.sep}};
    case OptionEnum::SEP_MIN_REFINE:
      return OptionInfo{"sep-min-refine", {}, opts.sep.sepMinimalRefineWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.sep.sepMinimalRefine}};
    case OptionEnum::SEP_PRE_SKOLEM_EMP:
      return OptionInfo{"sep-pre-skolem-emp", {}, opts.sep.sepPreSkolemEmpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.sep.sepPreSkolemEmp}};
    case OptionEnum::SEQ_ARRAY:
      return OptionInfo{"seq-array", {}, opts.strings.seqArrayWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.strings.seqArray, { "eager", "lazy", "none" }}};
    case OptionEnum::SETS_CARD_EXP:
      return OptionInfo{"sets-card-exp", {}, opts.sets.setsCardExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.sets.setsCardExp}};
    case OptionEnum::SETS_EXP:
      return OptionInfo{"sets-exp", {}, opts.sets.setsExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.sets.setsExp}};
    case OptionEnum::SETS_PROXY_LEMMAS:
      return OptionInfo{"sets-proxy-lemmas", {}, opts.sets.setsProxyLemmasWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.sets.setsProxyLemmas}};
    case OptionEnum::SHOW_CONFIG:
      return OptionInfo{"show-config", {}, opts.driver.showConfigurationWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.driver.showConfiguration}};
    case OptionEnum::SHOW_TRACE_TAGS:
      return OptionInfo{"show-trace-tags", {}, opts.driver.showTraceTagsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.driver.showTraceTags}};
    case OptionEnum::SIMP_ITE_COMPRESS:
      return OptionInfo{"simp-ite-compress", {}, opts.smt.compressItesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.compressItes}};
    case OptionEnum::SIMP_WITH_CARE:
      return OptionInfo{"simp-with-care", {}, opts.smt.simplifyWithCareEnabledWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.simplifyWithCareEnabled}};
    case OptionEnum::SIMPLEX_CHECK_PERIOD:
      return OptionInfo{"simplex-check-period", {}, opts.arith.arithSimplexCheckPeriodWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{200, opts.arith.arithSimplexCheckPeriod, {}, {}}};
    case OptionEnum::SIMPLIFICATION:
      return OptionInfo{"simplification", {"simplification-mode"}, opts.smt.simplificationModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"batch", opts.smt.simplificationMode, { "batch", "none" }}};
    case OptionEnum::SIMPLIFICATION_BCP:
      return OptionInfo{"simplification-bcp", {}, opts.smt.simplificationBoolConstPropWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.smt.simplificationBoolConstProp}};
    case OptionEnum::SOI_QE:
      return OptionInfo{"soi-qe", {}, opts.arith.soiQuickExplainWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.soiQuickExplain}};
    case OptionEnum::SOLVE_BV_AS_INT:
      return OptionInfo{"solve-bv-as-int", {}, opts.smt.solveBVAsIntWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"off", opts.smt.solveBVAsInt, { "bitwise", "bv", "iand", "off", "sum" }}};
    case OptionEnum::SOLVE_INT_AS_BV:
      return OptionInfo{"solve-int-as-bv", {}, opts.smt.solveIntAsBVWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<uint64_t>{0, opts.smt.solveIntAsBV, {}, 4294967295}};
    case OptionEnum::SOLVE_REAL_AS_INT:
      return OptionInfo{"solve-real-as-int", {}, opts.smt.solveRealAsIntWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.solveRealAsInt}};
    case OptionEnum::SORT_INFERENCE:
      return OptionInfo{"sort-inference", {}, opts.smt.sortInferenceWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.sortInference}};
    case OptionEnum::STANDARD_EFFORT_VARIABLE_ORDER_PIVOTS:
      return OptionInfo{"standard-effort-variable-order-pivots", {}, opts.arith.arithStandardCheckVarOrderPivotsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{-1, opts.arith.arithStandardCheckVarOrderPivots, {}, {}}};
    case OptionEnum::STATIC_LEARNING:
      return OptionInfo{"static-learning", {}, opts.smt.staticLearningWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.smt.staticLearning}};
    case OptionEnum::STATS:
      return OptionInfo{"stats", {}, opts.base.statisticsWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.base.statistics}};
    case OptionEnum::STATS_ALL:
      return OptionInfo{"stats-all", {}, opts.base.statisticsAllWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.base.statisticsAll}};
    case OptionEnum::STATS_EVERY_QUERY:
      return OptionInfo{"stats-every-query", {}, opts.base.statisticsEveryQueryWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.base.statisticsEveryQuery}};
    case OptionEnum::STATS_INTERNAL:
      return OptionInfo{"stats-internal", {}, opts.base.statisticsInternalWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.base.statisticsInternal}};
    case OptionEnum::STRICT_PARSING:
      return OptionInfo{"strict-parsing", {}, opts.parser.strictParsingWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.parser.strictParsing}};
    case OptionEnum::STRINGS_ALPHA_CARD:
      return OptionInfo{"strings-alpha-card", {}, opts.strings.stringsAlphaCardWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{196608, opts.strings.stringsAlphaCard, {}, 196608}};
    case OptionEnum::STRINGS_CHECK_ENTAIL_LEN:
      return OptionInfo{"strings-check-entail-len", {}, opts.strings.stringCheckEntailLenWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.strings.stringCheckEntailLen}};
    case OptionEnum::STRINGS_DEQ_EXT:
      return OptionInfo{"strings-deq-ext", {}, opts.strings.stringsDeqExtWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.strings.stringsDeqExt}};
    case OptionEnum::STRINGS_EAGER_EVAL:
      return OptionInfo{"strings-eager-eval", {}, opts.strings.stringEagerEvalWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.strings.stringEagerEval}};
    case OptionEnum::STRINGS_EAGER_LEN_RE:
      return OptionInfo{"strings-eager-len-re", {}, opts.strings.stringsEagerLenEntRegexpWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.strings.stringsEagerLenEntRegexp}};
    case OptionEnum::STRINGS_EAGER_REG:
      return OptionInfo{"strings-eager-reg", {}, opts.strings.stringEagerRegWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.strings.stringEagerReg}};
    case OptionEnum::STRINGS_EAGER_SOLVER:
      return OptionInfo{"strings-eager-solver", {}, opts.strings.stringEagerSolverWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.strings.stringEagerSolver}};
    case OptionEnum::STRINGS_EXP:
      return OptionInfo{"strings-exp", {}, opts.strings.stringExpWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.strings.stringExp}};
    case OptionEnum::STRINGS_FF:
      return OptionInfo{"strings-ff", {}, opts.strings.stringFlatFormsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.strings.stringFlatForms}};
    case OptionEnum::STRINGS_FMF:
      return OptionInfo{"strings-fmf", {}, opts.strings.stringFMFWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.strings.stringFMF}};
    case OptionEnum::STRINGS_INFER_AS_LEMMAS:
      return OptionInfo{"strings-infer-as-lemmas", {}, opts.strings.stringInferAsLemmasWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.strings.stringInferAsLemmas}};
    case OptionEnum::STRINGS_INFER_SYM:
      return OptionInfo{"strings-infer-sym", {}, opts.strings.stringInferSymWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.strings.stringInferSym}};
    case OptionEnum::STRINGS_LAZY_PP:
      return OptionInfo{"strings-lazy-pp", {}, opts.strings.stringLazyPreprocWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.strings.stringLazyPreproc}};
    case OptionEnum::STRINGS_LEN_NORM:
      return OptionInfo{"strings-len-norm", {}, opts.strings.stringLenNormWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.strings.stringLenNorm}};
    case OptionEnum::STRINGS_MBR:
      return OptionInfo{"strings-mbr", {}, opts.strings.stringModelBasedReductionWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.strings.stringModelBasedReduction}};
    case OptionEnum::STRINGS_MODEL_MAX_LEN:
      return OptionInfo{"strings-model-max-len", {}, opts.strings.stringsModelMaxLengthWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{65536, opts.strings.stringsModelMaxLength, {}, 2147483647}};
    case OptionEnum::STRINGS_PROCESS_LOOP_MODE:
      return OptionInfo{"strings-process-loop-mode", {}, opts.strings.stringProcessLoopModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"full", opts.strings.stringProcessLoopMode, { "abort", "full", "none", "simple", "simple-abort" }}};
    case OptionEnum::STRINGS_RE_DERIVE_CONF:
      return OptionInfo{"strings-re-derive-conf", {}, opts.strings.stringRegexpDeriveConflictsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.strings.stringRegexpDeriveConflicts}};
    case OptionEnum::STRINGS_RE_POSC_EAGER:
      return OptionInfo{"strings-re-posc-eager", {}, opts.strings.stringRegexpPosConcatEagerWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.strings.stringRegexpPosConcatEager}};
    case OptionEnum::STRINGS_REC_ARITH_APPROX:
      return OptionInfo{"strings-rec-arith-approx", {}, opts.strings.stringRecArithApproxWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.strings.stringRecArithApprox}};
    case OptionEnum::STRINGS_REGEXP_INCLUSION:
      return OptionInfo{"strings-regexp-inclusion", {}, opts.strings.stringRegexpInclusionWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.strings.stringRegexpInclusion}};
    case OptionEnum::STRINGS_REXPLAIN_LEMMAS:
      return OptionInfo{"strings-rexplain-lemmas", {}, opts.strings.stringRExplainLemmasWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.strings.stringRExplainLemmas}};
    case OptionEnum::SUB_CBQI:
      return OptionInfo{"sub-cbqi", {}, opts.quantifiers.quantSubCbqiWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.quantSubCbqi}};
    case OptionEnum::SUB_CBQI_TIMEOUT:
      return OptionInfo{"sub-cbqi-timeout", {}, opts.quantifiers.quantSubCbqiTimeoutWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<uint64_t>{0, opts.quantifiers.quantSubCbqiTimeout, {}, {}}};
    case OptionEnum::SYGUS:
      return OptionInfo{"sygus", {}, opts.quantifiers.sygusWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygus}};
    case OptionEnum::SYGUS_ABORT_SIZE:
      return OptionInfo{"sygus-abort-size", {}, opts.datatypes.sygusAbortSizeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<int64_t>{-1, opts.datatypes.sygusAbortSize, {}, {}}};
    case OptionEnum::SYGUS_ADD_CONST_GRAMMAR:
      return OptionInfo{"sygus-add-const-grammar", {}, opts.quantifiers.sygusAddConstGrammarWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusAddConstGrammar}};
    case OptionEnum::SYGUS_ARG_RELEVANT:
      return OptionInfo{"sygus-arg-relevant", {}, opts.quantifiers.sygusArgRelevantWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusArgRelevant}};
    case OptionEnum::SYGUS_AUTO_UNFOLD:
      return OptionInfo{"sygus-auto-unfold", {}, opts.quantifiers.sygusInvAutoUnfoldWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusInvAutoUnfold}};
    case OptionEnum::SYGUS_BOOL_ITE_RETURN_CONST:
      return OptionInfo{"sygus-bool-ite-return-const", {}, opts.quantifiers.sygusBoolIteReturnConstWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusBoolIteReturnConst}};
    case OptionEnum::SYGUS_CORE_CONNECTIVE:
      return OptionInfo{"sygus-core-connective", {}, opts.quantifiers.sygusCoreConnectiveWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusCoreConnective}};
    case OptionEnum::SYGUS_CREPAIR_ABORT:
      return OptionInfo{"sygus-crepair-abort", {}, opts.quantifiers.sygusConstRepairAbortWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusConstRepairAbort}};
    case OptionEnum::SYGUS_ENUM:
      return OptionInfo{"sygus-enum", {}, opts.quantifiers.sygusEnumModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"auto", opts.quantifiers.sygusEnumMode, { "auto", "fast", "random", "smart", "var-agnostic" }}};
    case OptionEnum::SYGUS_ENUM_FAST_NUM_CONSTS:
      return OptionInfo{"sygus-enum-fast-num-consts", {}, opts.quantifiers.sygusEnumFastNumConstsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{5, opts.quantifiers.sygusEnumFastNumConsts, {}, {}}};
    case OptionEnum::SYGUS_ENUM_RANDOM_P:
      return OptionInfo{"sygus-enum-random-p", {}, opts.quantifiers.sygusEnumRandomPWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<double>{0.5, opts.quantifiers.sygusEnumRandomP, 0.0, 1.0}};
    case OptionEnum::SYGUS_EVAL_UNFOLD:
      return OptionInfo{"sygus-eval-unfold", {}, opts.quantifiers.sygusEvalUnfoldModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"single-bool", opts.quantifiers.sygusEvalUnfoldMode, { "multi", "none", "single", "single-bool" }}};
    case OptionEnum::SYGUS_EXPR_MINER_CHECK_TIMEOUT:
      return OptionInfo{"sygus-expr-miner-check-timeout", {}, opts.quantifiers.sygusExprMinerCheckTimeoutWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{0, opts.quantifiers.sygusExprMinerCheckTimeout, {}, {}}};
    case OptionEnum::SYGUS_FAIR:
      return OptionInfo{"sygus-fair", {}, opts.datatypes.sygusFairWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"dt-size", opts.datatypes.sygusFair, { "direct", "dt-height-bound", "dt-size", "dt-size-bound", "none" }}};
    case OptionEnum::SYGUS_FAIR_MAX:
      return OptionInfo{"sygus-fair-max", {}, opts.datatypes.sygusFairMaxWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.datatypes.sygusFairMax}};
    case OptionEnum::SYGUS_FILTER_SOL:
      return OptionInfo{"sygus-filter-sol", {}, opts.quantifiers.sygusFilterSolModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.quantifiers.sygusFilterSolMode, { "none", "strong", "weak" }}};
    case OptionEnum::SYGUS_FILTER_SOL_REV:
      return OptionInfo{"sygus-filter-sol-rev", {}, opts.quantifiers.sygusFilterSolRevSubsumeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusFilterSolRevSubsume}};
    case OptionEnum::SYGUS_GRAMMAR_CONS:
      return OptionInfo{"sygus-grammar-cons", {}, opts.quantifiers.sygusGrammarConsModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"simple", opts.quantifiers.sygusGrammarConsMode, { "any-const", "any-term", "any-term-concise", "simple" }}};
    case OptionEnum::SYGUS_GRAMMAR_NORM:
      return OptionInfo{"sygus-grammar-norm", {}, opts.quantifiers.sygusGrammarNormWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusGrammarNorm}};
    case OptionEnum::SYGUS_GRAMMAR_USE_DISJ:
      return OptionInfo{"sygus-grammar-use-disj", {}, opts.quantifiers.sygusGrammarUseDisjWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusGrammarUseDisj}};
    case OptionEnum::SYGUS_INFERENCE:
      return OptionInfo{"sygus-inference", {}, opts.quantifiers.sygusInferenceWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"off", opts.quantifiers.sygusInference, { "off", "on", "try" }}};
    case OptionEnum::SYGUS_INST:
      return OptionInfo{"sygus-inst", {}, opts.quantifiers.sygusInstWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusInst}};
    case OptionEnum::SYGUS_INST_MODE:
      return OptionInfo{"sygus-inst-mode", {}, opts.quantifiers.sygusInstModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"priority-inst", opts.quantifiers.sygusInstMode, { "interleave", "priority-eval", "priority-inst" }}};
    case OptionEnum::SYGUS_INST_SCOPE:
      return OptionInfo{"sygus-inst-scope", {}, opts.quantifiers.sygusInstScopeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"in", opts.quantifiers.sygusInstScope, { "both", "in", "out" }}};
    case OptionEnum::SYGUS_INST_TERM_SEL:
      return OptionInfo{"sygus-inst-term-sel", {}, opts.quantifiers.sygusInstTermSelWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"min", opts.quantifiers.sygusInstTermSel, { "both", "max", "min" }}};
    case OptionEnum::SYGUS_INV_TEMPL:
      return OptionInfo{"sygus-inv-templ", {}, opts.quantifiers.sygusInvTemplModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"post", opts.quantifiers.sygusInvTemplMode, { "none", "post", "pre" }}};
    case OptionEnum::SYGUS_INV_TEMPL_WHEN_SG:
      return OptionInfo{"sygus-inv-templ-when-sg", {}, opts.quantifiers.sygusInvTemplWhenSyntaxWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusInvTemplWhenSyntax}};
    case OptionEnum::SYGUS_MIN_GRAMMAR:
      return OptionInfo{"sygus-min-grammar", {}, opts.quantifiers.sygusMinGrammarWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusMinGrammar}};
    case OptionEnum::SYGUS_OUT:
      return OptionInfo{"sygus-out", {}, opts.quantifiers.sygusOutWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"sygus-standard", opts.quantifiers.sygusOut, { "status", "status-and-def", "sygus-standard" }}};
    case OptionEnum::SYGUS_PBE:
      return OptionInfo{"sygus-pbe", {}, opts.quantifiers.sygusUnifPbeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusUnifPbe}};
    case OptionEnum::SYGUS_PBE_MULTI_FAIR:
      return OptionInfo{"sygus-pbe-multi-fair", {}, opts.quantifiers.sygusPbeMultiFairWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusPbeMultiFair}};
    case OptionEnum::SYGUS_PBE_MULTI_FAIR_DIFF:
      return OptionInfo{"sygus-pbe-multi-fair-diff", {}, opts.quantifiers.sygusPbeMultiFairDiffWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{0, opts.quantifiers.sygusPbeMultiFairDiff, {}, {}}};
    case OptionEnum::SYGUS_QE_PREPROC:
      return OptionInfo{"sygus-qe-preproc", {}, opts.quantifiers.sygusQePreprocWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusQePreproc}};
    case OptionEnum::SYGUS_QUERY_GEN:
      return OptionInfo{"sygus-query-gen", {}, opts.quantifiers.sygusQueryGenWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"basic", opts.quantifiers.sygusQueryGen, { "basic", "sample-sat", "unsat" }}};
    case OptionEnum::SYGUS_QUERY_GEN_DUMP_FILES:
      return OptionInfo{"sygus-query-gen-dump-files", {}, opts.quantifiers.sygusQueryGenDumpFilesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"none", opts.quantifiers.sygusQueryGenDumpFiles, { "all", "none", "unsolved" }}};
    case OptionEnum::SYGUS_QUERY_GEN_FILTER_SOLVED:
      return OptionInfo{"sygus-query-gen-filter-solved", {}, opts.quantifiers.sygusQueryFilterSolvedWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusQueryFilterSolved}};
    case OptionEnum::SYGUS_QUERY_GEN_THRESH:
      return OptionInfo{"sygus-query-gen-thresh", {}, opts.quantifiers.sygusQueryGenThreshWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{5, opts.quantifiers.sygusQueryGenThresh, {}, {}}};
    case OptionEnum::SYGUS_REC_FUN:
      return OptionInfo{"sygus-rec-fun", {}, opts.quantifiers.sygusRecFunWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusRecFun}};
    case OptionEnum::SYGUS_REC_FUN_EVAL_LIMIT:
      return OptionInfo{"sygus-rec-fun-eval-limit", {}, opts.quantifiers.sygusRecFunEvalLimitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{1000, opts.quantifiers.sygusRecFunEvalLimit, {}, {}}};
    case OptionEnum::SYGUS_REPAIR_CONST:
      return OptionInfo{"sygus-repair-const", {}, opts.quantifiers.sygusRepairConstWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusRepairConst}};
    case OptionEnum::SYGUS_REPAIR_CONST_TIMEOUT:
      return OptionInfo{"sygus-repair-const-timeout", {}, opts.quantifiers.sygusRepairConstTimeoutWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{0, opts.quantifiers.sygusRepairConstTimeout, {}, {}}};
    case OptionEnum::SYGUS_REWRITER:
      return OptionInfo{"sygus-rewriter", {}, opts.datatypes.sygusRewriterWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"extended", opts.datatypes.sygusRewriter, { "basic", "extended", "none" }}};
    case OptionEnum::SYGUS_RR_SYNTH_ACCEL:
      return OptionInfo{"sygus-rr-synth-accel", {}, opts.quantifiers.sygusRewSynthAccelWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusRewSynthAccel}};
    case OptionEnum::SYGUS_RR_SYNTH_CHECK:
      return OptionInfo{"sygus-rr-synth-check", {}, opts.quantifiers.sygusRewSynthCheckWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusRewSynthCheck}};
    case OptionEnum::SYGUS_RR_SYNTH_FILTER_CONG:
      return OptionInfo{"sygus-rr-synth-filter-cong", {}, opts.quantifiers.sygusRewSynthFilterCongWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusRewSynthFilterCong}};
    case OptionEnum::SYGUS_RR_SYNTH_FILTER_MATCH:
      return OptionInfo{"sygus-rr-synth-filter-match", {}, opts.quantifiers.sygusRewSynthFilterMatchWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusRewSynthFilterMatch}};
    case OptionEnum::SYGUS_RR_SYNTH_FILTER_NL:
      return OptionInfo{"sygus-rr-synth-filter-nl", {}, opts.quantifiers.sygusRewSynthFilterNonLinearWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusRewSynthFilterNonLinear}};
    case OptionEnum::SYGUS_RR_SYNTH_FILTER_ORDER:
      return OptionInfo{"sygus-rr-synth-filter-order", {}, opts.quantifiers.sygusRewSynthFilterOrderWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusRewSynthFilterOrder}};
    case OptionEnum::SYGUS_RR_SYNTH_INPUT_NVARS:
      return OptionInfo{"sygus-rr-synth-input-nvars", {}, opts.quantifiers.sygusRewSynthInputNVarsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{3, opts.quantifiers.sygusRewSynthInputNVars, {}, {}}};
    case OptionEnum::SYGUS_RR_SYNTH_REC:
      return OptionInfo{"sygus-rr-synth-rec", {}, opts.quantifiers.sygusRewSynthRecWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusRewSynthRec}};
    case OptionEnum::SYGUS_SAMPLE_FP_UNIFORM:
      return OptionInfo{"sygus-sample-fp-uniform", {}, opts.quantifiers.sygusSampleFpUniformWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusSampleFpUniform}};
    case OptionEnum::SYGUS_SAMPLE_GRAMMAR:
      return OptionInfo{"sygus-sample-grammar", {}, opts.quantifiers.sygusSampleGrammarWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusSampleGrammar}};
    case OptionEnum::SYGUS_SAMPLES:
      return OptionInfo{"sygus-samples", {}, opts.quantifiers.sygusSamplesWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{1000, opts.quantifiers.sygusSamples, {}, {}}};
    case OptionEnum::SYGUS_SI:
      return OptionInfo{"sygus-si", {}, opts.quantifiers.cegqiSingleInvModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"none", opts.quantifiers.cegqiSingleInvMode, { "all", "none", "use" }}};
    case OptionEnum::SYGUS_SI_ABORT:
      return OptionInfo{"sygus-si-abort", {}, opts.quantifiers.cegqiSingleInvAbortWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.cegqiSingleInvAbort}};
    case OptionEnum::SYGUS_SI_RCONS:
      return OptionInfo{"sygus-si-rcons", {}, opts.quantifiers.cegqiSingleInvReconstructWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"all", opts.quantifiers.cegqiSingleInvReconstruct, { "all", "all-limit", "none", "try" }}};
    case OptionEnum::SYGUS_SI_RCONS_LIMIT:
      return OptionInfo{"sygus-si-rcons-limit", {}, opts.quantifiers.cegqiSingleInvReconstructLimitWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{10000, opts.quantifiers.cegqiSingleInvReconstructLimit, {}, {}}};
    case OptionEnum::SYGUS_SI_RCONS_P:
      return OptionInfo{"sygus-si-rcons-p", {}, opts.quantifiers.cegqiSingleInvReconstructPWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<double>{0.5, opts.quantifiers.cegqiSingleInvReconstructP, 0.0, 1.0}};
    case OptionEnum::SYGUS_SIMPLE_SYM_BREAK:
      return OptionInfo{"sygus-simple-sym-break", {}, opts.datatypes.sygusSimpleSymBreakWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"agg", opts.datatypes.sygusSimpleSymBreak, { "agg", "basic", "none" }}};
    case OptionEnum::SYGUS_STREAM:
      return OptionInfo{"sygus-stream", {}, opts.quantifiers.sygusStreamWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusStream}};
    case OptionEnum::SYGUS_SYM_BREAK_LAZY:
      return OptionInfo{"sygus-sym-break-lazy", {}, opts.datatypes.sygusSymBreakLazyWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.datatypes.sygusSymBreakLazy}};
    case OptionEnum::SYGUS_SYM_BREAK_PBE:
      return OptionInfo{"sygus-sym-break-pbe", {}, opts.datatypes.sygusSymBreakPbeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.datatypes.sygusSymBreakPbe}};
    case OptionEnum::SYGUS_SYM_BREAK_RLV:
      return OptionInfo{"sygus-sym-break-rlv", {}, opts.datatypes.sygusSymBreakRlvWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.datatypes.sygusSymBreakRlv}};
    case OptionEnum::SYGUS_UNIF_COND_INDEPENDENT_NO_REPEAT_SOL:
      return OptionInfo{"sygus-unif-cond-independent-no-repeat-sol", {}, opts.quantifiers.sygusUnifCondIndNoRepeatSolWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.sygusUnifCondIndNoRepeatSol}};
    case OptionEnum::SYGUS_UNIF_PI:
      return OptionInfo{"sygus-unif-pi", {}, opts.quantifiers.sygusUnifPiWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"none", opts.quantifiers.sygusUnifPi, { "complete", "cond-enum", "cond-enum-igain", "none" }}};
    case OptionEnum::SYGUS_UNIF_SHUFFLE_COND:
      return OptionInfo{"sygus-unif-shuffle-cond", {}, opts.quantifiers.sygusUnifShuffleCondWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.quantifiers.sygusUnifShuffleCond}};
    case OptionEnum::SYGUS_VERIFY_INST_MAX_ROUNDS:
      return OptionInfo{"sygus-verify-inst-max-rounds", {}, opts.quantifiers.sygusVerifyInstMaxRoundsWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<int64_t>{10, opts.quantifiers.sygusVerifyInstMaxRounds, {}, {}}};
    case OptionEnum::SYGUS_VERIFY_TIMEOUT:
      return OptionInfo{"sygus-verify-timeout", {}, opts.quantifiers.sygusVerifyTimeoutWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<uint64_t>{0, opts.quantifiers.sygusVerifyTimeout, {}, {}}};
    case OptionEnum::SYMMETRY_BREAKER:
      return OptionInfo{"symmetry-breaker", {}, opts.uf.ufSymmetryBreakerWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.uf.ufSymmetryBreaker}};
    case OptionEnum::TC_MODE:
      return OptionInfo{"tc-mode", {}, opts.theory.tcModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"care-graph", opts.theory.tcMode, { "care-graph" }}};
    case OptionEnum::TERM_DB_MODE:
      return OptionInfo{"term-db-mode", {}, opts.quantifiers.termDbModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"relevant", opts.quantifiers.termDbMode, { "all", "relevant" }}};
    case OptionEnum::TERM_SORT_OVERLOAD:
      return OptionInfo{"term-sort-overload", {}, opts.parser.termSortOverloadingWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.parser.termSortOverloading}};
    case OptionEnum::THEORYOF_MODE:
      return OptionInfo{"theoryof-mode", {}, opts.theory.theoryOfModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"type", opts.theory.theoryOfMode, { "term", "type" }}};
    case OptionEnum::TIMEOUT_CORE_TIMEOUT:
      return OptionInfo{"timeout-core-timeout", {}, opts.smt.toCoreTimeoutWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::NumberInfo<uint64_t>{10000, opts.smt.toCoreTimeout, {}, {}}};
    case OptionEnum::TLIMIT:
      return OptionInfo{"tlimit", {}, opts.base.cumulativeMillisecondLimitWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<uint64_t>{0, opts.base.cumulativeMillisecondLimit, {}, {}}};
    case OptionEnum::TLIMIT_PER:
      return OptionInfo{"tlimit-per", {}, opts.base.perCallMillisecondLimitWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<uint64_t>{0, opts.base.perCallMillisecondLimit, {}, {}}};
    case OptionEnum::TRACE:
      return OptionInfo{"trace", {}, false, OptionInfo::Category::REGULAR, OptionInfo::VoidInfo{}};
    case OptionEnum::TRIGGER_ACTIVE_SEL:
      return OptionInfo{"trigger-active-sel", {}, opts.quantifiers.triggerActiveSelModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"all", opts.quantifiers.triggerActiveSelMode, { "all", "max", "min" }}};
    case OptionEnum::TRIGGER_SEL:
      return OptionInfo{"trigger-sel", {}, opts.quantifiers.triggerSelModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"min", opts.quantifiers.triggerSelMode, { "all", "max", "min", "min-s-all", "min-s-max" }}};
    case OptionEnum::TYPE_CHECKING:
      return OptionInfo{"type-checking", {}, opts.expr.typeCheckingWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{DO_SEMANTIC_CHECKS_BY_DEFAULT, opts.expr.typeChecking}};
    case OptionEnum::UF_CARD_EXP:
      return OptionInfo{"uf-card-exp", {}, opts.uf.ufCardExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.uf.ufCardExp}};
    case OptionEnum::UF_HO_EXP:
      return OptionInfo{"uf-ho-exp", {}, opts.uf.ufHoExpWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.uf.ufHoExp}};
    case OptionEnum::UF_HO_EXT:
      return OptionInfo{"uf-ho-ext", {}, opts.uf.ufHoExtWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.uf.ufHoExt}};
    case OptionEnum::UF_LAMBDA_QE:
      return OptionInfo{"uf-lambda-qe", {}, opts.uf.ufHoLambdaQeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.uf.ufHoLambdaQe}};
    case OptionEnum::UF_LAZY_LL:
      return OptionInfo{"uf-lazy-ll", {}, opts.uf.ufHoLazyLambdaLiftWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.uf.ufHoLazyLambdaLift}};
    case OptionEnum::UF_SS:
      return OptionInfo{"uf-ss", {}, opts.uf.ufssModeWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"full", opts.uf.ufssMode, { "full", "no-minimal", "none" }}};
    case OptionEnum::UF_SS_ABORT_CARD:
      return OptionInfo{"uf-ss-abort-card", {}, opts.uf.ufssAbortCardinalityWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::NumberInfo<int64_t>{-1, opts.uf.ufssAbortCardinality, {}, {}}};
    case OptionEnum::UF_SS_FAIR:
      return OptionInfo{"uf-ss-fair", {}, opts.uf.ufssFairnessWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.uf.ufssFairness}};
    case OptionEnum::UF_SS_FAIR_MONOTONE:
      return OptionInfo{"uf-ss-fair-monotone", {}, opts.uf.ufssFairnessMonotoneWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.uf.ufssFairnessMonotone}};
    case OptionEnum::UNATE_LEMMAS:
      return OptionInfo{"unate-lemmas", {}, opts.arith.arithUnateLemmaModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"all", opts.arith.arithUnateLemmaMode, { "all", "eqs", "ineqs", "none" }}};
    case OptionEnum::UNCONSTRAINED_SIMP:
      return OptionInfo{"unconstrained-simp", {}, opts.smt.unconstrainedSimpWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{false, opts.smt.unconstrainedSimp}};
    case OptionEnum::UNSAT_CORES_MODE:
      return OptionInfo{"unsat-cores-mode", {}, opts.smt.unsatCoresModeWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"off", opts.smt.unsatCoresMode, { "assumptions", "off", "sat-proof" }}};
    case OptionEnum::USE_APPROX:
      return OptionInfo{"use-approx", {}, opts.arith.useApproxWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.useApprox}};
    case OptionEnum::USE_FCSIMPLEX:
      return OptionInfo{"use-fcsimplex", {}, opts.arith.useFCWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.useFC}};
    case OptionEnum::USE_PORTFOLIO:
      return OptionInfo{"use-portfolio", {}, opts.driver.usePortfolioWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.driver.usePortfolio}};
    case OptionEnum::USE_SOI:
      return OptionInfo{"use-soi", {}, opts.arith.useSOIWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{false, opts.arith.useSOI}};
    case OptionEnum::USER_PAT:
      return OptionInfo{"user-pat", {}, opts.quantifiers.userPatternsQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ModeInfo{"trust", opts.quantifiers.userPatternsQuant, { "ignore", "interleave", "resort", "strict", "trust", "use" }}};
    case OptionEnum::USER_POOL:
      return OptionInfo{"user-pool", {}, opts.quantifiers.userPoolQuantWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ModeInfo{"trust", opts.quantifiers.userPoolQuant, { "ignore", "trust", "use" }}};
    case OptionEnum::VAR_ELIM_QUANT:
      return OptionInfo{"var-elim-quant", {}, opts.quantifiers.varElimQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.varElimQuant}};
    case OptionEnum::VAR_INEQ_ELIM_QUANT:
      return OptionInfo{"var-ineq-elim-quant", {}, opts.quantifiers.varIneqElimQuantWasSetByUser, OptionInfo::Category::REGULAR, OptionInfo::ValueInfo<bool>{true, opts.quantifiers.varIneqElimQuant}};
    case OptionEnum::VERBOSE:
      return OptionInfo{"verbose", {}, false, OptionInfo::Category::COMMON, OptionInfo::VoidInfo{}};
    case OptionEnum::VERBOSITY:
      return OptionInfo{"verbosity", {}, opts.base.verbosityWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::NumberInfo<int64_t>{0, opts.base.verbosity, {}, {}}};
    case OptionEnum::VERSION:
      return OptionInfo{"version", {}, opts.driver.showVersionWasSetByUser, OptionInfo::Category::COMMON, OptionInfo::ValueInfo<bool>{false, opts.driver.showVersion}};
    case OptionEnum::WF_CHECKING:
      return OptionInfo{"wf-checking", {}, opts.expr.wellFormedCheckingWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::ValueInfo<bool>{true, opts.expr.wellFormedChecking}};
    case OptionEnum::WRITE_PARTITIONS_TO:
      return OptionInfo{"write-partitions-to", {"partitions-out"}, opts.parallel.partitionsOutWasSetByUser, OptionInfo::Category::EXPERT, OptionInfo::VoidInfo{}};
  }
  // clang-format on
  return OptionInfo{"",
                    {},
                    false,
                    OptionInfo::Category::UNDOCUMENTED,
                    OptionInfo::VoidInfo{}};
}

#undef DO_SEMANTIC_CHECKS_BY_DEFAULT

}  // namespace cvc5::internal::options
